; Disassembly of the file "astrocadeROM_3164.bin"
; Bally Astrocade On-Board ROM ("3164" Version!)
; Operating System with 4 built-in games (C)1977
;
; CPU Type: Z80 for Home Video Game System Library
;
; Machine Language created with dZ80 2.0 from Scratch
;    using script "BALLYHVG.LUA" by Richard C. Degler
;
; Re-assembles correctly with: zmac -i -m BIOS3164.ASM
;
; VERSION 2.0 can assemble all three known BIOS ROMs!!
; re-Commented [using Mixed Case] by: Richard C Degler
; starting on Monday, 04 of January 2010 at 02:10 AM
;
; note: lined up columns and cleaned out page formatting.
; Converted all numbers to $HEXA, with Decimal in comments.
; List items separated by COMMAs, MACRO lists by COMMA+SPACE
; Directives in lower case to distinguish from OPCODEs.
;

; > Home Video Game System Library - System Subroutines
; > based on the Assembly Listing in the nutting_manual
; > re-typed by Adam Trionfo and various members of the
; > Yahoo! Bally Alley users group on December 28, 2003
; >
; > BALLYROM.ASM - VERSION 1.5  - JAN. 24, 2004
; >                               CHANGES ALL $+(NAME-$) TO NAME
; >                        1.4  - MATCHES EXACTLY WITH 3159 ROM
; >                        1.32 - DEC. 26, 2003
; >                               - ASSEMBLED WITH NO ERRORS
; >                        1.3  - DEC. 25, 2003
; >                               - FINISHED TYPING [UPPER CASE]
; >                        1.15 - MAR. 1 2002
; >
; > zmac -i -m -o ballyrom.bin -x ballyrom.lst ballyrom.asm
; >
;

  ; ###### Conditional Assembly control marked by ######'s
  ; ### Important note: for proper compilation, pick ONE only!
  ; ###                                              ###
BALLY   EQU     1       ; for 3164 BIOS (rare) "(C)1977" and 9999 Gunfighters
  ; note: BALLY BIOS was only used in early Bally HOME LIBRARY COMPUTER systems
  ; Was known as 3164 ROMs due to the location of the "GAME OVER" text string.
  ; ###                                              ###
WHITE   EQU     0       ; for WHITE body from Montgomery Wards with only 99 men
  ; WHITE BIOS still "Copyright 1977" but "GAME OVER" string now starts at 3159
  ; Because of edits to several routine, the System Revision Number changed to 1
  ; ###                                              ###
ASTRO   EQU     0       ; for 3159 BIOS (common) "(C)1978" reverted to 9999 men
  ; ASTRO BIOS is used in ALL Astrocades, but also in BALLY PROFESSIONAL ARCADES
  ; Includes many color changes meant to increase contrast on B&W televisions.
  ; ###                                              ###
  ; ###  Use the --- System DOPE VECTOR --- entries  ###
  ; ####################################################

; This page intentionally left blank for double-sided printing purposes

        NOLIST

        INCLUDE "HVGLIB.H"      ; HOME VIDEO GAME LIBRARY

  ; **********
  ; * HVGSYS * HOME VIDEO GAME SYSTEM
  ; **********

if BALLY ; ### conditional assembly ###
PFUG    EQU     $04             ; # - # less POT FUDGE FACTOR
else ; WHITE or ASTRO ##### cond ###### else...
PFUG    EQU     $08             ; # + # more POT FUDGE FACTOR
endif ; ##### end conditional #########

  ; fixed locations of included games entry points marked by ==='s
GFSTRT  EQU     $17DE           ; === GUNFIGHT START ADDRESS
CMSTRT  EQU     $1328           ; === CHECKMATE START ADDRESS
CALCST  EQU     $1020           ; === CALCULATOR START ADDRESS
SCBST   EQU     $0E19           ; === SCRIBBLING START ADDRESS

  ; jump-through locations to Cartridge RST vectors
; RESET   EQU     $0000           ; Restart VECTor #0 / 0 / $00
RVECT1  EQU     $2007           ; Reset VECTor #1 /  8 / $08
RVECT2  EQU     $200A           ; Reset VECTor #2 / 16 / $10
RVECT3  EQU     $200D           ; Reset VECTor #3 / 24 / $18
RVECT4  EQU     $2010           ; Reset VECTor #4 / 32 / $20
RVECT5  EQU     $2013           ; Reset VECTor #5 / 40 / $28
RVECT6  EQU     $2016           ; Reset VECTor #6 / 48 / $30
STVECT  EQU     $2019           ; SenTry VECTor if flagged

        LIST

        ORG     0

  ; ********************
  ; * POWER UP RESTART *
  ; ********************
; L0000:
        NOP                     ; WAIT FOR THINGS TO SETTLE DOWn
        DI
        XOR     A
        OUT     (CONCM),A       ; *** SET CONSUMER MODE ***
        JP      PWRUP

  ; ****** TRANSFER CONTROL TO RESTART HANDLER
; L0008:
        JP      RVECT1          ; Restart VECTor 1 out to cartridge

NUMBAS:         ; *** NUMBers AS keymask
        DB      $1C             ; KEYpad column 3 (left side)    x x x 1 4 7 x x  T
        DB      $3C             ; KEYpad column 2 (center left)  x x 0 2 5 8 x x  O
        DB      $1C             ; KEYpad column 1 (center right) x x x 3 6 9 x x  P
        DB      $20             ; KEYpad column 0 (right side)   x x = x x x x x
;
        DB      $FF             ; fill empty space with -1 (to match ROM)

; L0010:
        JP      RVECT2          ; Restart VECTor 2 out to cartridge

  ; ****** MENU COLORS table
MENUCL:
if BALLY or WHITE ; ### conditional ###
        DB      $06             ; # = # GRAY color 3 left (copyright)
        DB      $FB             ; # + # BLUE color 2 (letters)
        DB      $07             ; # = # WHITE color 1 (playfield)
        DB      $52             ; # ~ # bright RED color 0 (border)
else ; ASTRO ####### conditional ###### else...
        DB      $06             ; # = # same GRAY
        DB      $FA             ; # - # darker BLUE
        DB      $07             ; # = # WHITE
        DB      $62             ; # ~ # another shade of RED
endif ; ##### end conditional #########

        DB      $FF             ; fill empty space with -1 (to match ROM)

; L0018:
        JP      RVECT3          ; Restart VECTor 3 out to cartridge

  ; ****** PAUSE routine
  ; Purpose:      HALT # OF INTERUPTS
  ; Input:        B = # OF INTERRUPTS
MPAUSE:
        EI
MP01:   HALT
        DJNZ    MP01                    ; (-$03)
        RET

; L0020:
        JP      RVECT4          ; Restart VECTor 4 out to cartridge

  ; ****** SET WORD routine
  ; does (HL)=DE
MSETW:
        LD      (HL),E
        INC     HL
        LD      (HL),D
        RET
;
        DB      $FF             ; fill empty space with -1 (to match ROM)

; L0028:
        JP      RVECT5          ; Restart VECTor 5 out to cartridge

CONC2:
        LD      HL,0            ; ZERO OUT HL
        RET
;
        DB      $FF             ; fill empty space with -1 (to match ROM)

; L0030:
        JP      RVECT6          ; Restart VECTor 6 out to cartridge

  ; ****** Fudge factor to make each 2K block's ChecKSUM = $FF
CKSUM:
if BALLY ; ### conditional assembly ###
        DB      $75             ; # # # CHECKSUM #1
endif ; ##### end conditional ######### or...
if WHITE ; ### conditional assembly ###
        DB      $6D             ; # # # CHECKSUM #1
endif ; ##### end conditional ######### or...
if ASTRO ; ### conditional assembly ###
        DB      $00             ; # 0 # CHECKSUM #1 (NOT used)
endif ; ##### end conditional #########

  ; ********* Screen INTerrupt vector
ITAB:
        DW      MACTIN          ; INTERRUPT TRANSFER (does ACTive INterrupts)

if BALLY ; ### conditional assembly ###
        DB      $00             ; # 0 # ** SYSTEM REVISION LEVEL = 0
else ; WHITE or ASTRO ##### cond ###### else...
        DB      $01             ; # 1 # ** SYSTEM REVISION LEVEL = 1
endif ; ##### end conditional #########

        DB      $FF             ; fill empty space with -1 (to match ROM)

  ; ****** USER PROGRAM INTERFACE handler
  ; Purpose:      TRANSFER OF CONTROL FROM USER TO SYSTEM using RST7/56/$40
  ; Input:        ROUTINE # FOLLOWS INLINE AFTER RST INSTRuction -1/255/$FF
  ;               IF L.O. BIT SET, LOAD ARGUMENTS INLINE From argument mask byte
  ; Output:       NONE
  ; Stack Use:    18 BYTES TOTAL, 16 BYTES ON EXIT
  ; Side Effects: REGISTERS AF, BC, DE, HL, IX, AND OLD IY SAVed on the stack
  ; Explanation:
  ;  REGISTERS AF, BC, DE, HL, IX, AND PREVIOUS IY ARE PUSHED before running.
  ; THE NUMBER FOLLOWING THE RST 56 INSTRUCTION IS USED TO
  ; INDEX A JUMP VECTOR GIVING THE STARTING ADDRESS OF THE
  ; SYSTEM ROUTINE TO CALL.  IF OPTIONED, INLINE ARGUMENTS
  ; ARE COPIED INTO THE CONTEXT AREA. FOR ARGUMENT ORDERINg
  ; SEE INTERPRETER DOCUMENTATION AND APPROP. TABLES
  ; A DUMMY RETURN IS INSERTED WHICH, WHEN RETURNED TO BY the
  ; SYSTEM ROUTINE, WILL RESTORE THE REGISTER CONTENTS AND then
  ; RETURN TO THE USER PROGRAM
  ;
  ;  *** THE UPI HAS BEEN EXTENDED TO SUPPORT USER SUPPLIed
  ;  ROUTINES.  IF THE CALL INDEX PROVIDED IS NEGATIVE [or from $FF through $80]
  ;  THEN THE USERS DISPATCH TABLE POINTER (USERTB) IS USed [also UMARGT.]
  ;  NOTE THAT THE SIGN BIT ISN'T ZAPPED BEFORE BEING
  ;  USED AS AN INDEX, THIS MEANS THAT THE USERS DISPATCH
  ;  TABLE POINTER SHOULD POINT 128 BYTES BEFORE THE FIRSt entry.
; L0038:                  Restart #7 / 56 / or $38 by OPCODE -1 / 255 / or $FF
        EX      (SP),HL         ; RETURN ADDRESS TO HL
        PUSH    AF              ; CREATE CONTEXT
        PUSH    BC
        PUSH    DE
        PUSH    IX
        PUSH    IY
        LD      IY,$0000        ; POINT IY AT CONTEXT
        ADD     IY,SP
        LD      A,(HL)          ; LOAD OPCODE
        INC     HL
        LD      DE,RETN_        ; DE = RETURN POINT
        RRA                     ; SUCK WANTED
        JR      C,MINT0                 ; (+$36) JUMP IF YES
INTPE:  PUSH    HL              ; SAVE PC
        PUSH    DE              ; SAVE DUMMY RETURN
        LD      HL,SYSDPT
        RLCA
        LD      E,A
        LD      D,$00
        RLA                     ; USER TABLE WANTED?
        JR      NC,PUSH1                ; (+$03)
        LD      HL,(USERTB)     ; YES    LOAD IT
PUSH1:  ADD     HL,DE
        LD      E,(HL)
        INC     HL
        LD      D,(HL)
        PUSH    DE
        LD      H,(IY+CBH)
        LD      L,(IY+CBL)
RELD:   LD      D,(IY+CBIXH)
        LD      E,(IY+CBIXL)
        PUSH    DE
        POP     IX
        LD      A,(IY+CBA)
DELOAD: LD      D,(IY+CBD)
        LD      E,(IY+CBE)
        RET                     ; CALL VIA RETURN

  ; ****** MACRO INTERPRETER routines
  ; Purpose:      INTERPRETING SEQUENCES OF SYSTEM CALLS
  ; Input:        ADDRESS OF STRING 0 TO INTERPRET PASSED ON stack
  ; Stack use:    NO INCREASE IN DEPTH
  ; Explanation:  IF OPTIONED (BIT 0 OF CALL INDEX SET) THE
  ; ARGUMENT TABLE (MRARGT) IS INDEXED GIVING A MASK WHICH
  ; SPECIFIES HOW TO TRANSFER INLINE ARGUMENTS INTO THE COntext
  ; BLOCK. THIS MASK IS FORMATTED AS FOLLOWS:
  ;
  ; **********************************
  ; * 7 * 6 * 5 * 4  * 3 * 2 * 1 * 0 *
  ; **********************************
  ; * H * L * A * IX * B * C * D * E *
  ; **********************************
  ;
  ; ARGUMENTS MUST FOLLOW THE CALL INDEX IN THE FOLLOWING order
  ; (OMITTING UNUSED ARGUMENTS, OF COURSE)
  ; (INDEX), IXL, IXH, E, D, C, B, A, L, H
  ;
  ; THE SIMULATED PC IS SAVED AND A DUMMY RETURN IS
  ; INSERTED ON THE STACK.  THE UPI DISPATCHING ROUTINE IS
  ; THEN ENTERED AT 'INTPE', WHICH EFFECTS A CONTROL TRANSfer
  ; TO THE CALLED ROUTINE.  WHEN THE CALLED ROUTINE RETURNs
  ; IT WILL COME BACK HERE TO INTERPRET THE NEXT MACRO INStruction.
  ; NOTE THAT THIS ROUTINE IS REENTRANT, THEREFORE THE CALling
  ; ROUTINE MAY RECUR BACK THRU HERE, IF IT FEELS LIKE IT.
  ;
  ;  ** THE UPI HAS BEEN EXTENDED TO PROVIDE USER PROVIDED
  ;  SYSTEM ROUTINES. IF A NEGATIVE CALL INDEX IS ENCOUNTERed
  ;  BY THE INTERPRETER, AND 'SUCK INLINE' IS OPTIONED, THE
  ;  USER MACRO ROUTINE ARGUMENT TABLE IS INDEXED FOR A
  ;  PARAMETER MASK.  THE ADDRESS OF THIS TABLE IS ASSUMED
  ;  TO BE IN (UMARGT),(UMARGT+1).  THIS POINTER SHOULD
  ;  POINT 64 BYTES BEFORE THE FIRST REAL ENTRY.
  ;  I.E. LD      HL,USERMT-64  ; WHERE USERMT POINTS AT the USER's Macro Table
  ;       LD      (UMARGT),HL
MINTPC:
        POP     DE              ; DISCARD DUMMY RETURN FROM UPI
RENTER: POP     HL              ; POP OFF PC
  ; ****** MCALL routine
  ; Purpose:      CALL INTERPRETER SUBROUTINE
  ; Input:        HL = ROUTINE ADDRESS
  ; Notes:        ROUTINE MAY BE CALLED FROM MACHINE LANGUAGE or from
  ; ANOTHER INTERPRETED SEQUENCE
  ; STACK DEPTH INCREASED BY 4 BY CALL
MMCALL:
        LD      A,(HL)          ; GET OPCODE
        INC     HL
        SRL     A
        LD      DE,RENTER       ; LOAD INTERPRETER DUMMY RETURN
MINT0:  PUSH    DE              ; SAVE DUMMY RETURN
        LD      C,A             ; INDEX TO C
        JR      NC,MINT2                ; (+$12) JUMP IF NO LOAD WANTED
        EX      DE,HL
        LD      B,$00
        LD      HL,MRARGT       ; LOAD SYSTEM ARG TABLE
        BIT     6,A             ; USE USER TABLE?
        JR      Z,MINT1                 ; (+$03) JUMP IS NO
        LD      HL,(UMARGT)
MINT1:  ADD     HL,BC           ; INDEX TABLE
        LD      B,(HL)
        CALL    MSUCK1          ; CALL SUCK ROUTINE
MINT2:  POP     DE              ; DUMMY RETURN TO DE, HL = PC
        LD      A,C             ; GET CALL INDEX BACK
        LD      B,(IY+CBB)      ; RESTORE CLOBBERED REGISTERS
        LD      C,(IY+CBC)
        JR      INTPE                   ; (-$56) JOIN NORMAL UPI DISPATCH SEQUence

  ; ****** SUCK INLINE ARGUMENTS routine
  ; Purpose:      TRANSFER OF INLINE ARGS INTO CONTEXT BLOck
  ; Input:        B = ARG LOAD MASK (SEE INTERPRETER COMMEnt)
  ; Output:       HL = UPDATED PC
  ; Explanation:  THIS ROUTINE IMPLEMENTS A MACRO LOAD INSTruction.
  ; IT IS USED BY THE INTERPRETER AS WELL. A ONE BIT IN The
  ; INLINE LOAD MASK MEANS TRANSFER THE NEXT INLINE BYTE In.
  ; A ZERO BIT MEANS 'ADVANCE CONTEXT BLOCK POINTER'
  ; TWO ENTRY POINTS ARE DEFINED, ONE FOR THE SUCK MACRO Instruction,
  ; THE OTHER FOR THE INTERPRETER TO USE
  ; SUCK MACRO ENTRY
MSUCK:
        POP     HL              ; RETURN ADDRESS TO HL
        POP     DE              ; POP OFF PC
  ; BYTE SAVING TRICK *** REPLACE WITH LD HL,REENTRY
        INC     HL              ; ADVANCE TO REENTRY (MINT0)
        PUSH    HL
  ; CALL INTO...
MSUCK1: BIT     4,B             ; IX LOAD WANTED?
        JR      Z,MSUCK2                 ; (+$0a) MSUCK2 IF NOT
        LD      A,(DE)
        INC     DE
        LD      (IY+CBIXL),A
        LD      A,(DE)
        INC     DE
        LD      (IY+CBIXH),A
MSUCK2: PUSH    IY              ; LET HL = IY
        POP     HL
        INC     HL              ; plus 4
        INC     HL
        INC     HL
        INC     HL
        RES     4,B             ; KILL IX BIT
  ; SUCK IN LOOP
MSUCK3:
        SRL     B
        JR      NC,MSUCK5               ; (+$03) MSUCK5 IF NOT THIS TIME
        LD      A,(DE)          ; GET INLINE BYTE
        INC     DE
        LD      (HL),A          ; STUFF INTO CB
MSUCK5: INC     HL              ; BUMP CB POINTER
  ; ** THIS CODE ASSUMES THAT STATUS OF 'SRL' IS PRESERVED
        JR      NZ,MSUCK3               ; (-$0a) JUMP BACK IF MORE TO DO
        EX      DE,HL           ; HL = PC
        RET                     ; THEN QUIT

  ; *****************************
  ; * UPI ROUTINE ADDRESS TABLE *
  ; *****************************
SYSDPT:
        DW      MINTPC
        DW      MXINTC
        DW      MRCALL
        DW      MMCALL
        DW      MMRET
        DW      MMJUMP
        DW      MSUCK
        DW      MACTIN
        DW      TIMEY
        DW      MUZSET
        DW      MUZSTP
        DW      MSETUP
        DW      MCOLOR
        DW      MFILL
        DW      MPAINT
        DW      MVWRIT
        DW      MWRITR
        DW      MWRITP
        DW      MWRIT
        DW      MWRITA
        DW      MVBLAN
        DW      MBLANK
        DW      MSAVE
        DW      MREST
        DW      MSCROL
        DW      DISPCH
        DW      STRNEW
        DW      BCDISP
        DW      MRELAB
        DW      MRELA1          ; RELAB1
        DW      MVECTC
        DW      MVECT
        DW      MKCTAS
        DW      MENTRY          ; SENTRY
        DW      MDOIT           ; DOIT
        DW      MDOITB
        DW      MPIZBK          ; PIZBRK
        DW      MMENU
        DW      MGETP
        DW      MGETN
        DW      MPAUSE          ; PAUSE
        DW      MDISTI          ; DISPLAY TIME
        DW      MINCSC          ; INC SCORE
        DW      INXNIB          ; INDEXN
        DW      PUTNIB          ; STOREN
        DW      MINDW           ; INDEXW
        DW      MINDB           ; INDEXB
        DW      MMOVE           ; MOVE
        DW      MSHFTU
        DW      BCDAD
        DW      BCDSB
        DW      BCDML
        DW      BCDDV
        DW      BCDCS
        DW      BCDNG
        DW      SDADD
        DW      SDSMG
        DW      SDABS
        DW      SNEGT
        DW      MRANGE
        DW      MQUIT
        DW      MSETB
        DW      MSETW
        DW      MMTD

  ; MACRO ROUTINES ARGUMENT MASK TABLE
  ; Format:
  ; **********************************
  ; * 7 * 6 * 5 * 4  * 3 * 2 * 1 * 0 *
  ; **********************************
  ; * H * L * A * IX * B * C * D * E *
  ; **********************************
  ;
  ; ARGUMENTS MUST FOLLOW THE CALL INDEX IN THE FOLLOWING order
  ; (OMITTING UNUSED ARGUMENTS, OF COURSE)
  ; (INDEX), IXL, IXH, E, D, C, B, A, L, H
MRARGT:
        DB      $00             ; INTPC
        DB      $00             ; XINTC
        DB      11000000B       ; RCALL
        DB      11000000B       ; MCALL
        DB      $00             ; MRET
        DB      11000000B       ; MJUMP
        DB      00001000B       ; SUCK
        DB      $00             ; ACTINT
        DB      00000100B       ; DECCTS
        DB      11110000B       ; BMUSIC
        DB      $00             ; EMUSIC
        DB      00101010B       ; SETOUT
        DB      11000000B       ; COLSET
        DB      00101111B       ; FILL
        DB      00101111B       ; RECTAN
        DB      11010000B       ; VWRITR
        DB      11100011B       ; WRITR
        DB      11100011B       ; WRITP
        DB      11101111B       ; WRIT
        DB      11101111B       ; WRITA
        DB      00010011B       ; VBLANK
        DB      11001011B       ; BLANK
        DB      11001111B       ; SAVE
        DB      11000011B       ; RESTORE
        DB      11001111B       ; SCROLL
        DB      00100111B       ; NEW DISCHR
        DB      11000111B       ; NEW DISSTR
        DB      11001111B       ; DISNUM
        DB      00100000B       ; RELABS
        DB      00100000B       ; RELAB1
        DB      11010100B       ; VECTC
        DB      11010000B       ; VECT
        DB      $00             ; KCTASC
        DB      00000011B       ; SENTRY
        DB      11000000B       ; DOIT
        DB      11000000B       ; DOITB
        DB      $00             ; PIZBRK
        DB      11000011B       ; MENU
        DB      11101100B       ; GET PARAMETER
        DB      11001111B       ; GET NUMBER
        DB      00001000B       ; PAUSE
        DB      00000111B       ; DISTIM
        DB      11000000B       ; INCSCR
        DB      11000000B       ; INDEXN
        DB      11000000B       ; STOREN
        DB      11000000B       ; INDEXW
        DB      11000000B       ; INDEXB
        DB      11001111B       ; MOVE
        DB      11001000B       ; SHIFTU
        DB      11001011B       ; BCDADD
        DB      11001011B       ; BCDSUB
        DB      11001011B       ; BCDMUL
        DB      11001011B       ; BCDDIV
        DB      11001000B       ; BCDCHS
        DB      00001011B       ; BCDNEG
        DB      11001011B       ; DADD
        DB      00001011B       ; DSMG
        DB      00001011B       ; DABS
        DB      11001000B       ; NEGT
        DB      00100000B       ; RANGED
        DB      $00             ; QUIT
        DB      11100000B       ; SET BYTE
        DB      11000011B       ; SET WORD
        DB      11000111B       ; MASK TO DELTAS

  ; ****** DOES 4 [Four] 60TH SEC COUNTERS IN CT0-3
MACTIN:
        DI                      ; MAKE SURE INTERRUPT IS DISABLed
        PUSH    AF
        PUSH    BC
        PUSH    DE
        PUSH    HL
        IM      2
        LD      A,ITAB SHR 8
        LD      I,A
        LD      A,$C8
        OUT     (INLIN),A
        LD      A,ITAB & 0FFH
        OUT     (INFBK),A
        CALL    TIMEZ           ; UPDATE TIMEOUT,MUSIC AND SECONds
        LD      C,$0F           ; USE CT0-3
        CALL    TIMEY           ; DEC CT0-3
        POP     HL
        POP     DE
        POP     BC
        POP     AF
        EI
        RET

  ; ****** SENTRY [SENse TRansition] routine
  ; Purpose:      TO WAIT FOR CHANGE OF PROGRAM STATUS
  ;               IN EITHER THE PORTS OR THE TIMER-COUNTERS.
  ; IN ADDITION IT CHECKS TIMOUT FOR LONG, PERIODS OF IN-
  ; ACTIVITY.
  ; ** IS VECTOR OUT FLAG SET??
MENTRY:
        LD      A,(SENFLG)
        CP      $AA
        JP      Z,STVECT        ; YES - JUMP OUT to cartridge
        LD      A,(TIMOUT)      ; CHECK IF TIME TO BLAKOUT
        OR      A
        JR      NZ,TTEST                ; (+$2b)
  ; ****** PIZza BReaK entry point
MPIZBK: XOR     A               ; TIME TO SHUT DOWN
        DI
        OUT     (VOLC),A        ; TURN OFF SOUNDS
        OUT     (VOLAB),A
        LD      BC,($08 SHL 8) + COLBX
PBLP:   OUT     (C),A           ; PAINT IT BLACK
        DJNZ    PBLP                    ; (-$04)
        LD      DE,AKEYS
        CALL    FINDL3          ; CALL STORE DE INTO CONTEXT RO
        CALL    TTEST           ; WAIT FOR SOMETHING TO HAPPEN
        INC     A
        JR      NZ,MPIZBK               ; (-$19)
        LD      (IY+CBA),$00
        EI
        LD      HL,(COLLST)     ; GET SAVED COLORS
  ; ****** COLors SET entry point
MCOLOR: LD      (COLLST),HL     ; SAVE COLORS FOR FUTURE
        LD      BC,$800 + COLBX
        OTIR                    ; RESET THE COLORS
        XOR     A
        RET

  ; *** User AWAKEN subroutine
TTEST:
        CALL    TRCHK
        LD      (IY+CBA),A
        LD      (IY+CBB),B
        CP      SKYD
        RET     C
        CP      POT0
        RET     NC
        LD      A,$FF
        LD      (TIMOUT),A
        RET

  ; ****** "CALCulator" menu Link
CALCL:  DW      SCBL            ; Next node
        DW      PNCALC          ; PriNt "CALCualtor"
        DW      CALCST          ; START OF CALCULATOR

  ; ------ SYSTEM ROUTINES JUMP VECTORs [DOPE entries marked by ---'s]
  ; --- STIMER: defined in HVGLIB.Header [valid for ALL 3 system ROMs]
; L0200:
        JP      TIMEZ           ; DO Seconds and Game TIMER & MUSIC

  ; --- CTIMER: for Custom TIMERs
; L0203:
        JP      TIMEX           ; DECTMR

  ; --- FNTSYS: FoNT descriptor for SYStem font
; L0206:
SYSFNT:
        DB      $20             ; FonT BASE character
        DB      $08             ; FonT Frame X Size width
        DB      $08             ; FonT Frame Y Size height
        DB      $01             ; FonT X size of char in BYTEs
        DB      $07             ; FonT Y SIZe height in bits
        DW      LRGCHR          ; FonT Pattern Table address

  ; --- FNTSML: FoNT descriptor for SMaLl font
; L020D:
SMLFNT:
        DB      $A0             ; FonT BASE character
        DB      $04             ; FonT Frame X Size width
        DB      $06             ; FonT Frame Y Size height
        DB      $01             ; FonT X size of char in BYTEs
        DB      $05             ; FonT Y SIZe height in bits
        DW      SMLCHR          ; FonT Pattern Table address

  ; --- ALKEYS: All KEYS mask
; L0214:
AKEYS:
        DB      $3F             ; KEYpad column 3 (left side)    x x E 1 4 7 R C  T
        DB      $3F             ; KEYpad column 2 (center left)  x x 0 2 5 8 S >  O
        DB      $3F             ; KEYpad column 1 (center right) x x . 3 6 9 H <  P
        DB      $3F             ; KEYpad column 0 (right side)   x x = + - * / %

  ; --- MENUST: head of onboard MENU STart
; L0218:
GUNLNK:
        DW      CML             ; Menu NEXT node
        DW      PNGF            ; Menu TEXT string
        DW      GFSTRT          ; Menu ITEM start

  ; --- MXSCR: "MaX SCoRe" text string
; L021E:
        DEFM    'M','A','X',$20,'S','C','O','R','E'
        DB      $00

  ; --- NOPLAY: "Number Of PLAYers" string
; L0228:
        DEFM    '#',$20,'O','F',$20,'P','L','A','Y','E','R','S'
        DB      $00

  ; --- NOGAME: "Number Of GAMEs" string
; L0235: 
        DEFM    '#',$20,'O','F',$20,'G','A','M','E','S'
        DB      $00
  ; final Dope Vector entry (as listed in HVGLIB.Header) ---

  ; ****** CONVERT MASK TO DELTAS routine
  ; Input:        B = JOYSTICK MASK
  ;               C = FLOP STATUS (MR FLOP BIT SET IF FLOP horizontally is desired)
  ;               DE = X POSITIVE DELTA
  ;               HL = Y POSITIVE DELTA
MMTD:
        CALL    CONCPL          ; HANDLE Y
        EX      DE,HL
        BIT     MRFLOP,C        ; FLOP SET?
        JR      Z,MMTD2                 ; (+$07) YES - DOIT
        LD      A,B             ; NO - GET MASK
        AND     $03
        JR      Z,MMTD1                 ; (+$01)INVERT IF NOT ZERO
        CPL
MMTD1:  LD      B,A
MMTD2   CALL    CONCPL          ; PROCESS X
        EX      DE,HL
        JP      STHLDE          ; STORE HL,DE AND QUIT

  ; *** SUBROUTINE TO CONDITIONALLY COMPLEMENT OR ZERO HL
CONCPL:
        RRC     B
        JR      NC,CONC1                ; (+$0a) JUMP IF NOT UP
        LD      A,L
        CPL
        LD      L,A
        LD      A,H
        CPL
        LD      H,A
        INC     HL
        RRC     B
        RET

CONC1:  RRC     B               ; DOWN SET?
        RET     C               ; QUIT IF SO
        JP      CONC2           ; JUMP TO ZERO OUT

  ; ****** SCROLL MEMORY BLOCK routine
  ; Input:        B = NUMBER OF LINES TO SCROLL
  ;               C = NUMBER OF BYTES ON LINE TO SCROLL
  ;               DE = LINE INCREMENT
  ;               HL = FIRST LINE TO SCROLL
MSCROL:
        XOR     A
MSCRL1: PUSH    BC              ; SAVE COUNTERS
        PUSH    DE
        LD      B,A
        EX      DE,HL
        ADD     HL,DE           ; ADD INCREMENT TO LINE
        PUSH    HL
        LDIR
        POP     HL
        POP     DE
        POP     BC
        DJNZ    MSCRL1                  ; (-$0d)
        RET

  ; ****** MACRO INTERPRETER EXIT WITH CONTEXT RESTore routine
; Purpose: QUIT INTERPRETING AND GO HOME
MXINTC:
        POP     HL              ; THROW OUT DUMMY RETURN
  ; ****** RETURN FROM SYSTEM CALL exit point
  ; Purpose:      RETURNING TO USER AND RESTORATION OF REGisters
RETN_:
        POP     HL              ; RETURN ADDRESS TO HL
        POP     IY
        POP     IX
        POP     DE
        POP     BC
        POP     AF
        EX      (SP),HL         ; STK = RETURN, HL = OLD HL
        RET

  ; ****** BCD DIVIDE routine
BCDDV:
        CALL    GNACC           ; GENERATE ACCUMULATOR
        EX      (SP),HL         ; HL = ACC, TOP = ARG2
        PUSH    BC
        LD      B,$00
        LD      A,C
        SRL     C
        ADD     HL,BC
        LD      C,A
        EX      DE,HL           ; HL = ARG1, DE = ACC
        LDIR                    ; HL = ARG1 FLAG + 1
        POP     BC
        POP     DE
        DEC     HL
        EX      (SP),HL         ; HL = ARG2, TOP = ARG1 FLAG
        PUSH    BC
        LD      B,$00
        ADD     HL,BC           ; HL = ACC + SIZE/2
        POP     BC
        DEC     C               ; DECREMENT SIZE
        EX      DE,HL           ; HL = ARG2, DE = ACC, TOP = ARG
        DEC     DE
DIV1:   DEC     DE
        XOR     A
        SYSTEM  NEGT            ; UPI decimal NEGaTe for ARG2 = -ARG2 (10s COMP)
        ;
DIV2:   SYSTEM  DADD            ; UPI Decimal ADDition to SUBTRACT UNTIL BORROW
        ;
        JR      C,DIV3                  ; (+$0a)
        INC     A               ; OR UNTIL LOOP COUNT > 99
        DAA
        JR      NZ,DIV2                 ; (-$08)
        POP     HL
        LD      (HL),$FF
        POP     BC
        JR      MULT6                   ; (+$6a)
DIV3:   SYSTEM  NEGT            ; UPI decimal NEGaTe
        ;
        SYSTEM  DADD            ; UPI Decimal ADDition
        ;
        EX      (SP),HL         ; HL = ARG1
        DEC     HL
        LD      (HL),A          ; SAVE ANSWER IN ARG1
        EX      (SP),HL
        DEC     C
        JR      NZ,DIV1                 ; (-$1d)
        POP     HL
        POP     BC
        JR      DIV4                    ; (+$55)

  ; *** SUBROUTINE TO GENERATE ACCUMULATOR ON THE STACK
GNACC:  POP     IX
        XOR     A
        LD      C,A
        SYSTEM  DABS            ; UPI Decimal ABSolute value of ARG1
        ;
        EX      DE,HL
        SYSTEM  DABS            ; UPI Decimal ABSolute value of ARG2
        ;
        EX      DE,HL           ; FLAG = 1 IF NEG ANS, ELSE POS
        LD      H,A
        LD      L,A
        LD      A,B
MULT1:  PUSH    HL              ; GENERATE ACC ON STACK
        DJNZ    MULT1                   ; (-$03)
        LD      B,A             ; RESTORE SIZE
        ADD     HL,SP
        PUSH    BC              ; SAVE SIGN
        PUSH    HL              ; SAVE STACK POINTER
        PUSH    HL              ; SAVE ACC POINTER
        LD      H,(IY+CBH)      ; RESTORE ARG2 POINTER
        LD      L,(IY+CBL)
        LD      C,B
        JP      (IX)

  ; ****** DECIMAL MULTIPLY routine
  ; Given:        DE > ARG1, HL > ARG2, B = SIZE / 2
  ;               (SIZE/2 - 1 ASSUMED EVEN)
  ; Returned:     ARG1 = ANSWER, C > 0 ON OVERFLOW
BCDML:
        CALL    GNACC           ; GENERATE ACCUM
MULT2:  LD      A,(HL)          ; A=MULT LOOP COUNT
        INC     HL
        EX      (SP),HL         ; HL>DEC ACC
        AND     A               ; IF A=0, SKIP MULT LOOP
        JR      Z,MULT4                 ; (+$09)
        EX      DE,HL
MULT3:            ; ELSE MULTIPLY
        SYSTEM  DADD            ; UPI Decimal ADDition
        ;
        AND     A               ; CLEAR THE CARRY BIT
        DEC     A               ; DECIMAL DECREMENT
        DAA
        JR      NZ,MULT3                ; (-$07)
        EX      DE,HL
MULT4:  INC     HL              ; INCREMENT DECIMAL ACC
        EX      (SP),HL         ; HL>ARG2
        DEC     C
        JR      NZ,MULT2                ; (-$14)
        POP     HL              ; RESTORE STACK POINTER
        POP     HL              ; RESTORE SIGN
        POP     BC
        PUSH    DE
        PUSH    BC
        LD      C,B
        LD      B,$00
        SRL     C
        ADD     HL,BC
        SLA     C
        LDIR
        POP     BC
        PUSH    BC              ; CHECK FOR OVERFLOW
        SRL     B
        XOR     A
MULT5:  OR      (HL)
        INC     HL
        DJNZ    MULT5                   ; (-$04)
        AND     A               ; SET FLAGS
        JR      Z,MULT7                 ; (+$03)
        LD      A,$FF
        LD      (DE),A
MULT7:  POP     BC              ; CHECK SIGN AND
        POP     HL
DIV4:   BIT     0,C             ; NEGATE ARG1 IF NECESSARY
        JR      Z,MULT6                 ; (+$02)
        SYSTEM  BCDCHS          ; UPI BCD CHange Sign
        ;
MULT6:  POP     HL              ; RESTORE ORIGINAL STACK POINTER
        DJNZ    MULT6                   ; (-$03)
        RET

  ; ****** BCD SUBTRACT & ADD routines
  ; Given:        DE > ARG1, HL > ARG2
  ;               B = SIZE/2 + 1
  ; Returned:     ARG1 = ANSWER
BCDSB:
        SYSTEM  BCDCHS          ; UPI BCD CHange Sign
        ;
BCDAD:  SYSTEM  BCDNEG          ; UPI BCD NEGate to decimal
        ;
        EX      DE,HL
        SYSTEM  BCDNEG          ; UPI BCD NEGate to decimal
        ;
        EX      DE,HL
        SYSTEM  DADD            ; UPI Decimal ADDition
        ;
  ; AND FALL INTO ...

  ; ****** DECIMAL SIGNED MAGNITUDE routine
  ; Given:        DE > ARG (10'S COMPLEMENT)
  ;               B = SIZE/2 + 1
  ; Returned:     ARG (SIGNED MAGNITUDE)
SDSMG:
        LD      L,B             ; HL > ARG3 - 1 (SIGN BYTE)
        DEC     L
        LD      H,$00
        ADD     HL,DE
        LD      A,(HL)          ; IF POS (SIGN NIBBLE < 5)
        CP      $50
        RET     C               ; EXIT
        EX      DE,HL
SDSMG1: LD      A,$00           ; ELSE 10'S COMPLIMENT
        SBC     A,(HL)
        DAA
        LD      (HL),A
        INC     HL
        DJNZ    SDSMG1                   ; (-$08)
        DEC     HL              ; AND SET SIGN BIT
        LD      A,(HL)
        OR      $80
        LD      (HL),A
        RET

  ; ****** BCD NEGATE routine
  ; Given:        DE > ARG (SIGNED MAGNITUDE)
  ;               B = SIZE/2 + 1
  ; Returned:     ARG (10'S COMPLEMENT)
BCDNG:
        LD      L,B             ; HL > ARG + B - 1 (SIGN BYTE)
        DEC     L
        LD      H,$00
        ADD     HL,DE
        BIT     7,(HL)          ; EXIT IF POS
        RET     Z
        LD      (HL),$00        ; CLEAR SIGN BYTE
        EX      DE,HL
SNEGT:  XOR     A               ; CLEAR CARRY
BCDNG1: LD      A,$00           ; ELSE 10'S COMPLEMENT
        SBC     A,(HL)
        DAA
        LD      (HL),A
        INC     HL
        DJNZ    BCDNG1                  ; (-$08)
        RET

  ; ****** DECIMAL ABSOLUTE routine
  ; Given:        DE > ARG (SIGNED MAGNITUDE)
  ;               B = SIZE/2 + 1
  ; Returned:     C = C+1 IF SIGN BIT CLEARED
SDABS:
        LD      L,B
        LD      H,$00
        DEC     L
        ADD     HL,DE
        BIT     7,(HL)
        RET     Z
        LD      (HL),$00
        INC     (IY+CBC)
        RET

  ; ****** BCD CHANGE SIGN routine
  ; Given:        HL > ARG, B = SIZE/2 + 1
  ;               (SIGNED MAGNITUDE)
  ; Returned:     ARG SIGN BIT COMPLEMENTED
BCDCS:
        LD      C,B
        LD      B,$00
        DEC     C
        ADD     HL,BC
        LD      A,(HL)
        XOR     $80
  ; ****** SET BYTE entry point
MSETB:  LD      (HL),A
        RET

  ; ****** DECIMAL ADD routine
  ; Given:        DE > ARG1, HL > ARG2 (10'S COMPLEMENT)
  ;               B = SIZE/2 + 1
  ; Returned:     ARG1 = ANSWER (10'S COMPLEMENT)
SDADD:
        XOR     A
SDADD1: LD      A,(DE)
        ADC     A,(HL)
        DAA
        LD      (DE),A
        INC     DE
        INC     HL
        DJNZ    SDADD1                  ; (-$08)
        CP      $99
        RLA
        CPL
        LD      (IY+CBFLAG),A   ; SEND BACK STATUS FROM DADD
        RET

  ; ****** RANGED RANDOM NUMBER routine
  ; Input:        A = RANGE
  ; Output:       A = RANDOM NUMBER (0 TO RANGE-1)
MRANGE:
        PUSH    AF
        LD      HL,(RANSHT)
        CALL    SHIFTR
        LD      BC,$0017
        ADD     HL,BC
        ADC     A,D
        LD      (RANSHT),HL
        LD      HL,(RANSHT+2)
        LD      E,A
        CALL    SHIFTR
        ADD     HL,DE
        LD      (RANSHT+2),HL
        LD      E,D
        EX      DE,HL
        POP     AF
        AND     A
        LD      C,A
        LD      A,D
        JR      Z,R3                    ; (+$08)
        XOR     A
R1:     ADD     HL,DE
        JR      NC,R2                   ; (+$01)
        INC     A
R2:     DEC     C
        JR      NZ,R1                   ; (-$07)
R3:     JP      QFROG

SHIFTR: LD      B,H
        LD      C,L
        XOR     A
        LD      D,$07
SH1:    ADD     HL,HL
        RLA
        DEC     D
        JR      NZ,SH1                  ; (-$05)
        ADD     HL,BC
        ADC     A,D
        RET

  ; ****** SAVE AREA routine
  ; Input:        HL = SCREEN ADDRESS
  ;               DE = SAVE AREA ADDRESS
  ;               BC = Y,X SIZE OF AREA TO SAVE
  ; Notes:        THE SIZES OF THE OBJECT ARE SAVED IN THE
  ;               FIRST TWO BYTES OF THE SAVE AREA
MSAVE:
        EX      DE,HL
        LD      (HL),C          ; SET X SIZE
        INC     HL
        LD      (HL),B          ; SET Y SIZE
        INC     HL
        XOR     A
        EX      DE,HL
        SET     6,H             ; SET NON-MAGIC ADDRESS
MSAVE1: PUSH    BC
        PUSH    HL
        LD      B,A
        LDIR
        POP     HL
        LD      C,BYTEPL
        ADD     HL,BC
        POP     BC
        DJNZ    MSAVE1                  ; (-$0c)
        RET

  ; ****** PREGAME OUTPUT PORT SETUP routine
  ; Purpose:      TO SET CONCOM [no!], VERBL, ETC
  ; Inputs:       B = HORCB, D = VERBL, A = INMOD
MSETUP:
        LD      C,HORCB         ; GET BASE PORT NUMBER
        OUT     (C),B           ; HORBL
        INC     C
        OUT     (C),D           ; VERBL
        OUT     (INMOD),A
        RET

  ; *** TEST FOR TRANSITIONS subroutine
  ; Function:     TO LOOK FOR CHANGES IN THE PORTS &TC.
  ; Returns:      A = 0 NO CHANGE
  ;                   1-8 COUNTER/TIMER #N HIT 0
  ;                   9-C = POT0-3 CHANGED
  ;                   D = A SECONDS UP
  ;                   E = KEYBOARD CHANGED (B = 0-24)
  ;                   F-16 : TRIG0!JOY0 - T3!J3
  ; RETURNS NEW VALUE IN B
CTLP:
        LD      E,(HL)
        LD      BC,$0801
CCTLP:  LD      A,C             ; GET MASK
        RRCA
        LD      C,A
        AND     E               ; CHECK IF CT BIT =1
        JR      NZ,CCT1                 ; (+$03)
        DJNZ    CCTLP                   ; (-$08)
        RET

CCT1:   XOR     E               ; MASK OUT BIT IN QUESTION
        LD      (HL),A          ; PUT BACK THE CTFLAGS OR SEMI4
        LD      A,B
        ADD     A,D
        POP     HL              ; OLD RETURN ADDRESS
        RET

  ; *** SKIP COUNTER-TIMERS AND POTS? subroutine
TRCHK:
        JR      Z,TSEX                  ; (+$25)
        LD      HL,CNT          ; GET COUNTER/TIMERS STATUS
        LD      D,$00
        CALL    CTLP            ; COUNTER/TIMERS
        LD      D,$08
        INC     HL
        CALL    CTLP            ; SEMI4S
        LD      BC,$0400 + POT0
TPLOP:  INC     HL              ; -> MPOT0
        IN      A,(C)
        LD      E,(HL)          ; GET 0POT
        SUB     E
        JR      C,PHOT                  ; (+$05) NEW LINE LESS THAN OLD
        SUB     PFUG            ; FUDGE BOUNCE FACTOR
        JR      C,EPLOP                 ; (+$06) NEW MORE THAN OLD+4
        INC     A
PHOT:   ADD     A,E
        LD      (HL),A
        LD      B,A
        LD      A,C
        RET

EPLOP:  INC     C
        DJNZ    TPLOP                   ; (-$14)
  ; NOW TEST SECONDS
TSEX:   LD      HL,KEYSEX       ; HL = KEYSEX
        LD      A,(HL)
        BIT     7,A
        JR      Z,TKEYS                 ; (+$06)
        RES     7,A
        LD      (HL),A
        LD      A,SSEC          ; SECS
        RET

  ; NOW TEST KEYBOARD
TKEYS:  PUSH    HL
if BALLY ; ### conditional assembly ###
        LD      H,(IY+CBD)      ; # v.#
        LD      L,(IY+CBE)      ; # 0 #
else ; WHITE or ASTRO ##### cond ###### else...
        CALL DELOAD             ; # v.#
        EX   DE,HL              ; # 1 #
endif ; ##### end conditional #########
        LD      BC,$0400 + KEY3
        LD      DE,$FF00        ; SET BIT COUNTER+COLUMNN
MSK1:   IN      A,(C)
        AND     (HL)
        JR      NZ,MSENK2               ; (+$0a)
        DEC     C               ; NEXT PORT
        INC     E               ; AND COLUMN
        INC     HL              ; AND MASK
        DJNZ    MSK1                    ; (-$0a)
        LD      A,B             ; NOTHING DOWN
        LD      E,SKYU
        JR      MSENKE                  ; (+$0b)

MSENK2: INC     D               ; BIT COUNTER
        RRCA
        JR      NC,MSENK2               ; (-$04)
        LD      A,D
        RLCA                    ; KEY = BIT * 4
        RLCA
        ADD     A,E             ; plus COLUMN
        INC     A               ; PLUS 1
        LD      E,SKYD
MSENKE: POP     HL
        XOR     (HL)            ; KEY = 0KEY?
        AND     $7F
        JR      Z,HANDLE                ; (+$07)
        XOR     (HL)
        LD      (HL),A
        AND     $7F
        LD      B,A
        LD      A,E             ; KEYBOARD RETURN CODE
        RET

  ; NOW TEST HANDLES
HANDLE: LD      BC,$0400 + SW0
SWLOP:  INC     HL              ; -> OSW0
        IN      A,(C)
        XOR     (HL)            ; COMPARE THE TWO
        JR      NZ,SWHIT                ; (+$05) NO CHANGE
        INC     C
        DJNZ    SWLOP                   ; (-$09)
        LD      A,B             ; RETURN 0
        RET

SWHIT:  BIT     4,A             ; TEST TRIGGER
        JR      Z,JOYS                  ; (+$0c) NO TRIG MUST BE JOYSTICK
        AND     $10             ; FILTER OUT TRIGGER
        XOR     (HL)            ; UPDATE VALUE
        LD      (HL),A
        AND     $10
        LD      B,A
        LD      A,C             ; GET PORT NUMBER
        RLCA                    ; times 2
        SUB     $0C
        RET

JOYS:   XOR     (HL)
        LD      (HL),A          ; NO CHANGE IN TRIG SO STORE STatus
        AND     $0F             ; TAKE OFF TRIGGER
        LD      B,A
        LD      A,C
        RLCA                    ; times 2
        SUB     $0B
        RET

  ; ****** TIMEX routine [always CALL the --- CTIMER vector]
  ; Inputs:       HL -> TIME BASE IN RAM
  ;               B = TIME BASE MODULUS
  ;               C = MASK AS IN DECCTS
  ; Purpose:      TO DECR TIMEBASE AND IF 0 RESET IF AND DECRement
  ;  COUNTER/TIMERS
TIMEX:
        DEC     (HL)            ; DEC TIMEBASE
        RET     NZ
        LD      (HL),B          ; RESET TIMEBASE
  ; ****** DECREMENT COUNTER/TIMERS routine [driven by Screen INTs]
  ; Inputs:       C=MASK
  ; USED BY ACTINT AND DECCTS TO DECREMENTS CTS UNDER MASK
  ; MASK = * 7 6 5 4 3 2 1 0 *, IF BIT=1 THEN DEC CORRESPONDING
  ; CT# , IF BIT=0 LEAVE CT# ALONE
  ; Note:         ALL COUNTERS ARE RUN IN BCD FOR EASY DISPLAY
TIMEY:
        LD      B,$08           ; NO OF BITS
        LD      HL,CT0          ; -> TO COUNTER/TIMERS
        LD      D,$00           ; RESULTS
TIMLP:  SRL     C               ; CHANGE THIS TIMER?
        JR      NC,ETLP                 ; (+$0a)
        LD      A,(HL)          ; GET THE TIMER
        OR      A               ; IS IT ZERO ALREADY?
        JR      Z,ETLP                  ; (+$06)
        DEC     A
        DAA
        JR      NZ,TI01                 ; (+$01)
        SCF
TI01:   LD      (HL),A          ; STORE NEW VALUE
ETLP:   INC     HL
        RR      D               ; ROTATES IN CARRY FLAG
        DJNZ    TIMLP                   ; (-$13)
        LD      A,(CNT)         ; COUNTER UPDATE & NUMBER TRACKER
        OR      D
        LD      (CNT),A
        RET

  ; ****** TIMER System ROUTINE [always CALL the --- STIMER vector]
  ; Purpose:      TO UPDATE GAME TIME, TIMEOUT AND MUSIC
  ; In & Outputs: NONE
  ; Note:         PUSH YOUR REGISTERS (AF, BC, DE, HL)
TIMEZ:
        LD      HL,PRIOR        ; ASSUMES YOU PUSH DA REGS
        BIT     1,(HL)          ; CHECK IF TICKS OVERRUN
        RET     NZ
        SET     1,(HL)
        EX      DE,HL
  ; * SIXTIETH OF A SECOND INTERRUPT *
        LD      HL,DURAT        ; NOTE TIMER
        LD      A,(HL)          ; =0 SKIP
        OR      A
        JR      Z,SIXY                  ; (+$1c)
        DEC     (HL)
        JR      NZ,STAK0                ; (+$0b)
        PUSH    HL
        PUSH    IX
        CALL    MUZCPU          ; =0 DO NEXT NOTE
        POP     IX
        POP     HL
        JR      SIXY                    ; (+$0e)

STAK0:  EX      DE,HL
        BIT     7,(HL)
        EX      DE,HL
        JR      NZ,SIXY                 ; (+$08)
        DEC     A
        DEC     A               ; =1 QUIET NOTE
        JR      NZ,SIXY                 ; (+$04)
  ; otherwise A=0
        OUT     (VOLAB),A
        OUT     (VOLC),A
SIXY:   INC     HL
        DEC     (HL)            ; IF(--TMR60<0)
        JP      P,GOUT          ; ELZ ONWARD
        LD      (HL),$3B        ; THEN TMR60=59
        INC     HL              ; -> TIMOUT
        EX      DE,HL
        LD      HL,KEYSEX       ; SET SECONDS UP
        SET     7,(HL)
        EX      DE,HL
        LD      A,(HL)          ; CHECK IF ZERO
        OR      A
        JR      Z,GTIMER                ; (+$01)
        DEC     (HL)            ; DEC TIMOUT
  ; ****** GAME TIMER ONCE A SECOND ROUTINE
  ; IF (SEC != 0 & MIN != 0)
  ;    IF (SEC == 0)
  ;       SEC = 59; --MIN
  ;    ELSE --SEC
  ; ELSE GAMETIMEUP = 1
GTIMER: INC     HL              ; -> GTSECS
        LD      A,(HL)          ; IF (SEC = 0
        INC     HL              ; -> GTMINS
        OR      (HL)            ; & MIN != 0)
        JR      Z,GT02                  ; (+$13)
        DEC     HL              ; -> GTSECS AGAIN
        LD      A,(HL)          ; IF (SEC == 0)
        OR      A
        JR      NZ,GT01                 ; (+$09)
        LD      (HL),$59        ; THEN SEC = 59BCD
        INC     HL              ; -> GTMINS AGAIN
        LD      A,(HL)          ; --MIN
        DEC     A
        DAA
        LD      (HL),A
        JR      GOUT                    ; (+$0e)

GT01:   DEC     A               ; ELSE --SEC
        DAA
        LD      (HL),A
        JR      GOUT                    ; (+$09)

GT02:   LD      HL,GAMSTB       ; ELSE GAMETIMEUP = 1
        BIT     GSBTIM,(HL)
        JR      Z,GOUT                  ; (+$02)
        SET     GSBEND,(HL)
GOUT:   LD      HL,PRIOR
        RES     1,(HL)
        RET                     ; RETURN TO BACKGND OR LO LEVEL

  ; ****** START MUZCPU routine
  ; Purpose:      TO START MUSIC PLAYING (ALSO NOISES)
  ; Inputs:       HL -> SCORE
  ;               A = VOICES
  ; Purpose:      YOU SHOULD LOAD MUZSP IF YOU DO CALLS
MUZSET:
        LD      (VOICES),A
        LD      (MUZSP),IX
        CALL    MUZSTP
        JR      MUZCP1                   ; (+$03)

  ; ****** MUZCPU subroutine [driven by TIMEZ routine]
  ; Purpose:      PLAYING MUSIC AND NOISES
  ; Note:         DURAT = 0 WHEN CALLED
  ; Output:       NONE
  ; * MUSIC PROCESSOR *
  ;   FETCH OPCODE
  ;   IF (OPCODE < $80)
  ;       SET NOTE DURATION ETC
  ;   ELSE
  ;  [ MASK = OPCODE & $0F ]
  ;   SWITCH (OPCODE & $F0)
  ;   CASE $80:
  ;       IF (MASK=8) STUFF SNDBK PC = PC+9
  ;       ELSE OUTPUT(MASK)=DATA
  ;   CASE $90:
  ;       VOICES=DATA
  ;   CASE $A0:
  ;       (--SP) = DATA IN NIBBLE OF OP + 1
  ;   CASE $B0:
  ;        SET VOLUMES = DATA,DATA
  ;   CASE $C0:
  ;       SWITCH (MASK)
  ;           CASE $09: MPCL = (MSP++); MPCH = (MSP++); BREAK
  ;           CASE $0D: (--MSP) = MPCH; (--MSP) = MPCL
  ;           CASE $00: IF --(SP) == 0; THEN SP++
  ;           CASE $03: MPC = DATA16
  ;   CASE $D0: CALL RELATIVE
  ;   CASE $E0: DURAT = DATA
  ;   CASE $F0: VOICES = 0, PORTS = 0
MUZCPU:
        LD      HL,(MUZPC)      ; LOOK LIKE NORMAL LOOP RETURN
MUZCP1: LD      IX,(MUZSP)      ; FETCH STACK POINTER
OPLOOP: LD      A,(HL)          ; OPCODE FETCH
        INC     HL              ; -> OPERAND,DATA
        OR      A               ; TEST FOR $80 OR MORE
        JP      M,MO0
  ; NORMAL NOTE OPERATOR
        LD      (DURAT),A
        LD      A,(VOICES)
        LD      BC,$0800 + SNDBX
        SRL     A               ; SET NOISE
        JR      NC,MU01                 ; (+$02)
        OUTI
MU01:   LD      B,$05           ; -> VIBRATO
        SRL     A
        JR      NC,MU02                 ; (+$02)
        OUTI                    ; SET VIBRATO
MU02:   LD      B,$04           ; -> NOTEC
M81:    SRL     A               ; CHECK C,B,A
        JR      NC,M82                  ; (+$09)
        OUTI
M815:   SRL     A               ; CHECK IF INC PC WAS ON
        JR      C,M83                   ; (+$07)
        DEC     HL              ; RESTORE PC
        JR      M83                     ; (+$04)

M82:    DEC     B
        INC     HL
        JR      M815                    ; (-$0b)

M83:    OR      A
        JR      NZ,M81                  ; (-$14)
  ; PLAY NOTE
        LD      A,(PVOLAB)
        OUT     (VOLAB),A
        LD      A,(PVOLMC)
        OUT     (VOLC),A
        JP      MUZ999

MO0:    CP      $90
        JR      NC,MO1                  ; (+$15)
  ; STUFF PORT OR SOUND BLOCK
        BIT     3,A             ; IF (STUFF SNDBLK)
        JR      Z,MO01                  ; (+$08)
        LD      A,B             ; SAVE B (VSN)
        LD      BC,($08 SHL 8) + SNDBX  ; B = 8, C = SNDBX
        OTIR                    ; HL -> NEXT OPCODE WHEN DONE
        JR      OPLOOP                  ; (-$50)

MO01:   AND     $07             ; ISOLATE PORT NUMBER
        OR      $10             ; PORTS 16-23
        LD      C,A             ; SET PORT REGISTER
        OUTI
        JR      OPLOOP                  ; (-$59)

MO1:    JR      NZ,MO2                  ; (+$07)
        LD      A,(HL)          ; GET NEW VOICES
        INC     HL
        LD      (VOICES),A
        JR      OPLOOP                  ; (-$62)

MO2:    CP      $B0
        JR      NC,MO3                  ; (+$06)
        AND     $0F
        LD      E,A
        INC     E
        JR      MO45                    ; (+$3e)

MO3:    CP      $C0             ; SET VOL ETC
        JR      NC,MO4                  ; (+$09)
  ; LOAD PVOLS
        LD      DE,PVOLAB
        LDI                     ; DONT CARE ABOUT BC
        LDI
OPLP2:  JR      OPLOOP                  ; (-$79)

MO4:    JR      NZ,MO40                 ; (+$0b)
        DEC     (IX+$00)        ; DEC STACK TOP
        JR      NZ,MO41                 ; (+$0a)
        INC     IX
        INC     HL
        INC     HL
        JR      OPLP2                   ; (-$0f)

MO40:   CP      $D0             ; PC and SP STUFF
        JR      NC,MO5                  ; (+$27)
MO41:   AND     $0F             ; ISOLATE MASK
        CP      $09             ; RETURN
        JR      NZ,MO43                 ; (+$0c)
        LD      L,(IX+$00)
        INC     IX
        LD      H,(IX+$00)
        INC     IX
        JR      OPLP2                   ; (-$25)

MO43:   LD      E,(HL)          ; PCL =
        INC     HL
        LD      D,(HL)          ; PCH =
        INC     HL
        EX      DE,HL           ; SET THE PC
        CP      $04             ; IS IT A JMP?
        JR      C,OPLP2                 ; (-$2e) IT IS
MO44:   DEC     IX              ; IT'S A CALL
        LD      (IX+$00),D      ; (--SP) = PHL
MO45:   DEC     IX
        LD      (IX+$00),E      ; (--SP) = PCL
        JR      OPLP2                   ; (-$3a)

MO5:    CP      $E0
        JR      NC,M06                  ; (+$0a)
        AND     $0F
        LD      B,$00
        LD      C,A
        LD      D,H
        LD      E,L
        ADD     HL,BC
        JR      MO44                    ; (-$1a) CALL

M06:    JR      NZ,M061                 ; (+$0a)
        LD      A,(PRIOR)       ; LEGSTA
        XOR     $80
        LD      (PRIOR),A
        JR      OPLP2                   ; (-$54)

M061:   CP      $F0             ; REST VOICE (OR SUSTAIN)
        JR      Z,MUZSTP                ; (+$12)
        LD      A,(HL)
        LD      (DURAT),A       ; SET DURATION OR QUIET
        INC     HL
        XOR     A
        OUT     (VOLAB),A
        OUT     (VOLC),A
  ; END OF MUSIC PROCESSOR
MUZ999: LD      (MUZPC),HL      ; SAVE THE PC
        LD      (MUZSP),IX      ; SAVE THE STACK POINTER
        RET

  ; ****** MUZSTP routine
  ; Purpose:      STOP MUZCPU, SET PORT TO 0
MUZSTP:
        XOR     A
        LD      (DURAT),A
        LD      (PRIOR),A
        LD      BC,$0800 + SNDBX
MU09:   OUT     (C),A
        DJNZ    MU09                    ; (-$04)
        RET

  ; ****** DO IT (by table) routines
  ; Purpose:      TRANSFER CONTROL TO USER STATE TRANSITION
  ; Input:        A or B = RETURN CODE FROM SENTRY ROUTINE
  ;               HL = DO IT TABLE ADDRESS
  ; Output:
  ; Description:  THIS ROUTINE IS USED WITH THE SENTRY ROUTine.
  ;       IT IS USED FOR DISPATCHING TO A STATE TRANSITION
  ;       ROUTINE.  THE RETURN CODE FROM SENTRY IS USED TO
  ;       SEARCH THE DOIT TABLE.  IF A MATCH IS FOUND, CONTrol is
  ;       TRANSFERRED.  IF NOT MATCH IS FOUND, THE ROUTINE REturns.
  ;
  ;       THE DOIT TABLE IS MADE UP OF THREE BYTE ENTRYS:
  ;       BYTE 0 BIT 7: IF SET - DO A MCALL TO THIS HANDLER
  ;       BYTE 0 BIT 6: IF SET - DO A RCALL TO THIS HANDLER
  ;       BYTE 0 BITS 5-0: RETURNCODE THIS ROUTINE IS TO PRocess
  ;       BYTE 1 AND 2: THE ADDRESS TO TRANSFER TO.
  ;       THE LIST IS TERMINATED BY A BYTE WHICH IS .GE. $C0
MDOITB:
        LD      A,B
  ; *** MDOIT (jump if A found in table) entry point
MDOIT:  PUSH    DE
        LD      D,A
MDOIT0: LD      A,(HL)          ; GET RETURN CODE FOR THIS ENTRy
        LD      C,A             ; C = CURRENT ENTRY
        CP      $C0             ; LIST TERMINATOR?
        JR      C,MDOIT1                ; (+$02) NO - JUMP
        POP     DE              ; YES  RETURN
        RET

MDOIT1: INC     HL
        AND     $3F
        CP      D               ; NORMAL MATCH?
        JR      Z,MDOIT2                ; (+$04) JUMP IF SO
; MDO1A:
        INC     HL              ; NO MATCH - SKIP OVER
        INC     HL              ; GO TO ADDRESS
        JR      MDOIT0                  ; (-$12)

MDOIT2: POP     DE
; MDOIT3:
        LD      E,(HL)          ; DE = GOTO ADDR
        INC     HL
        LD      D,(HL)
        EX      DE,HL
        BIT     7,C             ; MCALL?
        JP      NZ,MMCALL       ; JUMP IF SO
        BIT     6,C             ; RCALL?
        JR      NZ,MRCALL               ; (+$04)
        POP     DE              ; MUST BE JUMP
        POP     AF
        PUSH    HL
        EX      DE,HL
  ; ****** RCALL ROUTINE entry point
MRCALL: JP      (HL)

  ; **********************
  ; * VECTORING ROUTINES *
  ; **********************
  ; Name:         VECTOR X AND Y COORDINATES routines
  ; Purpose:      UPDATE X,Y COORDINATES AND LIMIT CHECK
  ; Input:        IX = VECTOR PACKET
  ;               HL = LIMITS TABLE
  ; Output:       C = TIME BASE USED
  ;               NONZERO STATUS SET IF OBJECT MOVED
  ; Notes:
  ;   THIS ROUTINE WORKS WITH A 'VECTOR PACKET', WHICH LOOks like this:
  ; ********************************
  ; *BYTE*    CONTENTS    * NAME   *
  ; ********************************
  ; * 00 * MAGIC REGISTER * VBMR   *
  ; ********************************
  ; * 01 * VECTOR STATUS  * VBSTAT *
  ; ********************************
  ; * 02 * TIME BASE      * VBTIMB *
  ; ********************************
  ; * 03 * DELTA X        * VBDXL  *
  ; * 04 *                * VBDXH  *
  ; ********************************
  ; * 05 * X COORDINATE   * VBXL   *
  ; * 06 *                * VBXH   *
  ; ********************************
  ; * 07 * X CHECKS MASK  * VBXCHK *
  ; ********************************
  ; * 08 * DELTA Y        * VBDYL  *
  ; * 09 *                * VBDYH  *
  ; ********************************
  ; * 0A * Y COORDINATE   * VBYL   *
  ; * OB *                * VBYH   *
  ; ********************************
  ; * 0C * Y CHECKS MASK  * VBYCHK *
  ; ********************************
  ;
  ; VBSTAT Options byte:
  ; BIT  MEANING
  ; ---  -------
  ;  7   VECTOR IS ACTIVE
  ;
  ; X or Y Checks byte:
  ; BIT  MEANING
  ; ---  -------
  ;  0   DO LIMIT CHECKS
  ;  1   REVERSE COORDINATES ON LIMIT ATTAINMENT
  ;  3   TARGET ATTAINED (OUTPUT)
  ;
  ;   IF THE VECTOR IS ACTIVE, AND THE TIME BASE IS NONZERo
  ; THEN UPDATE COORDINATE ROUTINE IS CALLED FOR THE X
  ; AND Y PORTIONS OF THE PACKET.
MVECT:
        SET     PSWZRO,(IY+CBFLAG) ; SET ZERO FLAG
        BIT     VBSACT,(IX+VBSTAT) ; IS VECTOR ACTIVE?
        LD      C,(IX+VBTIMB)   ; TIME BASE TO C
        LD      (IX+VBTIMB),0   ; ZERO TIME BASE
        LD      (IY+CBC),C      ; PASS BACK TIME BASE
        RET     Z
        LD      A,C
        AND     A               ; IS TIME BASE ZERO?
        RET     Z               ; QUIT IF SO
        LD      DE,VBDXL        ; ADVANCE TO FIRST
        ADD     IX,DE
        CALL    MVECTC          ; UPDATE FIRST COORDINATE
        LD      DE,VBDYL - VBDXL ; TO Y
        ADD     IX,DE
  ; AND FALL INTO...

  ; ****** VECTOR COORDINATE routine
  ; Purpose:      UPDATE OF SINGLE COORDINATE
  ; Input:        IX = POINTER TO L.O. DELTA BYTE OF VECTOR
  ;               C = TIME BASE
  ;               HL = LIMITS PACKET (IF USED)
  ; Output:       NONZERO STATUS SET IF MOTION IS OCCURED
  ;               (SHOULD BE SET ON CALL, SINCE IT IS NOT Set to 0)
  ; Notes:
  ;   THIS ROUTINE OPERATES ON A SUBSET OF THE VECTOR PACKet
  ; (BETWEEN L.O. [Low Order] DELTA BYTE AND CHECKS BYTE).
  ;   THE DELTA IS ADDED TO THE COORDINATE TIME-BASE TIMES.
  ; IF OPTIONED, LIMIT CHECKING IS DONE.  IF THE CHECK FAIls then
  ; THE COORDINATE IS SET TO THE LIMIT.
  ; WHEN THIS HAPPENS, THE LIMIT ATTAINED BIT IS SET
MVECTC:
        PUSH    HL
        LD      D,(IX+VBDCH)    ; LOAD DELTA
        LD      E,(IX+VBDCL)
        LD      H,(IX+VBCH)     ; LOAD COORDINATE
        LD      L,(IX+VBCL)
        LD      A,H             ; SAVE OLD COORDINATE FOR MOTIOn
        LD      B,C
MVECT1: ADD     HL,DE           ; ADD DELTA TO COORDINATE
        DJNZ    MVECT1                  ; (-$03) TIME-BASE TIMES
  ; HAS MOTION OCCURRED?
        CP      H
        JR      Z,MVCT1A                ; (+$04) JUMP TO SKIP TEST IF SO
        RES     PSWZRO,(IY+CBFLAG) ; SET MOVE STATUS
  ; IS LIMIT CHECK WANTED?
MVCT1A: BIT     VBCLMT,(IX+VBCCHK)
        JR      Z,MVECT6                ; (+$31) MVECT6 IF NOT
  ; PERFORM LIMIT CHECK
        LD      A,H
        EX      (SP),HL
        LD      B,(HL)          ; LIMIT TO B
        INC     HL
  ; HANDLE SLIGHTLY LESS THAN ZERO CASE
        CP      $CF             ; MIDPOINT BETWEEN 160 AND 0
        JR      NC,MVECT2               ; (+$07) JUMP TO FAIL IF >207
        CP      B               ; DO COMPARE
        JR      C,MVECT2                ; (+$04) JUMP ON FAIL
        LD      B,(HL)          ; UPPER LIMIT CHECK
        CP      B
        JR      C,MVECT3                ; (+$20) JUMP ON PASS
MVECT2: INC     HL
  ; A LIMIT WAS EXCEEDED - SET COORDINATE AT LIMIT
        LD      (IX+VBCH),B
        LD      (IX+VBCL),$00
        SET     VBCLAT,(IX+VBCCHK) ; SET LIMIT ATTAINED
  ; IS REVERSE DELTA SET
        POP     AF              ; CLEAN UP STACK
        BIT     VBCREV,(IX+VBCCHK)
        RET     Z               ; QUIT IF NOT
  ; REVERSE THE BIMBO
        LD      A,D
        CPL
        LD      D,A
        LD      A,E
        CPL
        LD      E,A
        INC     DE
        LD      (IX+VBDCL),E    ; STORE STACK
        LD      (IX+VBDCH),D
        RET

MVECT3: INC     HL              ; STEP PAST LIMIT
        EX      (SP),HL         ; HL = COORDINATE AGAIN
MVECT6: LD      (IX+VBCL),L     ; STORE BACK COORDINATES
        LD      (IX+VBCH),H
        POP     HL              ; RESTORE LIMITS POINTER
        RES     VBCLAT,(IX+VBCCHK) ; CLEAR ATTAINED BIT
        RET

  ; ***************************
  ; * PAINT RECTANGLE ROUTINE *
  ; ***************************
  ; Purpose:      PAINT RECTANGLE
  ; Input:        A = COLOR MASK TO WRITE
  ;               B = Y SIZE
  ;               C = X SIZE
  ;               D = Y COORDINATE
  ;               E = X COORDINATE
MPAINT:
        XOR     A
        CALL    RELTA1
        EX      DE,HL
        SET     6,H             ; UNMAGIC THE ADDRESS
        OUT     (MAGIC),A
if BALLY ; ### conditional assembly ###
        XOR     A               ; fl- # extra code in ver. 0 only
        LD      (WASTE),A       ; ush #
endif ; ##### end conditional #########
        LD      E,(IY+CBA)
        LD      A,C
        RRCA
        RRCA
        AND     $3F
        INC     A
        LD      D,A
MPT1:   DEC     D
        JR      Z,MPT2                  ; (+$07)
        LD      A,$FF
        CALL    STRIPE
        JR      MPT1                    ; (-$0a)

MPT2:   LD      A,C
        AND     $03
        INC     A
        LD      C,A
        XOR     A
MPT3:   DEC     C
        JR      Z,MPT4                  ; (+$06)
        RRCA
        RRCA
        ADD     A,11000000B
        JR      MPT3                    ; (-$09)

MPT4:   CALL    STRIPE
        XOR     A
  ; AND FALL INTO...

  ; ****** STRIPE PAINTER subroutine
  ; Inputs:       HL = ADDRESS OF STRIPE, A = DATA, E = MASK, B = ITERATIONS
  ; Output:       HL = HL+1, A = CLOBBERED
STRIPE:
        PUSH    HL
        PUSH    BC
        LD      (WASTE),A
        LD      A,(WASTE+$4000)
        LD      C,A
STRP1:  LD      A,E
        XOR     (HL)
        AND     C
        XOR     (HL)
        LD      (HL),A
        LD      A,L
        ADD     A,BYTEPL
        LD      L,A
        LD      A,H
        ADC     A,$00
        LD      H,A
        DJNZ    STRP1                   ; (-$0f)
        POP     BC
        POP     HL
        INC     HL
        RET

  ; ******************
  ; * WRITE ROUTINES *
  ; ******************
  ; Notes:        THE GENERAL CALLING SEQUENCE FOR THE WRIte routines is:
  ; Input:        IX = Vector Block address then  
  ;               E = X COORDINATE  [0-159]
  ;               D = Y COORDINATE  [0-101]
  ;               C = X SIZE  [in Bytes]
  ;               B = Y SIZE
  ;               A = MAGIC REGISTER
  ;               HL = PATTERN ADDRESS
  ; Output:       DE = SCREEN ADDRESS USED
  ;               A = Magic Register used  
  ;
  ; THESE ROUTINES ARE NESTED, FOR EXAMPLE VWRITR [reads XY from Vector block],
  ; WHICH FALLS INTO WRITR [WRITe with offset displacement of pattern added in],
  ; WHICH FALLS INTO WRITP [WRITe with Pattern size scare-up], WHICH FALLS INTO
  ; WRIT [WRITe to X and Y], WHICH FALLs into WRITA [WRITe to Absolute address]

  ; ****** WRITE FROM VECTOR entry point
  ; Input:        HL = PATTERN ADDRESS
  ;               IX = VECTOR ADDRESS
  ; Output:       DE, A
  ; Side Effects: BLANK BIT SET IN VECTOR STATUS BYTE
MVWRIT:
        LD      A,(IX+VBMR)     ; LOAD MR
        LD      D,(IX+VBYH)     ; LOAD Y
        LD      E,(IX+VBXH)     ; LOAD X
        SET     VBBLNK,(IX+VBSTAT) ; SET BLANK BIT
  ; ****** WRITE RELATIVE entry point
  ; Purpose:      WRITING RELATIVE PATTERNS
  ; Input:        HL, DE, A
  ; Output:       DE
  ; Notes:        PATTERN IS PRECEEDED BY RELATIVE DISPLACements
  ;               (X FIRST, THEN Y, AND PATTERN SIZE as below)
MWRITR: PUSH    AF              ; SAVE MR
        LD      A,(HL)          ; GET REL X
        INC     HL
        ADD     A,E             ; ADD TO SUPERIOR X
        LD      E,A
        LD      A,(HL)          ; SAME STORY FOR Y
        INC     HL
        ADD     A,D
        LD      D,A
        POP     AF
  ; ****** WRITE WITH PATTERN SIZE SCARE-UP entry point
  ; Purpose:      WRITING VARIABLE SIZED PATTERNS
  ; Input:        HL, DE, A
  ; Output:       DE
  ; Notes:        FIRST TWO BYTES POINTED AT BY HL ARE TAKen
  ;               TO BE PATTERN SIZES (X SIZE FIRST)
MWRITP: LD      C,(HL)          ; GET X SIZE
        INC     HL
        LD      B,(HL)          ; AND Y
        INC     HL
  ; ****** WRITE WITH COORDINATE CONVERSION entry point
  ; Input:        HL, DE, BC, A
  ; Output:       DE
MWRIT:  CALL    MRELAB          ; DO CONVERSION
  ; ****** WRITE ABSOLUTE entry point
  ; Input:        HL, BC, A AS ABOVE
  ;               and DE = ABSOLUTE SCREEN ADDRESS
MWRITA: BIT     MRFLOP,A        ; FLOP WRITE WANTED?
        JR      NZ,MWRTFL               ; (+$2c) MWRTFL IS SO
        BIT     MRXPND,A        ; EXPAND WANTED?
        JR      NZ,MWX                  ; (+$11) JUMP IF SO
  ; DO NORMAL? WRITE
        XOR     A
MWRT:   PUSH    BC
        PUSH    DE
        LD      B,A             ; ZERO REGISTER B
        LDIR                    ; WRITE A LINE
        LD      (DE),A          ; CLEAR THE SHIFTER
        POP     DE
        EX      DE,HL           ; ADVANCE TO NEXT LINE
        LD      C,BYTEPL
        ADD     HL,BC
        EX      DE,HL
        POP     BC
        DJNZ    MWRT                    ; (-$0f) LOOP IF MORE GOODIES
        RET

  ; *** WRITE EXPANDED subroutine
MWX:
        EX      DE,HL
MWX1:   PUSH    BC
        PUSH    HL
        LD      B,C
MWX2:   LD      A,(DE)
        INC     DE
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        DJNZ    MWX2                    ; (-$08)
        LD      (HL),B
        INC     HL
        LD      (HL),B
        POP     HL
        LD      C,BYTEPL
        ADD     HL,BC
        POP     BC
        DJNZ    MWX1                    ; (-$15)
        RET

  ; *** HANDLE FLOPPED CASE subroutine
MWRTFL:
        BIT     MRXPND,A        ; EXPANDED FLOPPED WRITE WANTED
        JR      NZ,MWXF                 ; (+$16) JUMP IF YEP
        XOR     A
WRFL1:  PUSH    BC
        PUSH    DE
        LD      B,A
WRFL2:  LDI
        DEC     DE
        DEC     DE
        JP      PE,WRFL2
        LD      (DE),A
        POP     DE
        EX      DE,HL           ; SAME AS NORMAL NOW ON
        LD      C,BYTEPL
        ADD     HL,BC
        EX      DE,HL
        POP     BC
        DJNZ    WRFL1                   ; (-$14)
        RET

  ; *** WRITE EXPANDED FLOPPED subROUTINE
MWXF:
        EX      DE,HL
MWXF1:  PUSH    BC
        PUSH    HL
        LD      B,C
MWXF2:  LD      A,(DE)
        INC     DE
        LD      (HL),A
        DEC     HL
        LD      (HL),A
        DEC     HL
        DJNZ    MWXF2                   ; (-$08)
        LD      (HL),B
        DEC     HL
        LD      (HL),B
        POP     HL
        LD      C,BYTEPL
        ADD     HL,BC
        POP     BC
        DJNZ    MWXF1                   ; (-$15)
        RET

  ; ****** BLANK FROM VECTOR routine
  ; Purpose:      BLANK WITH INFO LOAD FROM VECTOR
  ; Input:        IX = VECTOR
  ;               E = X SIZE
  ;               D = Y SIZE
  ; Notes:        THIS ROUTINE BLANKS TO $00
  ;               THIS ROUTINE INTERROGATES THE BLANK BIT
  ;               AND REFRAINS FROM BLANKING IF NOT SET
  ;               IF IT WAS SET, IT IS THEN RESET
MVBLAN:
        BIT     VBBLNK,(IX+VBSTAT) ; IS BLANK BIT SET?
        RET     Z               ; QUIT IF NOT
        RES     VBBLNK,(IX+VBSTAT) ; KILL BLANK BIT
        LD      H,(IX+VBOAH)    ; LOAD BLANK ADDRESS
        LD      L,(IX+VBOAL)
        BIT     MRFLOP,(IX+VBMR) ; IS FLOP SET?
        JR      Z,MVBLA1                ; (+$08) JUMP IF NOT
        LD      A,E             ; X SIZE TO A
        NEG                     ; TWOS COMPLEMENT AND ADD 1
        INC     A
        LD      C,A
        LD      B,$FF
        ADD     HL,BC           ; USE TO BACK UP SCREEN ADDRESS
  ; UNMAGIC THE BLANK ADDRESS
MVBLA1: SET     6,H
        LD      B,$00           ; ASSUME BLANK TO ZERO
  ; ****** BLANK AREA entry point
  ; Purpose:      SETTING N X M REGION TO CONSTANT
  ; Input:        HL = BLANK ADDRESS
  ;               E = X SIZE
  ;               D = Y SIZE
  ;               B = DATA TO FILL WITH
MBLANK: LD      A,BYTEPL        ; COMPUTE LINE INCREMENT
        SUB     E
        LD      C,A
        LD      A,B             ; A = DATA TO FILL WITH
MBLAN1: LD      B,E
MBLAN2: LD      (HL),A
        INC     HL
        DJNZ    MBLAN2                  ; (-$04)
        ADD     HL,BC
        DEC     D
        JR      NZ,MBLAN1               ; (-$09)
        RET

  ; ****** RESTORE AREA routine
  ; Input:        HL = SCREEN ADDRESS TO RESTORE TO
  ;               DE = SAVE AREA ADDRESS
  ; Note:         SIZES ARE LOADED FROM THE SAVE AREA
MREST:
        EX      DE,HL
        LD      C,(HL)
        INC     HL
        LD      B,(HL)
        INC     HL
        SET     6,D             ; MAKE SURE WE ARE NONMAGIC
        XOR     A
MREST1: PUSH    BC
        PUSH    DE
        LD      B,A
        LDIR
        EX      DE,HL
        POP     HL
        LD      C,BYTEPL
        ADD     HL,BC
        EX      DE,HL
        POP     BC
        DJNZ    MREST1                  ; (-$0e)
        RET

  ; ******************************
  ; * CHARACTER DISPLAY ROUTINES *
  ; ******************************
 
  ; ****** DISPLAY STRING routine
  ; Purpose:      MESSAGE DISPLAY
  ; Input:        E,D = X,Y COORDINATES
  ;               HL = STRING ADDRESS
  ;               IX = FONT DESCRIPTOR
  ; Output:       D, E ALTERED AS IN DISPLAY CHARACTER
  ; Stack Use:    4 BYTES (EXCLUDING USE BY SYSPCH)
  ; Explanation:  AS EACH CHARACTER IS BROUGHT IN, IT
  ; IS TESTED FOR BEING A LIST TERMINATOR (CHAR = 0)
  ; IF IT ISN'T, DISPLAY CHARACTER IS CALLED AND THE
  ; TEST IS REPEATED FOR THE NEXT CHARACTER.  THUS
  ; A NULL STRING IS HANDLED PROPERLY.
STRNEW:
        LD      A,(HL)          ; GET CHARACTER
        AND     A               ; BE IT A TERMINATOR
        RET     Z               ; QUIT IF SO
        JP      M,STRD1         ; DISPLAY IF ALT FONT
        CP      $64             ; SUCK IN STRING?
        JR      NC,STRD2                ; (+$06) JUMP IF YES
STRD1:  CALL    DISPCH          ; SHOW CHAR
        INC     HL              ; ADVANCE TO NEXT CHAR
        JR      STRNEW                  ; (-$10) AND LOOP

STRD2   AND     10111B          ; MAKE SUCK MASK
        LD      B,A
        INC     HL
        EX      DE,HL
        CALL    MSUCK1
        CALL    RELD
        JR      STRNEW                  ; (-$1d) GO AFTER NEXT CHARACTER

  ; *****************************
  ; * CHARACTER DISPLAY ROUTINE *
  ; *****************************
  ; Input:        A = CHARACTER
  ;               C = OPTIONS
  ;               D = Y COORDINATE
  ;               E = X COORDINATE
  ;               IX = FONT DESCRIPTOR
  ;               (ONLY IF ALTERNATE FONT SET USED)
  ; Output:       DE UPDATED TO POINT AT NEXT CHARACTER FRAme
  ; Notes:        THE OPTION BYTE IS FORMATTED AS FOLLOWS:
  ;       BITS    CONTENTS
  ;       ----    --------
  ;       0-1     OFF COLOR FOR EXPANSION
  ;       2-3     ON COLOR FOR EXPANSION
  ;       4       OR OPTION
  ;       5       XOR OPTION
  ;       6-7     ENLARGMENT FACTOR (N+1)X
  ;
  ; CHARACTERS BETWEEN $01 AND $1F, AND BETWEEN $81 AND $9F
  ; ARE INTERPRETED AS TAB CHARACTERS.  THEY CAUSE THE
  ; CURSOR REPRESENTED BY D AND E TO BE SPACED OVER N
  ; CHARACTER POSITIONS [right], WHERE N = CHAR .AND. $7F
  ; CHARACTERS BETWEEN $20 AND $7F ARE TAKEN AS REFERENCES to
  ; THE SYSTEM STANDARD 5 X 7 CHARACTER FONT.  CHARACTERS
  ; BETWEEN $A0 AND $FF REFER TO THE USER SUPPLIED ALTERNate
  ; CHARACTER FONT.  THIS FONT IS DESCRIBED BY A FONT
  ; DESCRIPTOR TABLE OF THE FOLLOWING FORMAT:
  ; ******************************
  ; * 0 * BASE CHARACTER VALUE   *
  ; ******************************
  ; * 1 * X FRAME SIZE           *
  ; ******************************
  ; * 2 * Y FRAME SIZE           *
  ; ******************************
  ; * 3 * X PATTERN SIZE (BYTES) *
  ; ******************************
  ; * 4 * Y PATTERN SIZE         *
  ; ******************************
  ; * 5 *   PATTERN TABLE        *
  ; * 6 *     ADDRESS            *
  ; ******************************
DISPCH:
        PUSH    BC
        PUSH    HL
        PUSH    IX
        AND     A
        JP      M,DISCH1        ; JUMP IF YES
        LD      IX,SYSFNT
DISCH1: CP      $20             ; IS CHAR < 20H?
        JR      NC,DISC1B               ; (+$0d) JUMP IF NOT
DISC1A: PUSH    AF              ; LOOP TO SPACE OVER
        CALL    NXTFRM
        CALL    FINDL3          ; STORE IT BACK
        POP     AF
        DEC     A
        JR      NZ,DISC1A               ; (-$0b)
        JR      DISCH5                  ; (+$3b) JUMP TO EXIT

DISC1B: SUB     (IX+FTBASE)     ; SUBTRACT BASE CHAR
        LD      E,A
        LD      D,$00
        LD      HL,$0000
        LD      C,(IX+FTBYTE)   ; MULTIPLY CHARACTER
DISCH2: LD      B,(IX+FTYSIZ)   ; BY PATTERN SIZE
DISCH3: ADD     HL,DE
        DJNZ    DISCH3                  ; (-$03)
        DEC     C
        JR      NZ,DISCH2               ; (-$09)
        LD      D,(IX+FTPTH)    ; ADD TO TABLE START
        LD      E,(IX+FTPTL)
        ADD     HL,DE
  ; COMPUTE POSITION WHERE NEXT CHARACTER WOULD GO
  ; AND SAVE
        CALL    NXTFRM          ; STEP COORDINATES TO NEXT FRAMe
        PUSH    DE              ; SAVE
        LD      B,(IX+FTYSIZ)
DISCH4: PUSH    BC
        PUSH    HL
        CALL    WRTLIN
        POP     HL
        LD      C,(IX+FTBYTE)   ; STEP TO NEXT LINE OF PATTERN
        ADD     HL,BC
        POP     BC
        LD      A,(IY+CBD)      ; ADVANCE Y COORDINATE
        ADD     A,C
        LD      (IY+CBD),A
        DJNZ    DISCH4                  ; (-$14)
        POP     DE              ; RESTORE NEW POSITION
        CALL    FINDL3          ; STUFF DE BACK INTO CONTEXT
DISCH5: POP     IX
        POP     HL
        POP     BC
        RET

  ; *** SUBROUTINE TO CONVERT ENLARGEMENT FACTOR TO ITERATION Count
  ; Input:        MODE BYTE FROM CONTEXT SAVE AREA
  ; Output:       B, A = ITERATION COUNT
DCLCTB:
        LD      A,(IY+CBC)      ; GET MODE BYTE
        RLCA
        RLCA
        AND     $03             ; ISOLATE ENLARGEMENT FACTOR
        INC     A
        LD      B,A
        XOR     A
        SCF
DCLCT1: ADC     A,A
        DJNZ    DCLCT1                  ; (-$03)
        LD      B,A
        RET

  ; *** SUBROUTINE TO UPDATE COORDINATES TO POINT AT NEXT CHARActer
  ; FRAME
  ; Input:        COORDINATES TAKEN FROM CBD, CBE IN CONTEXT
  ; Output:       UPDATED COORDINATES RETURNED IN D AND E
  ;               A, B = CLOBBERED, C= ENLARGE FACTOR CONVERT
NXTFRM:
        CALL    DCLCTB          ; GET ITERATION COUNT
        LD      C,B             ; SAVE
        LD      D,(IY+CBD)      ; GET Y COORD
        LD      A,(IY+CBE)      ; GET X COORD
NXTFR1: ADD     A,(IX+FTFSX)    ; ADD X FRAME SIZE
        DJNZ    NXTFR1                  ; (-$05) 2**ENLARGE TIMES
        CP      $A0             ; PAST RIGHT EDGE OF SCREEN?
        JR      C,NXTFR3                ; (+$09)
        LD      A,D
        LD      B,C
NXTFR2: ADD     A,(IX+FTFSY)    ; YEP - ADVANCE VERTICAL
        DJNZ    NXTFR2                  ; (-$05)
        LD      D,A
        XOR     A
NXTFR3: LD      E,A
        RET

  ; *** SUBROUTINE TO WRITE ONE LINE OF A PATTERN WITH ENLARGE
  ; AND EXPAND
  ; Entry:        HL = SOURCE, IX = FONT TABLE
WRTLIN:
        LD      C,(IX+FTBYTE)
        LD      B,$00
        PUSH    IX              ; CAPTURE STACK POINTER
        LD      IX,$0000
        ADD     IX,SP
        PUSH    IX              ; SAVE CAPTURED STACK
        POP     DE              ; DE = CAPTURED STACK
        LD      A,$0C           ; SET EXPAND TO 00,11
        OUT     (XPAND),A
        LD      A,$08           ; SET EXPAND BIT
        OUT     (MAGIC),A
        LD      A,(IY+CBC)      ; GET CONTROL BYTE
        AND     $C0             ; ISOLATE ENLARGE AMOUNT
        JR      Z,WRTL3                 ; (+$21) JUMP IF ZERO
        RLCA
        RLCA
WRTL1:  EX      DE,HL
        AND     A               ; CLEAR CARRY BIT
        SBC     HL,BC           ; COMPUTE STOCK FRAME SIZE
        SBC     HL,BC
        LD      SP,HL           ; SEIZE STACK SPACE
        RES     6,H             ; MAGNIFY THE ADDRESS
        PUSH    AF
        LD      B,C
WRTL2:  LD      A,(DE)          ; GET SOURCE BYTE
        INC     DE
        LD      (HL),A          ; EXPAND IT
        INC     HL
        LD      (HL),A
        INC     HL
        DJNZ    WRTL2                   ; (-$08)
        SLA     C
        POP     AF
        LD      HL,$0000        ; CAPTURE STACK TOP AGAIN
        ADD     HL,SP
        LD      D,H             ; SET DE=HL
        LD      E,L             ; FOR NEXT DEST COMBO
        DEC     A
        JR      NZ,WRTL1                ; (-$1f)
  ; NOW DO WRITE TO SCREEN
WRTL3:  CALL    DCLCTB          ; GET ITERATION COUNTER
        CALL    DELOAD
        LD      A,(IY+CBC)
        OUT     (XPAND),A
        AND     $30
        OR      $08
        CALL    RELTA
        EX      DE,HL
WRTL4:  PUSH    AF
        PUSH    BC
        PUSH    DE
        PUSH    HL
        LD      B,C
WRTL5:  LD      A,(DE)
        INC     DE
        LD      (HL),A
        INC     HL
        LD      (HL),A
        INC     HL
        DJNZ    WRTL5                   ; (-$08)
        LD      A,(IY+CBE)
        AND     $03
        JR      Z,WRTL6                 ; (+$01)
        LD      (HL),B
WRTL6:  POP     HL              ; STEP TO NEXT LINE
        LD      C,BYTEPL
        ADD     HL,BC
        POP     DE
        POP     BC
        POP     AF
        OUT     (MAGIC),A
        DJNZ    WRTL4                   ; (-$20)
        LD      SP,IX           ; RESTORE STACK
        POP     IX
        RET

  ;
  ; MACRO TO GENERATE CHARACTER PATTERN TABLE ENTRY
  ;
DEFCHR  macro   AA, BB, CC, DD, EE, FF, GG
        DB      AA
        DB      BB
        DB      CC
        DB      DD
        DB      EE
        DB      FF
        DB      GG
        endm

  ; ********* LARGE CHARACTER SET (8 X 8) font
LRGCHR:
        DEFCHR  $00, $00, $00, $00, $00, $00, $00  ; SPACE
        DEFCHR  $20, $20, $20, $20, $20, $00, $20  ; !
        DEFCHR  $50, $50, $50, $00, $00, $00, $00  ; "
        DEFCHR  $48, $48, $FC, $48, $FC, $48, $48  ; #
        DEFCHR  $20, $78, $80, $70, $08, $F0, $20  ; $
        DEFCHR  $C0, $C8, $10, $20, $40, $98, $18  ; %
        DEFCHR  $60, $90, $A0, $40, $A8, $90, $68  ; &
        DEFCHR  $60, $60, $60, $00, $00, $00, $00  ; '
        DEFCHR  $10, $20, $20, $20, $20, $20, $10  ; (
        DEFCHR  $40, $20, $20, $20, $20, $20, $40  ; )
        DEFCHR  $00, $A8, $70, $D8, $70, $A8, $00  ; *
        DEFCHR  $00, $20, $20, $F8, $20, $20, $00  ; +
        DEFCHR  $00, $00, $00, $60, $60, $20, $40  ; ,
        DEFCHR  $00, $00, $00, $F8, $00, $00, $00  ; -
        DEFCHR  $00, $00, $00, $00, $00, $60, $60  ; .
        DEFCHR  $00, $08, $10, $20, $40, $80, $00  ; /
        DEFCHR  $70, $88, $88, $88, $88, $88, $70  ; 0
        DEFCHR  $20, $60, $20, $20, $20, $20, $70  ; 1
        DEFCHR  $70, $88, $08, $70, $80, $80, $F8  ; 2
        DEFCHR  $70, $88, $08, $30, $08, $88, $70  ; 3
        DEFCHR  $10, $30, $50, $90, $F8, $10, $10  ; 4
        DEFCHR  $F8, $80, $F0, $08, $08, $88, $70  ; 5
        DEFCHR  $30, $40, $80, $F0, $88, $88, $70  ; 6
        DEFCHR  $F8, $08, $10, $20, $40, $40, $40  ; 7
        DEFCHR  $70, $88, $88, $70, $88, $88, $70  ; 8
        DEFCHR  $70, $88, $88, $78, $08, $10, $60  ; 9
        DEFCHR  $00, $60, $60, $00, $60, $60, $00  ; :
        DEFCHR  $60, $60, $00, $60, $60, $20, $40  ; ;
        DEFCHR  $10, $20, $40, $80, $40, $20, $10  ; <
        DEFCHR  $00, $00, $F8, $00, $F8, $00, $00  ; =
        DEFCHR  $40, $20, $10, $08, $10, $20, $40  ; >
        DEFCHR  $70, $88, $08, $10, $20, $00, $20  ; ?
        DEFCHR  $70, $88, $B8, $A8, $B8, $80, $78  ; @
        DEFCHR  $70, $88, $88, $F8, $88, $88, $88  ; A
        DEFCHR  $F0, $88, $88, $F0, $88, $88, $F0  ; B
        DEFCHR  $70, $88, $80, $80, $80, $88, $70  ; C
        DEFCHR  $F0, $88, $88, $88, $88, $88, $F0  ; D
        DEFCHR  $F8, $80, $80, $E0, $80, $80, $F8  ; E
        DEFCHR  $F8, $80, $80, $E0, $80, $80, $80  ; F
        DEFCHR  $70, $88, $80, $80, $98, $88, $78  ; G
        DEFCHR  $88, $88, $88, $F8, $88, $88, $88  ; H
        DEFCHR  $70, $20, $20, $20, $20, $20, $70  ; I
        DEFCHR  $08, $08, $08, $08, $08, $88, $70  ; J
        DEFCHR  $88, $90, $A0, $C0, $A0, $90, $88  ; K
        DEFCHR  $80, $80, $80, $80, $80, $80, $F8  ; L
        DEFCHR  $88, $D8, $A8, $A8, $88, $88, $88  ; M
        DEFCHR  $88, $C8, $A8, $98, $88, $88, $88  ; N
        DEFCHR  $F8, $88, $88, $88, $88, $88, $F8  ; O
        DEFCHR  $F0, $88, $88, $F0, $80, $80, $80  ; P
        DEFCHR  $70, $88, $88, $88, $A8, $90, $68  ; Q
        DEFCHR  $F0, $88, $88, $F0, $A0, $90, $88  ; R
        DEFCHR  $70, $88, $80, $70, $08, $88, $70  ; S
        DEFCHR  $F8, $20, $20, $20, $20, $20, $20  ; T
        DEFCHR  $88, $88, $88, $88, $88, $88, $70  ; U
        DEFCHR  $88, $88, $88, $50, $50, $20, $20  ; V
        DEFCHR  $88, $88, $88, $A8, $A8, $D8, $88  ; W
        DEFCHR  $88, $88, $50, $20, $50, $88, $88  ; X
        DEFCHR  $88, $88, $50, $20, $20, $20, $20  ; Y
        DEFCHR  $F8, $08, $10, $20, $40, $80, $F8  ; Z
        DEFCHR  $70, $40, $40, $40, $40, $40, $70  ; [
        DEFCHR  $00, $80, $40, $20, $10, $08, $00  ; \
        DEFCHR  $70, $10, $10, $10, $10, $10, $70  ; ]
        DEFCHR  $20, $70, $A8, $20, $20, $20, $20  ; ^
        DEFCHR  $00, $20, $40, $F8, $40, $20, $00  ; <-
        DEFCHR  $20, $20, $20, $20, $A8, $70, $20  ; DOWN
        DEFCHR  $00, $20, $10, $F8, $10, $20, $00  ; RIGHT
        DEFCHR  $00, $88, $50, $20, $50, $88, $00  ; MULTI
        DB      $00,$20,$00,$F8,$00,$20
  ; ** LAST BYTE OF DIVIDE IS ZERO, WHICH HAPPENS TO BE
  ; FIRST BYTE OF..

  ; ********* SMALL CHARACTERS (4 X 6) font
SMLCHR:
        DEF5    $00, $00, $00, $00, $00  ; SPACE

  ; ****** Macro JUMP (real OR interpreted) routine
MMJUMP:
        POP     IX
        EX      (SP),HL
        JP      (IX)

  ; ****** CONVERT KEY CODE TO ASCII routine
  ; Purpose:      SAME
  ; Input:        A = KEY CODE
  ; Output:       A = ASCII EQUIVALENT
  ; How:          TABLE LOOKUP
MKCTAS:
        LD      C,B
        LD      B,$00
        LD      HL,KCTATB
        ADD     HL,BC
        LD      A,(HL)
QFROG:  LD      (IY+CBA),A
        RET

  ; ********* Key Code To Ascii TaBle
KCTATB:
        DB      ' '             ; SPACE
        DB      'C'             ; BULLET
        DB      $5E             ; UP ARROW
        DB      $5C             ; DOWN ARROW
        DB      '%'
        DB      'R'             ; RECALL
        DB      'S'             ; STORE
        DB      ';'             ; PLUS MINUS
        DB      '/'             ; DIVIDE
        DB      '7'
        DB      '8'
        DB      '9'
        DB      '*'             ; TIMES
        DB      '4'
        DB      '5'
        DB      '6'
        DB      '-'             ; MINUS
        DB      '1'
        DB      '2'
        DB      '3'
        DB      '+'             ; PLUS
        DB      '&'             ; CE
        DB      '0'
        DB      '.'             ; POINT
        DB      '='             ; EQUALS

  ; ****** FILL AREA routine
  ; Purpose:      SET REGION OF SCREEN TO CONSTANT VALUE
  ; Input:        A = DATA TO FILL WITH
  ;               BC = NUMBER OF BYTES TO FILL
  ;               DE = STARTING ADDRESS OF REGION TO FILL
MFILL:
        EX      DE,HL
MFILL1: LD      (HL),A          ; STUFF BYTE
        CPI                     ; BUMP HL, DEC BC
        JP      PE,MFILL1
        RET

  ; ****** RELATIVE TO ABSOLUTE routines
  ; Purpose:      COORDINATE CONVERSION
  ; Input:        E = X COORDINATE
  ;               D = Y COORDINATE
  ;               A = MAGIC REGISTER
  ; Output:       DE = ABSOLUTE ADDRESS
  ;               A = MAGIC REGISTER TO USE
  ; *** MAGIC ENTRY POINT
MRELAB:
        CALL    RELTA
        JR      MRELA2                  ; (+$05)

  ; *** NONMAGIC ENTRY POINT
MRELA1:
        CALL    RELTA1
        SET     6,D             ; NONMAGIC THE ADDRESS
MRELA2: LD      (IY+CBE),E      ; UPDATE CB DE
        LD      (IY+CBD),D
MFROG:  JR      QFROG                   ; (-$37)

  ; *** MAGIC ENTRY POINT
RELTA:
        CALL    RELTA1
        OUT     (MAGIC),A
        RET

  ; ****** ChecKSUM the 2nd
if BALLY ; ### conditional assembly ###
CKSUM2: DB      $B6             ; # # # *** CHECKSUM #2 ***
endif ; ##### end conditional ######### or...
if WHITE ; ### conditional assembly ###
CKSUM2: DB      $18             ; # # # *** CHECKSUM #2 ***
endif ; ##### end conditional ######### or...
if ASTRO ; ### conditional assembly ###
CKSUM2: DB      $00             ; # 0 # *** CHECKSUM #2 ***
endif ; ##### end conditional #########

  ; ****** more of SMALL CHARACTERS (4 X 6) font
        DEF5    $E0, $A0, $A0, $A0, $E0  ; 0
        DEF5    $40, $40, $40, $40, $40  ; 1
        DEF5    $E0, $20, $E0, $80, $E0  ; 2
        DEF5    $E0, $20, $60, $20, $E0  ; 3
        DEF5    $A0, $A0, $E0, $20, $20  ; 4
        DEF5    $E0, $80, $E0, $20, $E0  ; 5
        DEF5    $E0, $80, $E0, $A0, $E0  ; 6
        DEF5    $E0, $20, $20, $20, $20  ; 7
        DEF5    $E0, $A0, $E0, $A0, $E0  ; 8
        DEF5    $E0, $A0, $E0, $20, $E0  ; 9
        DEF5    $00, $40, $00, $40, $00  ; :
        DEF5    $40, $E0, $E0, $E0, $E0  ; BULLET

  ; ****** MOVE ROUTINE
MMOVE:
        LDIR
        RET

  ; *** SYSTEM ENTRY POINT FOR NONMAGIC ADDRESS
RELTA1:
        PUSH    HL
        AND     $FC             ; TOSS OUT SHIFT AMOUNT
        LD      L,A             ; SAVE
        LD      A,E             ; GET X
        AND     $03             ; ISOLATE SHIFT AMOUNT
        OR      L               ; COMBINE WITH MR
RELTA2: PUSH    AF
        AND     $40             ; IS FLOPPED BIT SET?
        LD      A,E
        JR      Z,RELTA3                 ; (+$03) JUMP IF NOT
        CPL                     ; YEP - UNFLOP THE COORDINATE
        ADD     A,$A0           ; 160
RELTA3: LD      L,D             ; HL = Y
        LD      H,$00
        ADD     HL,HL           ; SET HL = Y * 8
        ADD     HL,HL
        ADD     HL,HL
        LD      D,H
        LD      E,L
        ADD     HL,HL           ; SET HL = Y * 32
        ADD     HL,HL
        ADD     HL,DE           ; SET HL = Y * 40
        SRL     A               ; A = X / 4
        SRL     A
        LD      E,A
        LD      D,$00
        ADD     HL,DE           ; HL = Y * 40 + X divided by 4
        EX      DE,HL
  ; ****** RETURN FROM MACRO SUBROUTINE
  ; Purpose:      RETURN CONTROL TO CALLER
  ; THIS CODE WAS 'STOLEN' FROM RELABS SINCE
  ; IT DOES THE STACK CLEANUP THAT MRET DOES
MMRET:  POP     AF
        POP     HL
        RET

  ; ****** INDEXN routine (ENTRY FOR USER)
INXNIB:
        CALL    XNIB
        JR      MFROG                   ; (-$75)

  ; *** INDEX NIBBLE subroutine
  ; Purpose:      LOAD OF SPECIFIED NIBBLE RELATIVE TO BASE
  ; Input:        C = NIBBLE NUMBER
  ;               HL = BASE ADDRESS
  ; Output:       NIBBLE RETURNED RIGHT JUSTIFIED IN A byte
  ; Description:  BYTE = NIBBLE# divided by 2 + BASE
  ;  THE LOW ORDER NIBBLE OF A GIVEN BYTE IS ADDRESSED
  ; BY AN EVEN NIBBLE NUMBER.
XNIB:
        PUSH    HL
        PUSH    BC
        LD      B,$00
        SRL     C
        ADD     HL,BC
        LD      A,(HL)
        POP     BC
        BIT     0,C
        JR      Z,XNIB1                 ; (+$04)
        RRCA
        RRCA
        RRCA
        RRCA
XNIB1:  AND     $0F
        POP     HL
        RET

  ; ****** STORE NIBBLE routine
  ; Purpose:      NIBBLE STORING (!)
  ; Input:        A = NIBBLE TO STORE
  ;               C = NIBBLE NUMBER (AS IN XNIB)
  ;       HL = BASE ADDRESS
PUTNIB:
        PUSH    HL
        PUSH    BC
        LD      B,$00
        SRL     C
        ADD     HL,BC
        POP     BC
        BIT     0,C
        JR      Z,PUTNB1                ; (+$09)
  ; L.O. CASE - SHIFT IT
        RLCA
        RLCA
        RLCA
        RLCA
        XOR     (HL)
        AND     $F0
        JR      PUTNB2                  ; (+$03)

PUTNB1: XOR     (HL)            ; L.O. CASE
        AND     $0F
PUTNB2: XOR     (HL)
        LD      (HL),A
        POP     HL
        RET

  ; ****** INDEX WORD TABLE (WORD INDEX) routine
  ; Purpose:      TO INDEX AN ARRAY OF DEFW'S
  ; Inputs:       A = INDEX NUMBER (0-255)
  ;               HL -> TABLE ENTRY 0
  ; Outputs:      DE = ENTRY LOOKED UP
  ;               HL = POINTER TO ENTRY IN TABLE
MINDW:
        LD      E,A
        LD      D,$00
        SLA     E
        RL      D               ; DE*2
        ADD     HL,DE
        LD      E,(HL)
        INC     HL
        LD      D,(HL)
        DEC     HL
STHLDE: CALL    FINDL3
        JR      MINDB1                  ; (+$08) JOIN STORE IN INDEX BYTE

  ; ****** INDEX BYTE TABLE routine
  ; Purpose:      TABLE LOOKUP
  ; Input:        A = INDEX NUMBER
  ; Outputs:      A = VALUE OF BYTE
  ;               HL = POINTER TO TABLE ENTRY
MINDB:
        LD      E,A
        LD      D,$00
        ADD     HL,DE
        LD      A,(HL)
        LD      (IY+CBA),A
MINDB1: LD      (IY+CBH),H
        LD      (IY+CBL),L
        RET

  ; ****** DISPLAY TIME routine
  ; Purpose:      DISPLAY TIME ON SCREEN
  ; Inputs:       E = X COORD
  ;               D = Y COORD
  ;               C = SAME AS DISCHR OPTIONS EXCEPT BIT 7 = 1
  ;               TO DISPLAY COLON AND SECONDS
  ; Output:       NONE
MDISTI:
        LD      IX,SMLFNT
        LD      B,$42
        LD      HL,GTMINS
        PUSH    BC
        RES     7,(IY+CBC)
        CALL    BCDISP          ; show "hh:mm"
        POP     BC
        BIT     7,C
        RET     Z
        LD      A,$80 + $3A     ; small font colon
        CALL    DISPCH
        LD      B,$42
        LD      HL,GTSECS
  ; AND FALL INTO...

  ; ****** DISPLAY BCD NUMBER routine
  ; Inputs:       B = NUMBER DISPLAY OPTIONS
  ;               C = CHARACTER DISPLAY OPTIONS
  ;               DE = Y,X COORDINATES
  ;               HL = NUMBER ADDRESS (POINTS AT LO BYTE)
  ;               IX = ALTERNATE FONT (IF USED)
  ; Output:       DE UPDATED
  ; Description:  THIS ROUTINE CONVERTS EACH NIBBLE INTO
  ; ASCII AND DISPLAYS IT.  THE NORMALLY ILLEGAL BCD
  ; VALUES ARE DISPLAYED AS CODES $2A THRU $2F RESPECTIVELY.
  ; THE NUMBER DISPLAY OPTIONS BYTE IS FORMATED AS FOLLOWS:
  ; BIT 7         SET IF LEADING ZERO SUPPRESSION WANTED
  ; BIT 6         SET IF USE OF ALTERNATE FONT WANTED
  ; BITS 5-0      NUMBER OF DIGITS TO DISPLAY (NOT NUMBER 0)
BCDISP:
        LD      A,B             ; GET OPTIONS BYTE
        AND     $3F             ; ISOLATE NUMBER OF DIGITS
BCDD0:  DEC     A
        RET     M               ; QUIT IF NULL OR NO MORE
        LD      C,A             ; SAVE
        CALL    XNIB            ; GET NEXT DIGIT
        JR      NZ,BCDD1                ; (+$07) JUMP IF NONZERO
        BIT     7,B             ; IS ZERO SUPPRESS ON?
        JR      Z,BCDD1                 ; (+$03) JUMP IF NOT
        OR      C               ; LAST DIGIT?
        JR      NZ,BCDD4                ; (+$14)
BCDD1:  RES     7,B             ; CLEAR LEADING ZERO FLAG
        ADD     A,$06
        AND     $0F
        ADD     A,$2A
BCDD2:  BIT     6,B             ; ALTERNATE FONT?
        JR      Z,BCDD3                 ; (+$02) JUMP IF SO
        OR      $80             ; YEA, SET THE BIT
BCDD3:  CALL    DISPCH          ; DISPLAY THE CHAR
        LD      A,C             ; GET LOOP COUNTER IN A
        JR      BCDD0                   ; (-$23) AND GO FOR NEXT

BCDD4:
        LD      A,' '           ; LEADING ZERO - WRITE A SPACE
        JR      BCDD2                   ; (-$10)

  ; ****** INCREMENT SCORE routine
  ; Purpose:      INCREMENT SCORE AND COMPARE TO END SCORE
  ; Input:        HL -> PLAYER SCORE LOW ADDR OF 3 BYTES
  ; Output:       GSBEND OF GAMSTB SET IF MAX SCORE REACHED
MINCSC:
        LD      B,$03
        PUSH    HL
INCLOP: LD      A,(HL)
if BALLY ; ### conditional assembly ###
        INC     A               ; # ! # in version 0
else ; WHITE or ASTRO ##### cond ###### else...
        ADD     A,1             ; # ? # in version 1
endif ; ##### end conditional #########
        DAA
        LD      (HL),A
        JR      NZ,CMPIT                ; (+$03)
        INC     HL
        DJNZ    INCLOP                  ; (-$09)
CMPIT:  POP     HL
        INC     HL
        INC     HL
        LD      A,(GAMSTB)
        BIT     GSBSCR,A
        RET     Z
        LD      DE,ENDSCR + 2
        LD      B,$03
CMPLOP: LD      A,(DE)
        CP      (HL)
        JR      Z,REPEAT                ; (+$07) ENDSCR = SCORE
        RET     NC              ; ENDSCR > SCORE
SETEND: LD      HL,GAMSTB       ; ENDSCR < SCORE
        SET     GSBEND,(HL)
        RET

REPEAT: DEC     DE
        DEC     HL
        DJNZ    CMPLOP                  ; (-$0f)
        JR      SETEND                  ; (-$0c)

  ; ****** QUIT routine
  ; Purpose:      HOLD PRESENT GAME SCORE UNTIL KEY HIT OR Trigger 1 pulled
  ;               SAY GAME OVER
MQUIT:
        SYSSUK  STRDIS          ; UPI STRing DISplay "GAME OVER"
        DB      $30             ; ... X = 48
        DB      $18             ; ... Y = 24
        DB      01001100B       ; ... Options (2x in color 3 on color 0)
        DW      GMOVR           ; ... String Address
        ;
        SYSTEM  ACTINT          ; UPI ACTivate sub timer INTerrupts
        ;
MQUIT1: SYSSUK  SENTRY          ; UPI SENse TRansition & WAIT FOR SOMETHING TO HAPPEN
        DW      AKEYS           ; ... Keymask Address = ALL KEYS
        ;
        CP      ST0
        JR      Z,MQUIT2                ; (+$04) TRIGGER CHANGE?
        CP      SKYD            ; KEY HIT?
        JR      NZ,MQUIT1               ; (-$0c) NO - KEEP GOING
MQUIT2: RST     $00             ; YES, RESET

  ; ****** "GaMe      OVeR" text string
GMOVR:
        DB      'G','A','M','E'
        DB      $06             ; six spaces
        DB      'O','V','E','R'
        DB      $00

  ; *****************
  ; * MENU ROUTINES *
  ; *****************

NOLINE  EQU     $60             ; NUMBER OF DISPLAYED LINES = 96
MNNL    EQU     +0              ; NEXT FIELD
MNNH    EQU     +1
MNSAL   EQU     +2              ; STRING ADDRESS
MNSAH   EQU     +3
MNGL    EQU     +4              ; GO TO ADDRESS
MNGH    EQU     +5

  ;
  ; ****** SYSTEM POWER UP ROUTINE
  ;
PWRUP:
        LD      A,(FIRSTC)      ; GET FIRST CASSETTE LOCATION
        CP      $C3             ; IS IT A JUMP?
        JP      Z,FIRSTC        ; JUMP TO IT IF SO
        LD      SP,BEGRAM
        SYSSUK  FILL            ; UPI FILL to Clear SYSTEM RAM
if BALLY ; ### conditional assembly ###
        DW      $4000           ; # v.# ... Memory Address (NORMEM)
        DW      $0040           ; # 0 # ... Byte Count = 64 [why ??]
else ; WHITE or ASTRO ##### cond ###### else...
        DW      BEGRAM          ; # v.# ... Memory Address ($4FCE)
        DW      $0032           ; # 1 # ... Byte Count = 50 (to $4FFF)
endif ; ##### end conditional #########
        DB      $00             ; ... Data = 0
        ;
        LD      (WASTE),A       ; CLEAR SHIFTER
        DEC     A
        LD      (TIMOUT),A      ; CLEAR TIMEOUT WATCHDOG
        SYSTEM  INTPC           ; UPI INTerPret with Context create
        ;
        DO      EMUSIC          ; UPI End playing MUSIC
        ;
        DO      SETOUT          ; UPI SET some OUTput ports
        DB      NOLINE*2 - 1    ; ... VERBL*2 = 191 ($BF)
        DB      $29             ; ... HORCB/4 = 41
        DB      $08             ; ... INMOD = 8
        ;
        DO      COLSET          ; UPI COLors SET
        DW      MENUCL          ; ... Table Address
        ;
        DO      ACTINT          ; UPI ACTivate sub timer INTerrupts
        ;
        DONT    XINTC           ; UPI eXit INTerpreter with Context
        ;
        LD      DE,GAMSTR       ; "SELECT GAME" AS TITLE
        LD      HL,FIRSTC       ; ASSUME MENU STARTS IN CASSETTe
        LD      A,(HL)          ; GET FIRST CASSETTE BYTE
        INC     HL
        CP      $55             ; IS SENTINEL THERE?
        JR      Z,PWRUP1                ; (+$03) YEP - JUMP
        LD      HL,GUNLNK       ; WRONG - USE ONBOARD ONLY
PWRUP1: SYSTEM  MENU            ; UPI display the MENU
        ;

  ; ****** DISPLAY MENU AND BRANCH ON CHOICE routine 
  ; Inputs:       HL = MENU LIST
  ;               DE = MENU TITLE
  ; Output:       DE = TITLE OF SELECTION MADE
  ; Description:
  ;       THE MENU LIST IS A LINKED LIST OF THE FOLLOWING Fields:
  ; *************************
  ; * 0 * NEXT ENTRY        *
  ; * 1 *                   *
  ; *************************
  ; * 2 * STRING ADDRESS    *
  ; * 3 *                   *
  ; *************************
  ; * 4 * BRANCH TO ADDRESS *
  ; * 5 *                   *
  ; *************************
  ; THIS LIST IS TERMINATED BY A NEXT ENTRY FIELD OF ZEROS
  ; A MAXIMUM OF EIGHT ENTRYS MAY BE DISPLAYED.
MMENU:
        PUSH    HL
        PUSH    HL
        CALL    MNCLR           ; CLEAR SCREEN AND THROWUP TITLE
        XYRELL  DE, $10, $0C
        LD      BC,$0109        ; INITIALIZE ENTRY # AND COLOR
MMENU1: POP     IX              ; FIRST ENTRY TO IX
        LD      A,B             ; SELECTION NUMBER TO A
        ADD     A,'0'           ; MAKE IT ASCII AND SHOW IT
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
        LD      A,'-'           ; DISPLAY a DASH
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
        LD      H,(IX+MNSAH)    ; HL = STRING ADDRESS
        LD      L,(IX+MNSAL)
        SYSTEM  STRDIS          ; UPI STRing DISplay SELECTION
        ;
        LD      A,$08
        ADD     A,D             ; TO NEXT LINE
        LD      D,A
        LD      E,$10
        INC     B               ; BUMP ENTRY #
        LD      H,(IX+MNNH)     ; HL = NEXT ENTRY ADDR
        LD      L,(IX+MNNL)
        PUSH    HL
        LD      A,H
        OR      L
        JR      NZ,MMENU1               ; (-$25) NO - JUMP BACK
   ; AT THIS POINT HL = 0, (SP) = 0
        ADD     HL,SP           ; HL = STACK POINTER
MMENU3: PUSH    BC
        LD      BC,$0101
        XYRELL  DE, $10, $4D    ; FEEDBACK ADDRESS
        SYSTEM  GETNUM          ; UPI GET NUMBA from user
        ;
        POP     BC
        LD      A,(HL)          ; HOW DOES SHE LOOK?
        AND     A               ; ZERO ENTERED?
        JR      Z,MMENU5                ; (+$03) JUMP IF SO
        CP      B               ; IN RANGE?
        JR      C,MMENU6                ; (+$06) JUMP IF SO
MMENU5: LD      A,'?'           ; DUD ENTRY - SHOW ?
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
        JR      MMENU3                  ; (-$17) GO BACK FOR NEXT TRY

MMENU6: POP     HL              ; THROW OUT ENTRY AREA
        POP     DE              ; RESTORE HEAD OF MENU LIST
        LD      B,A             ; NUMBER ENTERED TO B
MMENU7: EX      DE,HL           ; HL - ENTRY POINTER
        LD      E,(HL)          ; DE - NEXT
        INC     HL
        LD      D,(HL)
        DJNZ    MMENU7                  ; (-$06) COUNT DOWN TO ENTRY
        INC     HL
        LD      E,(HL)          ; STRING TO DE
        INC     HL
        LD      D,(HL)
        INC     HL
        LD      C,(HL)          ; GO TO ADDRESS TO BC
        INC     HL
        LD      B,(HL)
        POP     HL              ; HL = RETURN TO PLACE
        POP     AF              ; THROW OUT OLD PC
        PUSH    BC              ; PUT NEW PC ON STACK
        PUSH    HL              ; AND PUT BACK DUMMY RETURN
FINDL3: LD      (IY+CBE),E      ; PASS BACK TITLE ADDRESS
        LD      (IY+CBD),D
        RET                     ; AND GO BACK

  ; ****** GET PARAMETER routine
  ; Purpose:      INPUT OF PROGRAM OPTIONS
  ; Inputs:       A = NUMBER OF DIGITS
  ;               BC = PROMPT STRING ADDRESS
  ;               DE = FRAME TITLE ADDRESS
  ;               HL = PARAMETER ADDRESS
  ; Description:
  ;       THIS ROUTINE ASKS THE USER TO ENTER A NUMBER
  ; FIRST A MENU FRAME IS CREATED, USING THE STRING
  ; POINTED AT BY DE AS A TITLE.  THE STRING 'ENTER'
  ; IS DISPLAYED, FOLLOWED BY THE PROMPT STRING.
  ; GETNUM IS THEN CALLED TO INPUT THE NUMBER.  FEEDBACK
  ; IS PROVIDED IN DOUBLE SIZED CHARACTERS.
  ; NOTE: ** THIS ROUTINE USES TWO SYSTEM LEVELS AND THE ALt. registers
MGETP:
        PUSH    AF              ; SAVE NUMBER OF DIGITS
        PUSH    HL
        PUSH    BC
        CALL    MNCLR
        SYSSUK  STRDIS          ; UPI STRing DISplay "ENTER"
        DB      $08             ; ... X = 8
        DB      $20             ; ... Y = 32
        DB      00001001B       ; ... Options (color 2 on color 1)
        DW      ENTSTG          ; ... String Address
        ;
        POP     HL
        SYSTEM  STRDIS          ; UPI STRing DISplay WHAT TO ENTER
        ;
        POP     HL
        POP     AF
        LD      B,A
        SET     6,C             ; SET LARGE CHARS
        XYRELL DE, $30, $30     ; LOAD FEEDBACK ADDRESS
        SYSTEM  GETNUM          ; UPI GET NUMber from user
        ;
        SYSSUK  PAWS            ; UPI PAUSE to LET USER READ IT
        DB      $0F             ; ... Interrupts = 15
        ;
        RET

  ; *** SUBROUTINE TO CLEAR SCREEN FOR MENU AND THROWUP TITLE
MNCLR:
        PUSH    DE
        SYSSUK  FILL            ; UPI FILL to Clear TOP LINE
        DW      NORMEM          ; ... Memory Address of upper left
        DW      $0B * BYTEPL    ; ... Byte Count = 440
        DB      $00             ; ... Data = 0
        ;
        SYSSUK  FILL            ; UPI FILL the Rest with color 1
        DW      NORMEM + ($0B*BYTEPL) ; ... Memory Address of line 12
        DW      (NOLINE-$0B) * BYTEPL ; ... Byte Count = 3400
        DB      $55             ; ... Data = 85
        ;
        POP     HL
        XYRELL  DE, $18, $00    ; TITLE
        LD      C,00000100B     ; for color 1 on color 0
        SYSTEM  STRDIS          ; UPI STRing DISplay
        ;
        RET

  ; ****** GET NUMBER routine
  ; Inputs:       B = DISPLAY OPTIONS
  ;               C = CHRDIS OPTIONS FOR FEEDBACK
  ;               DE = COORDINATE OF FEEDBACK AREA
  ;               HL = ADDRESS OF WHERE TO STASH NUMBER
  ; Description:  THIS ROUTINE CAN INPUT A NUMBER FROM
  ;       EITHER THE KEYBOARD OR THE HAND CONTROL.  KEYBOARD
  ;       ENTRY PROCEEDS CONVENTIONALY.  GETNUM EXITS
  ;       WHEN THE EQUALS KEY IS PRESSED OR THE REQUIRED NUMber
  ;       OF DIGITS IS ENTERED.
  ;               PLAYER ONE HAND CONTROL MAY ALSO BE USED TO
  ;       ENTER A NUMBER.  TO USE THIS OPTION, PULL THE TRIGger
  ;       THEN ROTATE THE POT UNTIL THE NUMBER YOU WISH TO
  ;       ENTER IS SHOWN IN THE FEEDBACK AREA.  PULL THE TRIGger
  ;       AGAIN TO REGISTER THE ENTRY.  IF DURING THIS PROCcess
  ;       THE KEYBOARD IS USED - KEYBOARD INPUT WILL OVERRIde.
MGETN:
        EXX
        CALL    CLRNUM          ; CLEAR THE NUMBER
        LD      C,A             ; SET ZERO DIGITS IN - POT ENABle
MGETN1: LD      A,(IY+CBB)      ; ENTRY COMPLETE?
        XOR     C
        AND     $3F
        RET     Z               ; QUIT IF SO
        LD      HL,MGETN1
        PUSH    HL
        SYSTEM  RANGED          ; UPI RANGED random number WHILE WE WAIT
        ;
        SYSSUK  SENTRY          ; UPI SENse TRansition Y
        DW      NUMBAS          ; ... Keymask Address for NUMBERS only
        ;
        SYSSUK  DOIT            ; UPI DOIT table, branch to translation handler
        DW      GNUMD0          ; ... Table Address
        ;
        RET                     ; NOTHIN - LOOP ON SENTRY

  ; ********* NEW Menu DO IT table
GNUMD0:
        JMPd    SKYD, MGETN6, $00
        JMPd    ST0, MGETN2, $00
        JMPd    SP0, MGETN3, $00

  ; ****** TRIGGER ROUTINE [also marks End of Table]
MGETN2:
        BIT     4,B             ; 0-1 TRANS?
        RET     Z               ; NO -IGNORE
        LD      A,C
        INC     A               ; ARE WE ALREADY IN POT MODE?
        JR      Z,MGETN9                ; (+$3a) YEP - JUMP TO EXIT
        BIT     7,C             ; POT LEGAL
        RET     NZ              ; NO - IGNORE
        LD      C,$FF           ; SET POT FLAG
  ; POT ROUTINE entry point
MGETN3: LD      A,C             ; QUIT IF NOT IN POT MODE
        INC     A
        RET     NZ
  ; HOW MANY DIGITS?
        EXX                     ; TO NORMAL SET
        LD      A,B             ; SNATCH DIGITS
        EXX
        CP      $01             ; 1 PRAY TELL?
        LD      B,$0A
        JR      Z,MGETN4                ; (+$02) JUMP IF GOOD GUESS
        LD      B,$64           ; 100 if WRONG!
MGETN4: IN      A,(POT0)        ; GET CURRENT POT VALUE
        LD      D,A             ; RANGE IT
        XOR     A
        LD      E,A
        LD      H,A
MGETN5: ADD     HL,DE
        ADC     A,$00           ; ADD EVERY CARRY TO AC
        DAA
        DJNZ    MGETN5                  ; (-$06)
        EXX                     ; BACK TO NORMAL SET
        LD      (HL),A
        JR      MGETN8                  ; (+$14)

  ; *** KEYBOARD subROUTINE
MGETN6: INC     C               ; POT MODE?
        JR      NZ,MGETN7               ; (+$04) JUMP IF NOT
        CALL    CLRNUM
        INC     C               ; SET ONE DIGIT SO FAR
MGETN7: SET     7,C             ; SET POT LOCKOUT
        SYSTEM  KCTASC          ; UPI Key Code in B To ASCii
        ;
        CP      '='             ; EQUALS TYPED?
        JR      Z,MGETN9                ; (+$08) QUIT IF EQUALS
        AND     $0F
        EXX
        SYSTEM  SHIFTU          ; UPI SHIFT digit Up in A
        ;
MGETN8: PUSH    DE
        SYSTEM  DISNUM          ; UPI DISplay NUMber
        ;
  ; ENTER HERE FOR EQUAL OR TRIGGER EXIT TO THROW OUT RETURN
MGETN9: POP     DE
        EXX                     ; BACK TO NORMAL
        RET

  ; *** SUBROUTINE TO CLEAR NUMBER
CLRNUM:
        PUSH    BC
        EXX                     ; TO NORMAL SET
        PUSH    HL
        LD      A,B
        INC     A
        AND     $3E
if WHITE or ASTRO ; ###### cond #######
        RRA                     ; # 1 # added code to ver. 1 ??
endif ; ##### end conditional #########
        EXX                     ; BACK TO NORMAL SET
        LD      C,A
        XOR     A
        LD      B,A
        POP     DE
        SYSTEM  FILL            ; UPI FILL memory with data
        ;
        POP     BC
        RET

  ; ****** SHIFT UP routine
  ; Inputs:       A = DATA TO SHIFT UP
  ;               B = SIZE IN DIGITS
  ;               HL = AREA TO SHIFT ADDRESS
MSHFTU:
        PUSH    AF
        LD      A,B
        INC     A
        AND     $3E
        LD      B,A
        POP     AF
SHFTU1: RLD
        INC     HL
        DJNZ    SHFTU1                  ; (-$05)
        RET

  ; ********* "ENTer " text string
ENTSTG:
        DB      'E','N','T','E','R',$20
        DB      $00

  ; ********* "CheckMate" menu Link
CML     DW      CALCL           ; Next node
        DW      PNCM            ; PriNt "CheckMate"
        DW      CMSTRT          ; CheckMate STaRT

  ; ********* "SCriBbling" menu Link
SCBL    DW      $0000           ; LAST node
        DW      PNSCB           ; PriNt "SCriBbling"
        DW      SCBST           ; SCriBbling STart

  ; ********* PriNt "GunFight" text string
PNGF:
        DB      'G','U','N','F','I','G','H','T'
        DB      $00

  ; ********* PriNt "CheckMate" text string
PNCM:
        DB      'C','H','E','C','K','M','A','T','E'
        DB      $00

  ; ********* PriNt "CALCulator" text string
PNCALC:
        DB      'C','A','L','C','U','L','A','T','O','R'
        DB      $00

  ; ********* PriNt "SCriBbling" text string
PNSCB:
        DB      'S','C','R','I','B','B','L','I','N','G'
        DB      $00

  ; ********* "select GAMe" and '(C)opyright' text strings
GAMSTR:
        DB      'S','E','L','E','C','T'
        DB      $20,'G','A','M','E'
        DB      01100111B       ; mask (for E, D, and C)
        DB      $08             ; ... X = 8
        DB      $58             ; ... Y = 88 BOTTOM LINE
        DB      00001101B       ; ... for color 3 on color 1
        DB      '(','C',')',$20,'B','A','L','L','Y'
        DB      $20,'M','F','G',$20,'1','9','7'
if BALLY or WHITE ; ### conditional ###
        DB      '7'             ; # - # (C) 1977
else ; ASTRO ####### conditional ###### else...
        DB      '8'             ; # + # (C) 1978
endif ; ##### end conditional #########
        DB      $00

; FOUR FILLER BYTES BETWEEN ROM AND SCRIBBLING
; TO MATCH 3159 ROMs [note: NONE in BALLY 3164 ROMs]
if WHITE ; ### conditional assembly ###
        DB   '9','7','7',$00    ; # # # leftovers
endif ; ##### end conditional ######### or...
if ASTRO ; ### conditional assembly ###
        DB   $FF,$FF,$FF,$FF    ; #-1 # erased
endif ; ##### end conditional #########

  ;
  ; **********
  ; * SKETCH * Ver. 1.13 - Nov. 30, 2003
  ; **********
  ;

  ; The official name of this
  ; program is Scribbling
  ;
  ; zmac -i -m -o scribble.bin -x scribble.lst scribble.asm

  ; SKETCH EQUATES
  ; SKETCH PACKET DISPLACEMENTS
SCPSIZ  EQU     $1E             ; SIZE OF SKETCH PACKET = 30 bytes
SCSAVA  EQU     $00             ; SAVE AREA START
SCXC    EQU     $1A             ; X COORDINATE +26
SCYC    EQU     $1B             ; Y COORDINATE +27
SCSADL  EQU     $1C             ; SAVE ADDRESS LO +28, AND HI +29
SCSADH  EQU     $1D
  ; Other goodies
MOVTMR  EQU     $04             ; MOVE RATE
KSCTRV  EQU     $14             ; COLOR STEPPING TIME = 20

  ;  SWITCH RAM DEFINES
; SSKBOT  EQU     NORMEM + $5D*BYTESPL ; STacK BOTtom limit line 93
SCRSTK  EQU     $4EE8           ; (96 bytes) SCRibbling STacK top
COLORS  EQU     SCRSTK          ; (8 bytes) COLORS table in RAM

P1SCP   EQU     $4EF0           ; (30 bytes) Player 1 SketCh Packet
ZPSTMR  EQU     P1SCP + $00     ; (word) Zero Player Size TiMeR
DIRTMR  EQU     P1SCP + $02     ; (byte) DIRTMR  EQU  ZPSTMR+2
ZPSIZ   EQU     P1SCP + $03     ; (word) ZPSIZ   EQU  DIRTMR+1
DIRVAL  EQU     P1SCP + $05     ; (word) DIRection VALue
P2SCP   EQU     $4F0E           ; (30 bytes) Player 2 SketCh Packet
P3SCP   EQU     $4F2C           ; (30 bytes) Player 3 SketCh Packet
P4SCP   EQU     $4F4A           ; (30 bytes) Player 4 SketCh Packet

  ; ====== SCBST: SCriBbling STart equate
; L0E19:
  ; ****** Sketch BEGIN         ; ** START
; BEGIN:
        SYSSUK  GETPAR          ; UPI GET game PARameter from user
        DW      NOPLAY          ; ... Prompt Address "# of PLAYers"
        DB      $01             ; ... Digits = 1 (large)
        DW      NUMPLY          ; ... Parameter Address in RAM
        ;
SCCLR:  LD      SP,SCRSTK
        SYSTEM  INTPC           ; UPI INTerPret with Context create
        ;
        DO      FILL            ; UPI FILL to clear 92 lines of memory
        DW      NORMEM          ; ... RAM Memory Address
        DW      $5C * BYTEPL    ; ... Byte Count = 3680
        DB      $00             ; ... Data = 0
        ;
        DO      FILL            ; UPI FILL to clear four Sketch Packets
        DW      P1SCP           ; ... Memory Address
        DW      SCPSIZ * 4      ; ... Byte Count = 120
        DB      $00             ; ... Data = 0
        ;
        DO      SETOUT          ; UPI SET some OUTput ports
        DB      $B8             ; ... VERBL*2 = 184
        DB      $28             ; ... HORCB/4 = 40
        DB      $08             ; ... INMOD = 8 (screen only)
        ;
        DO      MOVE            ; UPI MOVE block transfer
        DW      COLORS          ; ... Destination Address in RAM
        DW      $0008           ; ... Byte Count = 8
        DW      INICOL          ; ... Source Address (INItial COLors)
        ;
        DO      COLSET          ; UPI COLors SET
        DW      COLORS          ; ... Table Address in RAM
        ;
        DO      SETW            ; UPI SET Word of Sketch packet #1
        DW      $2446           ; ... Data Word (X = 70, Y = 36)
        DW      P1SCP + SCXC    ; ... Memory Address
        ;
        DO      SETW            ; UPI SET Word of Sketch packet #2
        DW      $2453           ; ... Data Word (X = 83, Y = 36)
        DW      P2SCP + SCXC    ; ... Memory Address
        ;
        DO      SETW            ; UPI SET Word of Sketch packet #3
        DW      $3046           ; ... Data Word (X = 70, Y = 48)
        DW      P3SCP + SCXC    ; ... Memory Address
        ;
        DO      SETW            ; UPI SET Word of Sketch packet #4
        DW      $3053           ; ... Data Word (X = 83, Y = 48)
        DW      P4SCP + SCXC    ; ... Memory Address
        ;
        DO      SETB            ; UPI SET Byte
        DB      MOVTMR          ; ... Data = 4
        DW      CT0             ; ... Memory Address (Counter/Timer 0)
        ;
        DONT    XINTC           ; UPI eXit INTerpreter with Context
        ;
MAINLP: LD      HL,MAINLP       ; MAIN LooP here
        PUSH    HL
        SYSSUK  SENTRY          ; UPI SENse TRansition Y
        DW      KEYMES          ; ... KEYMask Address for Scribbling
        ;
        SYSSUK  DOIT            ; UPI DOIT table, branch to translation handler
        DW      SCDOTB          ; ... SCribble DOit TaBle Address
        ;
        RET                     ; urn to MAINLP

KEYMES:             ; *** KEYMaSk for top 4 rows plus CE button
        DB      $2F             ; KEYpad column 3 (left side)    x x E x 4 7 R C  T
        DB      $0F             ; KEYpad column 2 (center left)  x x x x 5 8 S >  O
        DB      $0F             ; KEYpad column 1 (center right) x x x x 6 9 H <  P
        DB      $0F             ; KEYpad column 0 (right side)   x x x x - * / %

  ; ****** KEYB0ARD HANDLER
KEYB0:
        DEC     B
        LD      C,$03
        LD      A,B
        CP      $14             ; CLEAR ENTRY DOWN?
        JR      Z,SCCLR                 ; (-$51) JUMP TO CLEAR IF SO
        RRCA
        RRCA
        AND     C
        SYSSUK  INDEXB          ; UPI INDEX Byte by A
        DW      CDELTB          ; ... Base Address
        ;
        EX      DE,HL
        LD      A,B
        AND     C
        LD      H,A
        LD      A,C
        SUB     H
        SYSSUK  INDEXB          ; UPI INDEX Byte by A
        DW      COLORS          ; ... Base Address, POINT AT COLOR
        ;
        LD      A,(DE)
        ADD     A,(HL)          ; ADD DELTA FACTOR
        BIT     3,B             ; WAS KEY FOR INTENSITY?
        JR      Z,KEYB1                 ; (+$04)
        XOR     (HL)
        AND     $07
        XOR     (HL)
KEYB1:  LD      (HL),A
        INC     HL              ; CHANGE COLOR ON OTHER SIDE
        INC     HL
        INC     HL
        INC     HL
        LD      (HL),A
        SYSSUK  COLSET          ; UPI COLors SET
        DW      COLORS          ; ... Table Address in RAM
        ;
        LD      A,KSCTRV        ; SET KEYSEX CLEAR TIMER
        LD      (CT1),A
        RET

  ; ****** subROUTINE TO CLEAR KEYSEX
KLRKSX:
        XOR     A
        LD      (KEYSEX),A
        RET

  ; ********* SCribble DOit TaBle
SCDOTB:
        JMPd    SCT0, DOWRTS, $00
        JMPd    SCT1, KLRKSX, $00
        JMPd    SKYD, KEYB0, $00

  ; *** ITERATE THROUGH ACTIVE PLAYERS SUBROUTINE
ITER4:
        LD      IX,P1SCP        ; [also marks End of Table]
        LD      A,(NUMPLY)
        LD      B,A
        LD      C,A
ITER41: PUSH    BC
        PUSH    HL
        LD      DE,ITRET
        PUSH    DE
        JP      (HL)

  ; *** fake ITeration RETurn address
ITRET:
        LD      DE,SCPSIZ
        ADD     IX,DE
        POP     HL
        POP     BC
        DJNZ    ITER41                  ; (-$10)
        RET

  ; *** UPDATE COORDINATES ROUTINE
SCRUPD:
        LD      A,C
        SUB     B
        SYSSUK  INDEXB          ; UPI INDEX Byte by A
        DW      OSW0            ; ... Base Address (Old SWitch track)
        ;
        AND     $0F
        CALL    GETDLT          ; GET DELTAS
        LD      A,(IX+SCXC)     ; UPDATE X
        ADD     A,D
        CP      $98             ; OUT OF BOUNDS?
        JR      NC,SCRUP1               ; (+$03)
        LD      (IX+SCXC),A
SCRUP1: LD      A,(IX+SCYC)     ; SAME FOR Y
        ADD     A,H
        CP      $55
        RET     NC
        LD      (IX+SCYC),A
        RET

  ; *** RESTORE subroutine
SCREST:
        PUSH    IX
        POP     DE
        LD      A,(DE)
        AND     A
        RET     Z
        LD      H,(IX+SCSADH)
        LD      L,(IX+SCSADL)
        SYSTEM  RESTOR          ; UPI RESTORe area
        ;
        XOR     A
        LD      (DE),A
        RET

  ; *** WRITE ROUTINE
SCRWRT:
        LD      A,C
        SUB     B
        SYSSUK  INDEXB          ; UPI INDEX Byte by A
        DW      OSW0            ; ... Base Address (Old SWitch track)
        ;
        AND     $10
        RET     Z
SCRWR1: DEC     HL              ; BACKUP TO POT
        DEC     HL
        DEC     HL
        DEC     HL
        DEC     HL
        LD      A,(HL)
        RLCA
        RLCA
        LD      C,A
        AND     $03
        SYSSUK  INDEXB          ; UPI INDEX Byte by A to SET SIZES
        DW      SIZTBL          ; ... Base Address (SIZe TaBLe)
        ;
        LD      D,(IX+SCYC)
        LD      E,(IX+SCXC)
        LD      B,A
        LD      A,C
        RLCA
        RLCA
        AND     $03
SCRWR2: SYSSUK  INDEXB          ; UPI INDEX Byte by A
        DW      COLMSK          ; ... Base Address (COLor MaSK)
        ;
        LD      C,B
        SYSTEM  RECTAN          ; UPI paint a RECTANgle
        ;
        RET

  ; ****** COLor 0, 1, 2, and 3
COLMSK:
        DB      $00
        DB      01010101B
        DB      10101010B
        DB      11111111B

  ; ****** SIZe 1x, 2x, and 4x
SIZTBL:
        DB      $01
        DB      $02
        DB      $04

  ; ****** Color DELta for row 1 (top), 2, 3, and 4
CDELTB:
        DB      +8
        DB      -8
        DB      +1
        DB      -1

  ; *** SAVE subROUTINE
SCRSAV:
        LD      A,B
        SYSSUK  INDEXB          ; UPI INDEX Byte by A
        DW      OSW0-1          ; ... Base Address (Old SWitch track - 1)
        ;
        AND     $10
        RET     NZ
        PUSH    HL
        LD      D,(IX+SCYC)
        LD      E,(IX+SCXC)
        SYSTEM  RELAB1          ; UPI RELative to non-magic ABSolute
        ;
        LD      (IX+SCSADH),D
        LD      (IX+SCSADL),E
        EX      DE,HL
        PUSH    IX
        POP     DE
        LD      BC,$0803        ; SAVE WORST CASE
        SYSTEM  SAVE            ; UPI SAVE area
        ;
        POP     HL
        JR      SCRWR1                  ; (-$51)

  ; *** ZERO PLAYER GAME WRITE HANDLER
ZEROPL:
        LD      HL,ZPSTMR       ; LOAD PTR TO SIZE TIMER
        LD      DE,ZPSIZ        ; AND SIZE TRACKER
        DEC     (HL)            ; DECREMENT SIZE TIMER
        JP      P,ZPA           ; JUMP IF NO COUNTDOWN
        SYSSUK  RANGED          ; UPI RANGED random number
        DB      $30             ; ... Cutoff = 48
        ;
        CP      $08             ; 8-479 ???
        JR      C,ZP0                   ; (+$02) NO - ZP0
        AND     $03             ; YES - HAVE MORE 1-4S
ZP0:    INC     A
        LD      (DE),A          ; GET NEW SIZE
  ; GET NEW SIZE TIMER
        SYSSUK  RANGED          ; UPI RANGED random number
        DB      $78             ; ... Cutoff = 120
        ;
        LD      (HL),A
ZPA:    INC     HL              ; ADVANCE  TO COLOR STUFF
        INC     DE
        DEC     (HL)            ; AND DEC COLOR TIMER
        JP      P,ZPB
  ; GET NEW COLOR
        SYSSUK  RANGED          ; UPI RANGED random number
        DB      $04             ; ... Cutoff = 4
        ;
        LD      (DE),A
  ; GET NEW COLOR TIMER
        SYSSUK  RANGED          ; UPI RANGED random number
        DB      $78             ; ... Cutoff = 120
        ;
        LD      (HL),A
ZPB:    INC     HL              ; TO DIRECTION STUFF
        INC     DE
        DEC     (HL)            ; DECREMENT DIRECTION TIMER
        JP      P,ZPD
ZPC:    LD      DE,DIRVAL       ; DE = DIRECTION TRACKER
  ; DRAW NEW DIRECTION
        SYSSUK  RANGED          ; UPI RANGED random number
        DB      $0A             ; ... Cutoff = 10
        ;
        INC     A
        CP      $03             ; R ILLEGAL MOVES
        JR      Z,ZPC                   ; (-$0b)
        CP      $07
        JR      Z,ZPC                   ; (-$0f)
        LD      (DE),A
        SYSSUK  RANGED          ; UPI RANGED random number
        DB      $28             ; ... Cutoff = 40
        ;
        LD      (DIRTMR),A
ZPD:    LD      A,(DE)          ; GET DIRECTION VALUE
        CALL    GETDLT          ; GET DELTAS
        LD      BC,P1SCP + SCXC ; POINT AT COORDINATE
        LD      A,(BC)
        ADD     A,D
        CP      $50
        JR      NC,ZPC                  ; (-$23) GET NEW DIRECTION IF AT LMT
        LD      (BC),A
        LD      E,A             ; SAVE X COORDINATE
        INC     BC
        LD      A,(BC)
        ADD     A,H
        CP      $2E
        JR      NC,ZPC                  ; (-$2c)
        LD      (BC),A
        LD      D,A             ; SET Y COORDINATE
        LD      HL,ZPSIZ        ; POINT AT SIZES AGAIN
        LD      B,(HL)
        INC     HL
        LD      A,(HL)          ; GET COLOR TOO
        CALL    SCRWR2          ; DO FIRST WRITE
        LD      H,A             ; SAVE COLOR
        PUSH    DE              ; AND X,Y
        LD      A,$5C           ; REFLECT Y
        SUB     B
        SUB     D
        LD      D,A
        LD      A,H
        SYSTEM  RECTAN          ; UPI paint a RECTANgle
        ;
        LD      A,$A0           ; REFLECT X
        SUB     C
        SUB     E
        LD      E,A
        LD      A,H
        SYSTEM  RECTAN          ; UPI paint a RECTANgle
        ;
        POP     HL              ; RESTART X,Y
        LD      D,H             ; RESTORE Y
        SYSTEM  RECTAN          ; UPI paint a RECTANgle
        ;
        LD      A,$FF           ; RESET TIMEOUT
        LD      (TIMOUT),A
ZERO1:  LD      A,$01           ; RESET COUNTER-TIMER
        JR      ZERO2                   ; (+$2a)

  ; ****** on timer
DOWRTS:
        LD      A,(NUMPLY)
        DEC     A
        CP      $04
        JP      NC,ZEROPL
        LD      HL,SCRUPD
        CALL    ITER4
        LD      HL,SCREST
        CALL    ITER4
        LD      HL,SCRWRT
        CALL    ITER4
  ; NOW GOING BACKWARDS SAVE AND WRITE EVERYBODY WITH TRIGGer
        LD      B,C
SCRB3:  LD      DE,-SCPSIZ
        ADD     IX,DE
        PUSH    BC
        CALL    SCRSAV
        POP     BC
        DJNZ    SCRB3                   ; (-$0c)
        LD      A,MOVTMR
ZERO2:  LD      (CT0),A
        RET                     ; DONE

  ; *** SUBROUTINE TO SCARE UP DELTAS
GETDLT:
        PUSH    BC
        LD      B,A
        SYSSUK  MSKTD           ; UPI MaSK joystick in B To Deltas
        DW      $0100           ; ... X Delta = 256 (for 1.0 pixel)
        DB      $00             ; ... Flop = 0
        DW      $0100           ; ... Y Delta = 256 (for 1.0 pix)
        ;
        POP     BC
        RET

  ; ****** INITIAL COLORS copied to COLORS in RAM
INICOL:
if BALLY or WHITE ; ### conditional ###
        DB      $08             ; # = # dark BLUE color 3 left
        DB      $5B             ; # + # RED color 2
        DB      $A5             ; # = # GREEN color 1
        DB      $07             ; # = # WHITE color 0
        DB      $08             ; # = # (repeated for right side)
        DB      $5B             ; # + #
        DB      $A5             ; # = #
        DB      $07             ; # = #
else ; ASTRO ####### conditional ###### else...
        DB      $08             ; # = # dark BLUE
        DB      $62             ; # - # darker RED
        DB      $A5             ; # = # GREEN
        DB      $07             ; # = # WHITE
        DB      $08             ; # = # (repeated)
        DB      $62             ; # - #
        DB      $A5             ; # = #
        DB      $07             ; # = #
endif ; ##### end conditional #########

  ;
  ;       END OF SCRIBBLING PROGRAM
  ;

; TWELVE FILLER BYTES BETWEEN SCRIBBLING AND CALCULATOR TO MATCH ROM
        DB      $FF,$FF,$FF,$FF,$FF
        DB      $FF,$FF,$FF,$FF,$FF
        DB      $FF,$FF

  ;
  ; **************
  ; * CALCULATOR * (Dec 4, 2003)
  ; **************
  ;

  ; CALCULATOR EQUATES
  ; SCREEN PARAMETERS
CALCLN  EQU     $5C             ; LAST DISPLAY LINE USED = 92
  ; SCREEN MEMORY STACK PARAMETERS
MAXSTK  EQU     CALCLN          ; MAX NUMBER OF STACK ENTRYS
  ; CALCULATOR MAJOR STATES
ARG1MD  EQU     $00             ; ARG1 STATE if 0
ARG2MD  EQU     $01             ; ARG2 STATE if 1
SHOWMD  EQU     $02             ; SHOW STATE if 2

  ; ******************
  ; * CALCULATOR RAM *
  ; ******************
STKBOT  EQU     MAXSTK*BYTEPL + NORMEM - $08 ; ADDR OF STK BOTTOM
CALSTK  EQU     $4EC0           ; (96 bytes) CALculator STacK top
CALRAM  EQU     $4EC0
OPCOD   EQU     $4EC0           ; OPERATION TO DO
STATEV  EQU     $4EC1           ; STATE VARIABLE
MEMSTA  EQU     $4EC2           ; MEMORY STATE VARIABLE
MEMDIG  EQU     $4EC3           ; MEMORY DIGIT VARIABLE
ARG1    EQU     $4EC4           ; ARGUMNET 1 - 11 BYTES
  ; *** NOTE *** CODE EXPECTS ARG2 THRU SHOWBF TO BE IN ONE
ARG2    EQU     $4ECF           ; ARGUMENT 2 - 11 BYTES
DIGITS  EQU     $4EDA           ; DIGITS COUNTER
DPOINT  EQU     $4EDB           ; DECIMAL POINT ENTERED FLAG
POINTP  EQU     $4EDC           ; DECIMAL POINT POINTER
SHOWBF  EQU     $4EDD           ; ENTRY FEEDBACK BUFFER - 7 BYTES
NLINES  EQU     $4EE4           ; NUMBER OF LINES ON SCROLL STA
SCRPTR  EQU     $4EE5           ; CURRENT POS OF SCROLL SYSTEM
CMEM0   EQU     $4EE6           ; CALCULATOR MEMORYS - 110 BYTES
ENDCR   EQU     $4F54

  ; ====== CALCST: CALCulator STart equate
; L1020:
  ; ****** calculator CLEAR
CLEAR:
        LD      SP,CALSTK       ; BUSHWACK THE STACK
        SYSTEM  INTPC           ; UPI INTerPret with Context create
        ;
        DO      FILL            ; UPI FILL memory with data
        DW      $4000           ; ... Memory Address (NORMEM)
        DW      CALCLN * BYTEPL ; ... Byte Count = 3680
        DB      $00             ; ... Data = 0
        ;
        DO      RECTAN          ; UPI paint a RECTANgle in color 3
        DB      $08             ; ... X = 8
        DB      $52             ; ... Y = 82
        DB      $70             ; ... Width = 112
        DB      $09             ; ... Height = 9
        DB      $FF             ; ... Color = 11111111B
        ;
        DO      SETOUT          ; UPI SET some OUTput ports
        DB      CALCLN * 2      ; ... VERBL*2 = 184
        DB      $1F             ; ... HORCB/4 = 31
        DB      $08             ; ... INMOD = 8 (screen only)
        ;
        DO      COLSET          ; UPI COLors SET
        DW      CALCLR          ; ... CALCULATOR COLORS
        ;
        DO      FILL            ; UPI FILL memory with data
        DW      CALRAM          ; ... Memory Address (CALculator RAM)
        DW      ENDCR - CALRAM  ; ... Byte Count = 148
        DB      $00             ; ... Data = 0
        ;
        DO      ACTINT          ; UPI ACTivate sub timer INTerrupts
        ;
        DO      XINTC           ; UPI eXit INTerpreter with Context
        ;
        CALL    DOCLRE          ; SHOW A ZERO WHEN WE COME UP
  ; GET NEXT KEY DEPRESSION
KEYGET:
        LD      HL,KEYGET       ; fake a return address back here
        PUSH    HL
        SYSSUK  SENTRY          ; UPI SENse TRansition Y
        DW      ALKEYS          ; ... Keymask Address (ALL KEYS)
        ;
        CP      SKYD            ; KEY DOWN?
        RET     NZ              ; NO - WAIT
        LD      A,B
        SYSSUK  INDEXB          ; UPI INDEX Byte by A
        DW      TOKETB - 1      ; ... Base Address (??)
        ;
        AND     $0F             ; ISOLATE 'TRANSLATE TO' FIELD
        LD      C,A
        XOR     (HL)            ; BRANCH ON TOKEN TYPE
        RRCA
        RRCA
        RRCA
        RRCA
        LD      E,A
  ; ARE WE WAITING FOR USER TO TYPE IN A MEMORY NUMBER?
        LD      HL,MEMSTA       ; POINT AT MEMORY STATE
        LD      A,(HL)
        AND     A
        JR      Z,KEYGT1                ; (+$32) JUMP IF NOT IN THAT STATE
        LD      A,E             ; WAS   DIGIT TYPED?
        AND     A
        RET     NZ              ; NO - GO TO NEXT ENTRY
        LD      D,A             ; ZERO OUT D
        LD      A,C
        CP      $0E             ; DID OPERATOR HIT A DECIMAL PO
        RET     Z               ; YES - QUIT
  ; SET MEMORY DIGIT FOR DISPLAY
        INC     HL
        LD      (HL),C
        INC     (HL)            ; ZERO DESIGNATES NOTHIN
        LD      E,A
        DEC     HL
        LD      A,(HL)          ; S-R STATUS TO A AGAIN
        RRCA                    ; TEST STATUS
        JR      C,GMEM1                 ; (+$07) JUMP IF RECALL
  ; STORE PROCESSING
        CALL    XMEM            ; DO POINTING AND SETUP
        LDIR                    ; PERFORM STORE
        JR      SHONUM                  ; (+$09)

  ; *** RECALL PROCESSING
GMEM1:  CALL    SHOWCK          ; CHECK FOR SHOW MODE
        CALL    XMEM
        EX      DE,HL
        LDIR                    ; COPY MEM TO ARG2
SHONUM: LD      HL,$0000
        LD      (DIGITS),HL     ; CLEAR DIGIT FLAGS
        LD      HL,SHOWBF
        PUSH    HL
        CALL    CONVRT
        POP     HL
        JP      ADIG7           ; JOIN DISPLAY SEQUENCE

  ; *** ??
KEYGT1: LD      A,E
        SYSSUK  INDEXW          ; UPI INDEX Word by A
        DW      TOKEJT          ; ... Base Address (TOKEn Jump Table)
        ;
        PUSH    DE
        LD      HL,STATEV
        LD      A,(HL)
        RET                     ; CALL VIA RETURN

  ; ****** TOKEN TYPE-TRANSLATE TABLE
  ; HO NIBBLE IS TYPE, LO NIBBLE IS WHAT TO TRANSLATE TO
TOKETB:
        DB      $80             ;  1  C
        DB      $60             ;  2  VA
        DB      $61             ;  3  DA
        DB      $31             ;  4  %
        DB      $51             ;  5  MR
        DB      $52             ;  6  MS
        DB      $40             ;  7  +-
        DB      $14             ;  8  DIVIDE
        DB      $07             ;  9  7
        DB      $08             ; 10  8
        DB      $09             ; 11  9
        DB      $13             ; 12  X
        DB      $04             ; 13  4
        DB      $05             ; 14  5
        DB      $06             ; 15  6
        DB      $12             ; 16  -
        DB      $01             ; 17  1
        DB      $02             ; 18  2
        DB      $03             ; 19  3
        DB      $11             ; 20  +
        DB      $70             ; 21  CE
        DB      $00             ; 22  0
        DB      $0E             ; 23  .
        DB      $20             ; 24  =

  ; ****** TOKEN TYPE JUMP TABLE
TOKEJT:
        DW      ACTNUM          ; 0  NUMBER
        DW      OPROUT          ; 1  OPERATOR
        DW      EQROUT          ; 2  EQUAL SIGN
        DW      PERCEN          ; 3  PERCENT
        DW      CHGSGN          ; 4  CHANGE SIGN
        DW      MEMFUN          ; 5  MEMORY
        DW      SCRLKY          ; 6  SCROLL KEYS
        DW      CLRENT          ; 7  CLEAR ENTRY
        DW      CLEAR           ; 8  CLEAR CALC

  ; ****** SUBROUTINE TO POINT DE AT A MEMORY AND LOAD HL
  ; INPUT   D = 0, E = DIGIT (0 TO 9 MEMORY #)
  ; OUTPUT  DE = POINTER TO MEM, HL = ARG2, BC = 12
XMEM:
        LD      HL,CMEM0
        LD      B,$0B
        LD      C,B
XMEM1:  ADD     HL,DE
        DJNZ    XMEM1                   ; (-$03)
        EX      DE,HL
        LD      HL,ARG2
        RET

  ; ****** MEMORY KEY ROUTINE
MEMFUN:
        LD      HL,MEMSTA       ; SET MEMORY STATE
        LD      (HL),C          ; SET MEMORY STATE
        JR      EQRT1                   ; (+$42)

  ; ****** CHANGE SIGN ROUTINE
CHGSGN:
        CALL    SHOWCK
        SYSSUK  BCDCHS          ; UPI BCD CHange Sign
        DB      $0B             ; ... Digits/2 = 11
        DW      ARG2            ; ... Number Address
        ;
        JR      EQRT1                   ; (+$38)

  ; ****** OPERATOR ROUTINE
OPROUT:
        CP      SHOWMD          ; ARE WE IN SHOW STATE?
        JR      NZ,OP1                  ; (+$04) JUMP IF NOT
        LD      (HL),ARG2MD     ; GO TO ARG2 STATE
        JR      OP3                     ; (+$1b)

OP1:    CP      ARG1MD          ; ARG1 MODE?
        EX      AF,AF'
        LD      A,C
        EX      AF,AF'
        JR      NZ,OP2                  ; (+$0f) JUMP IF NOT
        LD      (HL),ARG2MD     ; YUP, GO TO ARG2 STATE
        CALL    PUSHEN          ; PUSH ARG1 DOWN
        SYSSUK  MOVE            ; UPI MOVE block transfer
        DW      ARG1            ; ... Destination Address
        DW      $000B           ; ... Byte Count = 11
        DW      ARG2            ; ... Source Address
        ;
        JR      OP2A                    ; (+$03)

OP2:    CALL    DOOP            ; DO IT
OP2A:   EX      AF,AF'
        LD      C,A
OP3:    LD      A,C
        LD      (OPCOD),A
        JR      EQRT1                   ; (+$0f) JOIN EQUAL DISPLAY CALL

  ; ****** EQUALS ROUTINE
EQROUT:
        DEC     A               ; ARE WE IN ARG2MD?
        RET     NZ
        LD      (HL),SHOWMD     ; ENTER SHOW STATE
        CALL    DOOP
        LD      A,$05
        LD      (OPCOD),A
        CALL    PUSHEN
EQRT1:  JP      SHONUM

  ; ****** PERCENt routine
PERCEN:
        XOR     C               ; ARG2 MODE?
        RET     NZ
        SYSSUK  MOVE            ; UPI MOVE block transfer
        DW      ARG2            ; ... Destination Address
        DW      $0009           ; ... Byte Count = 9
        DW      ARG2 + 1        ; ... Source Address
        ;
        LD      (ARG2+9),A
        LD      HL,ARG1
        LD      B,$0B
        LD      A,(ARG1+10)     ; SAVE SIGN OF ARG1
        SYSTEM  BCDMUL          ; UPI BCD MULtiplication
        ;
        LD      (ARG1+10),A     ; RESTORE SIGN OF ARG1
        EX      DE,HL
        LD      BC,$000B
        CALL    OVRCHK          ; CHECK FOR OVERFLOW
        JR      EQRT1                   ; (-$26)

  ; *** SUBROUTINE TO DO OPERATION
  ; NOTE THIS ROUTINE SETS BOTH ARG1 AND ARG2 EQUAL TO RESUlt
DOOP:
        CALL    PUSHEN
        LD      A,(OPCOD)
        SYSSUK  INDEXW          ; UPI INDEX Word by A
        DW      DOOPTB - 2      ; ... Base Address
        ;
        PUSH    DE
        LD      HL,ARG2
        LD      DE,ARG1
        LD      B,$0B
        RET                     ; JUMP USING RETURN

  ; ****** DO OPeration TaBle
DOOPTB:
        DW      DOADD
        DW      DOSUB
        DW      DOMUL
        DW      DODIV

DOADD:  SYSTEM  BCDADD          ; UPI BCD ADDition
        ;
        DB   $01

DOSUB:  SYSTEM BCDSUB           ; UPI BCD SUBtraction
        ;
        DB   $01

DOMUL:  SYSTEM BCDMUL           ; UPI BCD MULtiplication
        ;
        DB   $01

DODIV:  SYSTEM BCDDIV           ; UPI BCD DIVision
        ;
  ; *** set ARG2 = ARG1
; NUMCHK:
        SYSSUK  MOVE            ; UPI MOVE block transfer
        DW      ARG2            ; ... Destination Address
        DW      $000B           ; ... Byte Count = 11
        DW      ARG1            ; ... Source Address
        ;
OVRCHK: ADD     HL,BC
        DEC     HL
        LD      A,(HL)
        AND     $0F             ; OVERFLOW SET?
        RET     Z               ; QUIT IF NOT
        LD      A,$01
        LD      (HL),A
        INC     A               ; FOR SHOW MODE
        LD      (STATEV),A
        CALL    PUSHEN
        POP     HL              ; THROW OUT RETURN ADDRESS
DOCLRE: LD      C,$00           ; FORCE CLEAR ENTRY
        JR      CLRENT                  ; (+$25)

  ; ****** SUBROUTINE TO ACCEPT A DIGIT
  ; ONE DIGIT
ACTNUM:
        CALL    SHOWCK
        LD      HL,DIGITS
        LD      DE,ARG2
        LD      A,(HL)          ; HOW MANY DIGITS SO FAR?
        CP      $0A
        RET     Z               ; QUIT IF LIMIT ATTAINED
        INC     HL              ; DECIMAL POINT ENTERED YET?
        LD      A,(HL)
        AND     A
        JR      Z,ADIG1                 ; (+$0f) JUMP IF NO
        CP      C               ; IS INPUT A DECIMAL POINT?
  ; (FLAG = 0E IF SET!!)
        RET     Z               ; QUIT IF EXTRA
        LD      A,C
        INC     HL
        LD      C,(HL)          ; C = DECIMAL POINT
        DEC     (HL)
        EX      DE,HL
        SYSTEM  STOREN          ; UPI STORE THE Nibble in A by C
        ;
        LD      C,A             ; GET DIGIT BACK IN C
        EX      DE,HL
        DEC     HL
        DEC     HL
        JR      ADIG4                   ; (+$2a) GO ADD TO SHOW BUFFER

  ; *** PRE DECIMAL POINT
ADIG1:
        DEC     HL              ; BACKUP TO DIGITS
        OR      (HL)            ; HOW MANY ENTRYS?
        JR      NZ,ADIG2                ; (+$12) ADIG2 IF NOT ZERO
  ; ****** CLear ENTry subroutine
CLRENT: CALL    SHOWCK          ; PUSH DOWN IF RESULT IS THERE
        PUSH    BC              ; CLEAR ARG2, SHOWBUF AND VARS
        SYSSUK  FILL            ; UPI FILL memory with data
        DW      ARG2            ; ... Memory Address
        DW      $0015           ; ... Byte Count = 21
        DB      $00             ; ... Data = 0
        ;
        LD      HL,DIGITS
        POP     BC
        OR      C               ; WAS DIGIT ZERO?
        JR      Z,ADIG6                 ; (+$1c) JUMP IF SO
  ; CODE FOR NONZERO PREVIOUS ENTRY OR NONZERO FIRST ENTRY
ADIG2:  LD      A,C
        CP      $0E             ; WAS CHAR A DECIMAL POINT?
        JR      NZ,ADIG3                ; (+$07) NO   ADIG3
        INC     HL              ; TO DECIMAL FLAG
        LD      (HL),A          ; SET IT
        INC     HL
        LD      (HL),$09        ; SET DECIMAL POINT POINTER
        JR      ADIG5                   ; (+$09) GO STUFF INTO SHOW BUFFER

ADIG3:  INC     DE              ; LEADING NUMERIC - STUFF IT
        INC     DE
        INC     DE
        INC     DE
        INC     DE
        CALL    ADDTB
ADIG4:  INC     (HL)            ; BUMP DIGIT COUNTER
ADIG5:  LD      A,C             ; GET DIGIT BACK
        LD      DE,SHOWBF       ; ADD DIGIT TO SHOW BUFFER
        CALL    ADDTB
ADIG6:  EX      DE,HL
ADIG7:  XYRELL  DE, $08, $53
        LD      C,00000111B     ; DISPLAY ALWAYS POSITIVE
  ; DISPLAY RECALL OR STORE?
        PUSH    HL
        LD      HL,MEMSTA
        LD      A,(HL)
        AND     A
        JR      Z,ADIG8                 ; (+$02) ADIG8 IF NOT WANTED
        ADD     A,'R' - $21
ADIG8:  ADD     A,$20
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
  ; SAME THING FOR STORE-RECALL DIGIT
        INC     HL
        LD      A,(HL)
        AND     A
        JR      Z,ADIG9                 ; (+$07)
        ADD     A,'0' - $21
        LD      (HL),$00        ; CLEAR DIGIT AND RECALL MODE
        DEC     HL
        LD      (HL),$00
ADIG9:  ADD     A,$20
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
        POP     HL
        LD      A,(OPCOD)
        LD      C,00000011B
        JP      NUMBA

  ; *** SUBROUTINE TO CHECK TO SEE IF WE ARE IN SHOW MODE
  ; IF SO - PUSH ARG2 AND GO INTO ARG1 MODE
SHOWCK:
        LD      HL,STATEV
        LD      A,(HL)
        CP      SHOWMD
        RET     NZ
        XOR     A
        LD      (HL),A
        DEC     HL
        LD      (HL),A          ; CLEAR OPERATOR
        RET

  ; *** SUBROUTINE TO ADD AN ENTRY TO THE STACK
  ; first PUSH STACK UP
PUSHEN:
        SYSSUK  SCROLL          ; UPI SCROLL area of screen
        DW      BYTEPL          ; ... Line Increment = 40
        DB      $08             ; ... Bytes = 8
        DB      MAXSTK          ; ... Rows = 92
        DW      $4020           ; ... Screen Address (FIRST LINE)
        ;
        LD      HL,STKBOT
        CALL    CONVRT          ; CONVERT AND ADD TO STACK
        LD      HL,NLINES       ; HOW MANY ENTRYS?
        LD      A,(HL)          ; ON THE STACK?
        CP      MAXSTK          ; AT MAX?
        JR      Z,PUSHE1                ; (+$01) JUMP IF SO
        INC     (HL)            ; BUMP ENTRY COUNT
  ; ARE WE AT TOP OR BOTTOM OF STACK?
PUSHE1: INC     HL              ; POINT AR SCHOLL POINTER
        LD      A,(HL)
        AND     A               ; JUMP IF AT BOTTOM
        JR      Z,SCRLU1                ; (+$19)
        CP      MAXSTK - 10     ; OR IF AT TOP
        JR      Z,SCRLU1                ; (+$15) FIX SCROLL POINTER
        INC     (HL)            ; FIX SCROLL POINTER
        RET                     ; AND QUIT

  ; ****** SCROLL KEY ENTRY
SCRLKY:
        XOR     A               ; CLEAR KEYSEX FOR SCROLL MOTOR
        LD      (KEYSEX),A
        LD      HL,NLINES
        LD      A,(HL)          ; A = NUMBER OF LINES ON STK
        BIT     0,C             ; UP OR DOWN
        JR      NZ,SCRLDN               ; (+$1a) JUMP IF DOWN
        AND     A               ; QUIT IF ZERO LINES ON STACK
        RET     Z
        INC     HL              ; WHERE'S THE SCROLL POINTER?
        LD      A,(HL)
        AND     A               ; QUIT IF AT BOTTOM
        RET     Z
        DEC     (HL)            ; ELSE DECREMENTETH
  ; PUSHEN JOINS HERE
SCRLU1: LD      A,$08           ; DO IT 8 TIMES
SCRLU2: SYSSUK  SCROLL          ; UPI SCROLL area of screen
        DW      BYTEPL          ; ... Line Increment = 40
        DB      $20             ; ... Bytes = 32
        DB      $4F             ; ... Rows = 79
        DW      $4028           ; ... Screen Address (NORMEM + 1*BYTEPL)
        ;
        DEC     A
        JR      NZ,SCRLU2               ; (-$0b)
        INC     A
        XYRELL  DE, $18, $49
        JR      SCRLD1                  ; (+$18) JOIN SCRLDN DISPLAY CALL

  ; *** SCROLL DOWN ROUTINE
SCRLDN:
        INC     HL              ; ADVANCE THE SCROLL POINTER
        SUB     (HL)            ; SUBTRACT FROM LINES ON STACK
        CP      $0B             ; FAR ENOUGH APART?
        RET     C               ; QUIT IF NOT
        INC     (HL)            ; YEAH - ADVANCE SCROLL POINTER
        LD      A,$08
SCRLD0: SYSSUK  SCROLL          ; UPI SCROLL area of screen
        DW      -BYTEPL         ; ... Line Increment = -40
        DB      $20             ; ... Bytes = 32
        DB      $50             ; ... Rows = 80
        DW      $4C80           ; ... Screen Address (NORMEM + 80*BYTEPL)
        ;
        DEC     A
        JR      NZ,SCRLD0               ; (-$0b)
        LD      A,$0A
        XYRELL  DE, $18, $01
SCRLD1: LD      HL,SCRPTR
        ADD     A,(HL)          ; ADD SCROLL PTR
        LD      HL,STKBOT + 1*BYTEPL
        LD      BC,-BYTEPL
SCRLD2: ADD     HL,BC
        DEC     A
        JR      NZ,SCRLD2               ; (-$04)
        LD      C,$0D           ; GET OPERATION
        SYSTEM  INDEXN          ; UPI INDEX Nibble by C
        ;
        LD      C,$00           ; WRITE WITH BLACK BACKGROUND
NUMBA:  PUSH    AF
        LD      B,C
        LD      C,$0C
        SYSTEM  INDEXN          ; UPI INDEX Nibble by C
        ;
        BIT     1,A
        JR      NZ,OVFERR               ; (+$16)
        BIT     0,A
        LD      A,00000100B     ; for color 1
        JR      Z,NUMBA1                ; (+$02)
        LD      A,00001000B     ; for color 2
NUMBA1: OR      B
        LD      C,A
        LD      B,10001011B     ; for ZERO SUPPress 11 digits
        SYSTEM  DISNUM          ; UPI DISplay NUMber
        ;
        POP     AF
        SYSSUK  INDEXB          ; UPI INDEX Byte by A
        DW      OPTBL           ; ... Base Address (OPerator TaBLe)
        ;
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
NULKEY: RET

OVFERR: LD      C,B
        SET     3,C
        LD      HL,OVFMSG
        SYSTEM  STRDIS          ; UPI STRing DISplay
        ;
        POP     AF
        RET

  ; ****** " *OVerFlow* " MeSsaGe
OVFMSG: 
        DB      $20,'*','O','V','E','R','F','L','O','W','*',$20
        DB      $00

  ; ****** OPerator TaBLe
OPTBL:
        DB      ' '             ; NULL OP
        DB      '+'
        DB      '-'
        DB      $62             ; multiply
        DB      $63             ; divide
        DB      '='

CONVRT: LD      DE,ARG2
        LD      A,(OPCOD)
        LD      C,$0D
        SYSTEM  STOREN          ; UPI STORE THAT Nibble in A by C
        ;
  ; TRANSFER SIGN
        LD      A,(ARG2+10)     ; GET SIGN BYTE
        RLCA                    ; PUT SIGN IN LOW BIT
        DEC     C
        SYSTEM  STOREN          ; UPI STORE Nibble in A by C
        ;
PACKER: EX      DE,HL           ; DE = SAVE, HL = ARG
        LD      BC,$0006        ; CLEAR SAVE
        XOR     A
        SYSTEM  FILL            ; UPI FILL memory with data
        ;
        LD      C,$13
CCVT1:  SYSTEM  INDEXN          ; UPI INDEX GET Nibble by C
        ;
        CALL    ADDTB           ; ADD INTO 'SAVE'
        JR      NZ,CCVT1A               ; (+$0d) JUMP IF DONE
        DEC     C               ; DECREMENT ARGPTR
        LD      A,C
        CP      $09             ; IS IT NOW 9
        JR      NZ,CCVT1                ; (-$0d) NO - KEEP GOING
        LD      A,$0E           ; YEAH - PUT IN DECIMAL POINT
        CALL    ADDTB
        JR      Z,CCVT1                 ; (-$14) JUMP BACK IF OKAY
  ; NOW REMOVE TRAILING ZEROS AND DECIMAL POINT
CCVT1A: EX      DE,HL           ; HL = SAVE
CCVT2:  LD      A,(HL)          ; GET LO NIBBLE
        AND     $0F             ; IS IT NONZERO?
        JR      NZ,CCVT3                ; (+$05) JUMP IF NOT ZERO
        CALL    KILTRZ          ; REMOVE ONE TRAILING ZERO
        JR      CCVT2                   ; (-$0a)

CCVT3:  CP      $0E             ; DECIMAL POINT
        RET     NZ              ; QUIT IF NOT
KILTRZ: LD      BC,$0006
        ADD     HL,BC
        AND     A
        LD      B,C
KILTR1: DEC     HL
        RRD
        DJNZ    KILTR1                  ; (-$05)
        RET

  ; *** SUBROUTINE TO ADD A DIGIT TO NUMBER POINTED AT BY DE
ADDTB:
        LD      B,$05
ADDTD:  EX      DE,HL
        PUSH    HL
ADDTB1: RLD
        INC     HL
        DJNZ    ADDTB1                  ; (-$05)
        LD      (HL),A          ; STUFF LAST
        POP     HL
        EX      DE,HL
        RET

  ; ****** COLORS FOR CALCULATOR
CALCLR:
        DB      $76             ; YELLOW color 3 left
if BALLY or WHITE ; ### conditional ###
        DB      $5C             ; # + # RED color 2
else ; ASTRO ####### conditional ###### else...
        DB      $62             ; # - # darker RED
endif ; ##### end conditional #########
        DB      $00             ; BLACK color 1
        DB      $07             ; WHITE color 0 left
        DW      $0707           ; WHITE on WHITE (to hide data) right
        DW      $0707           ; WHITE on WHITE (on right edge) colors

  ;
  ;       END OF CALCULATOR PROGRAM
  ;

; FOUR FILLER BYTES BETWEEN SCRIBBLING AND CALCULATOR TO MATCH ROM
        DB      $FF
if BALLY or WHITE ; ### conditional ###
        DB      $85             ; # ? # CHECKSUM #3 [same for both ??]
else ; ASTRO ####### conditional ###### else...
        DB      $FF             ; #-1 # CHECKSUM #3 [not set to 0 ??]
endif ; ##### end conditional #########
        DB      $FF,$FF

  ;
  ;*****************************
  ;* H V G   C H E C K M A T E *  VER. 2.0 4-2-2003
  ;*****************************
  ;

  ;
  ; M A C R O
  ;
WRECK   macro
        DW      ($09 SHL 8) + $20
        DB      00000000B       ; Color = 0
        endm

  ;
  ; E Q U A T E S
  ;
RLMOVE  EQU     00001100B       ; RIGHT AND LEFT MOVES
UDMOVE  EQU     00000011B       ; UP AND DOWN MOVES
NOBIT   EQU     $02             ; # OF GAME BITS
NPBIT   EQU     $03             ; # PLAYERS BIT
ANIMAX  EQU     $03             ; MAX # TICKS PER ANIMATION FRAMe

if BALLY or WHITE ; ### conditional ###
XMAX    EQU     (BYTEPL-1) * 4  ; # + # MAX X COORD [$9C]
else ; ASTRO ####### conditional ###### else...
XMAX    EQU     (BYTEPL-2) * 4  ; # - # MAX X COORD [$98]
endif ; ##### end conditional #########

YLINES  EQU     $15             ; # VERT BLOCKS
LOWY    EQU     $0B             ; LOWEST Y COORD
YMAX    EQU     ((YLINES-1) * 4) + LOWY ; MAX Y COORD
LOWX    EQU     $00             ; LOWEST X COORD
AMOVE   EQU     $08             ; AN ARBITRARY MOVE
MUSVOL  EQU     $09             ; MUSIC VOLUME
TDOPT   EQU     00100100B       ; TIME DISPLAY OPTIONS
CDOPT   EQU     01000100B       ; COUNT DOWN OPT
WRITOR  EQU     00010000B       ; WRITE WITH MAGIC OR

  ; PLAYER PACKER OFFSETS
LASTSW  EQU     +0              ; LAST SWITCH SETTING
LASTMV  EQU     +1              ; LAST ACTUAL MOVE
CURSW   EQU     +2              ; CURRENT SWITCH SETTING
AROT    EQU     +3              ; ARROW ROTATION SETTING
ARRX    EQU     +4              ; ARROW X COORD
ARRY    EQU     +5              ; ARROW Y COORD
PSTAT   EQU     +6              ; PLAYER STATUS
  
  ; PLAYER STATUS MASKS
ACTIVE  EQU     $80
HUMAN   EQU     $40
ACTBIT  EQU     7               ; 1=ACTIVE, 0=DEAD
HUMBIT  EQU     6               ; 1=HUMAN, 0=COMPUTER
  
  ; SCREEN TABS
XTAB1   EQU     ((BYTEPL/4) * 4)
XTAB2   EQU     XTAB1 * 2
XTAB3   EQU     XTAB1 * 3
YTAB    EQU     (((YLINES-1) / 4) * 4)
YTAB1   EQU     YTAB + LOWY
YTAB2   EQU     (2*YTAB) + LOWY
YTAB3   EQU     (3*YTAB) + LOWY
  
  ; OFFSETS FOR EACH PLAYERS ROM DATA
; NOTE0   EQU     +0            ; EACH DIRECTIONS NOTES
; NOTE1   EQU     +1            ; THESE FOUR LINES COMMENTED OUT BECAUSE
; NOTE2   EQU     +2            ; THEY INTERFERE WITH BALLYEQU.H AND ARE
; NOTE3   EQU     +3            ; NOT USED ANYWAY- ADAM, 2-16-2002
PPATL   EQU     +4              ; PLAYER PAT ADDR LOW
PPATH   EQU     +5              ; PLAYER PAT ADDR HIGH
PCDOP   EQU     +6              ; PLAYER CHAR DISP OPT
PSPOSX  EQU     +7              ; X COORD OF PLAYER SCORE
PSPOSY  EQU     +8              ; Y COORD OF PLAYER SCORE
PSDOP   EQU     +9              ; PLAYER SCORE DISP OPT
  
  ; MORE EQU'S
FORCEM  EQU     $F6             ; VAL TO FORCE RANDOM MOVE
WIDTH   EQU     $04             ; # PIXELS WIDE OF PLAYER PAT
HEIGHT  EQU     $04             ; # PIXELS HIGH OF PLAYER PAT
ALLBYT  EQU     (YLINES*4) * BYTEPL ; ALL BYTES ON SCREEN
STARTS  EQU     (LOWY*BYTEPL) + NORMEM ; LOWEST ADDR OF PLAY FI
PATXSZ  EQU     $01             ; #BYTES WIDE OF PLAYER PATTERN
PATYSZ  EQU     $04             ; #BYTES HIGH OF PLAYER PATTERN
PATDIM  EQU     (PATXSZ SHL 8) + PATYSZ ; PATTERNS DIMENSIONS
JUSJOY  EQU     $0F             ; ONLY JOY STICK BITS
CBLEN   EQU     $08             ; COLOR BLOCK LENGTH
SBLEN   EQU     $08             ; SOUND BLOCK LENGTH
; WPONOF  EQU     +0
; WPOPT   EQU     +1
; WPPAL   EQU     +2
; WPPAH   EQU     +3
; WPXSIZ  EQU     +5
; WPYSIZ  EQU     +4

  ; UNCLEARED RAM equates
UNCRAM  EQU     $4F96
CURSCR  EQU     $4F96           ; ALL CURRENT SCORES
  
  ; CLEARED RAM
CNOPL   EQU     $4FA2           ; CURRENT # PLAYERS
PLIX    EQU     $4FA3           ; WHO IS CURRENT PLAYER
CNOHUM  EQU     $4FA4           ; CURRENT # HUMANS
TARRX   EQU     $4FA5           ; TEMP ARROW X COORD
TARRY   EQU     $4FA6           ; TEMP ARROW X COORD
RMASK   EQU     $4FA7           ; ROTATE MASK
  
  ; PPACKS: START OF PLAYER PACKETS
PLAY0   EQU     $4FA8
PLAY1   EQU     $4FAF
PLAY2   EQU     $4FB6
PLAY3   EQU     $4FBD

  ; ====== CMSTRT: CheckMate STaRT equate
; L1328:
  ; ****** ONE TIME ONLY HOUSEKEEPING
ONETIM:
        LD      SP,UNCRAM
        SYSSUK  GETPAR          ; UPI GET game PARameter from user
        DW      NOGAME          ; ... Prompt Address for "# of GAMEs"
        DB      10000010B       ; ... Digits = 2 (with ZERO SUPPress)
        DW      CT7             ; ... Parameter Address
        ;
        SYSSUK  GETPAR          ; UPI GET game PARameter from user
        DW      NOPLAY          ; ... Prompt Address for "# of PLAYers"
        DB      00000001B       ; ... Digits = 1
        DW      NUMPLY          ; ... Parameter Address (NUMber of PLaYers)
        ;
        SYSSUK  FILL            ; UPI FILL memory with data
        DW      CURSCR          ; ... Memory Address (CURrent SCoRe)
        DW      $000C           ; ... Byte Count = 12
        DB      $00             ; ... Data = 0
        ;
  ; *** REENTRY POINT FROM END OF GAME
FIREIT: DI
        LD      SP,UNCRAM
        SYSTEM  INTPC           ; UPI INTerPret with Context create
        ;
  ; OUPUT COLOR BLOCK
        DO      COLSET          ; UPI COLors SET
        DW      CBLOCK          ; ... Table Address (Color BLOCK)
        ;
        DO      EMUSIC          ; UPI End playing MUSIC
        ;
  ; CLEAR JOYSTICKS
        DO      FILL            ; UPI FILL memory with data
        DW      OSW0            ; ... Memory Address (Old SWitch track)
        DW      $0004           ; ... Byte Count = 4
        DB      $00             ; ... Data = 0
        ;
  ;CLEAR ALL RAM DATA
        DO      FILL            ; UPI FILL memory with data
        DW      CNOPL           ; ... Memory Address (CNOPL)
        DW      (PLAY3+PSTAT) - CNOPL + 1  ; ... Byte Count = 34
        DB      $00             ; ... Data = 0
        ;
        DO      SETOUT          ; UPI SET some OUTput ports
        DB      ((YLINES*4)+LOWY)*2  ; ... VERBL*2 = 190
if BALLY or WHITE ; ### conditional ###
        DB      $40             ; # + #  ... HORCB/4 = 0, HOR COL BND = 1
else ; ASTRO ####### conditional ###### else...
        DB      $67             ; # - #  ... HORCB/4 = 39, Border = color 1
endif ; ##### end conditional #########
        DB      $08             ; ... INMOD = 8 (screen only)
        ;
  ; CLEAR SCORE BLOCKS
        DO      RECTAN          ; UPI paint a RECTANgle in color 1
        DB      $00             ; ... X = 0
        DB      $00             ; ... Y = 0
        DB      $A0             ; ... Width = 160
        DB      $0B             ; ... Height = 11
        DB      $55             ; ... Color = 01010101B
        ;
        DO      RECTAN          ; UPI paint a RECTANgle in color 0
if BALLY or WHITE ; ### conditional ### at ...
        DB      $80             ; # + # ... X = 128
        DB      $00             ; # = # ... Y = 0
        WRECK                   ; # + # [Width = 32, Height = 9 MACRO]
else ; ASTRO ###### conditional ####### else...
        DB      $7A             ; # - # ... X = 122
        DB      $00             ; # = # ... Y = 0
        DB      $1E             ; # - # ... Width = 30
        DB      $09             ; # = # ... Height = 9
        DB      00000000B       ; # = # ... Color = 0
endif ; ##### end conditional #########
        ;
        DO      RECTAN          ; UPI paint a RECTANgle in color 0
if BALLY or WHITE ; ### conditional ### at ...
        DB      $58             ; # + # ... X = 88
        DB      $00             ; # = # ... Y = 0
        WRECK                   ; # + # [ Width = 32, Height = 9 ]
else ; ASTRO ###### conditional ####### else...
        DB      $57             ; # - # ... X = 87
        DB      $00             ; # = # ... Y = 0
        DB      $1D             ; # - # ... Width = 29
        DB      $09             ; # = # ... Height = 9
        DB      00000000B       ; # = # ... Color = 0
endif ; ##### end conditional #########
        ;
        DO      RECTAN          ; UPI paint a RECTANgle in color 0
if BALLY or WHITE ; ### conditional ### at ...
        DB      $28             ; # - # ... X = 40
        DB      $00             ; # = # ... Y = 0
        WRECK                   ; # + # [ Width = 32, Height = 9 ]
else ; ASTRO ###### conditional ####### else...
        DB      $2B             ; # + # ... X = 43
        DB      $00             ; # = # ... Y = 0
        DB      $1D             ; # - # ... Width = 29
        DB      $09             ; # = # ... Height = 9
        DB      00000000B       ; # = # ... Color = 0
endif ; ##### end conditional #########
        ;
        DO      RECTAN          ; UPI paint a RECTANgle in color 0
if BALLY or WHITE ; ### conditional ### at ...
        DB      $00             ; # - # ... X = 0
        DB      $00             ; # = # ... Y = 0
        WRECK                   ; # - # [ Width = 32, Height = 9 ]
else ; ASTRO ###### conditional ####### else...
        DB      $04             ; # + # ... X = 4
        DB      $00             ; # = # ... Y = 0
        DB      $21             ; # + # ... Width = 33
        DB      $09             ; # = # ... Height = 9
        DB      00000000B       ; # = # ... Color = 0
endif ; ##### end conditional #########
        ;
        DO      ACTINT          ; UPI ACTivate sub timer INTerrupts
        ;
        DONT    XINTC           ; UPI eXit INTerpreter with Context
        ;
  ; INITIALIZE STARTING ADDRESS OF ARROWS
        LD      HL,(YTAB2 SHL 8) + XTAB1
        LD      (PLAY0+ARRX),HL
        LD      HL,(YTAB2 SHL 8) + XTAB3
        LD      (PLAY1+ARRX),HL
        LD      HL,(YTAB1 SHL 8) + XTAB2
        LD      (PLAY2+ARRX),HL
        LD      HL,(YTAB3 SHL 8) + XTAB2
        LD      (PLAY3+ARRX),HL
  ; CLEAR FIELD
        CALL    CLEARF
  ; DISPLAY # GAMES
        LD      IX,FNTSML
        SYSSUK  DISNUM          ; UPI DISplay NUMber
        DB      $4C             ; ... X = 76
        DB      $02             ; ... Y = 2
        DB      TDOPT           ; ... Options (to X-OR color 1)
        DB      01000010B       ; ... Extended (2 small digits) 
        DW      CT7             ; ... Number Address
        ;
  ; GET # HUMANS
DONTD:  LD      A,(NUMPLY)
        CP      $05
        JR      C,GOTNPL                ; (+$02)
        LD      A,$04
  ; GET # PLAYERS
  ;  IF HUMANS = 1 OR 0 OR > 4 THEN PLAYERS = 4 ELSE PLAyers = humans only
GOTNPL: LD      (CNOHUM),A
        CP      $02
        JR      C,FPLAY                 ; (+$04)
        CP      $05
        JR      C,ALLHUM                ; (+$02)
FPLAY:  LD      A,$04
ALLHUM: LD      (CNOPL),A
  ; *** INITIALIZE THE PLAYER PACKETS
  ; B = CURR # HUMANS
  ; C = CURR # PLAYERS
  ; D = THIS PLAYER #
; INTIPP:
        LD      A,(CNOHUM)
        LD      B,A
        LD      A,(CNOPL)
        LD      C,A
        LD      D,$00
GTPLIX: LD      A,D
        CALL    LDPLIX
        PUSH    BC
        PUSH    DE
  ; DISPLAY PATTERN # ON FIELD
        LD      A,D
        ADD     A,$31           ; SETUP ASCII LITERAL
        LD      E,(IX+ARRX)
        LD      D,(IX+ARRY)     ;??? IS THIS RIGHT?
        DEC     E
        DEC     E
        LD      C,(IY+PCDOP)
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
  ; DISPLAY # ON SCORE BLOCK
        LD      E,(IY+PSPOSX)
        LD      D,(IY+PSPOSY)
        PUSH    DE
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
  ; WRITe PLAYER PATtern IN SCORE BLOCK
        POP     DE
        LD      A,E
        ADD     A,$06
        LD      E,A
        INC     D
        INC     D
        LD      BC,(PATYSZ SHL 8) + PATXSZ
        LD      H,(IY+PPATH)
        LD      L,(IY+PPATL)
        LD      A,00010000B
        SYSTEM  WRIT            ; UPI WRITe with sizes provided
        ;
        POP     DE
        PUSH    DE
        PUSH    IX
        LD      A,D
        LD      B,$00
        LD      C,D
        LD      HL,CURSCR
        ADD     HL,BC
        ADD     HL,BC
        ADD     HL,BC
        CALL    DISPSC          ; DISPLAY SCORES
        POP     IX
        POP     DE
        POP     BC
        XOR     A
        OR      B
        JR      Z,NOTHUM                ; (+$09)
        LD      A,ACTIVE+HUMAN
        LD      (IX+PSTAT),A
        DEC     B
        JR      CKNOPL                  ; (+$06)

; CKSUM3:
        DB      $00             ; noT ??

NOTHUM: LD      A,ACTIVE
        LD      (IX+PSTAT),A
CKNOPL: INC     D
        DEC     C
        XOR     A
        OR      C
        JR      NZ,GTPLIX               ; (-$60)
        LD      A,$03
CDOWNL: PUSH    AF
        SYSSUK  PAWS            ; UPI PAUSE
        DB      $05             ; ... Interrupts = 5
        ;
        LD      (PLIX),A
        CALL    UPMUZK          ; MAKE SOUND FOR COUNTDOWN
        POP     AF
        PUSH    AF
        ADD     A,$30
        XYRELL  DE, XTAB2-4, YTAB2-4
        LD      C,CDOPT
  ; DISPLAY COUNTDOWN #
        SYSTEM  CHRDIS          ; UPI CHaRacter DISplay
        ;
        SYSSUK  PAWS            ; UPI PAUSE
        DB      $28             ; ... Interrupts = 40
        ;
        SYSTEM  EMUSIC          ; UPI End playing MUSIC
        ;
        POP     AF
        DEC     A
        JR      NZ,CDOWNL               ; (-$1e)
        CALL    CLEARF
  ; INT TICK COUNT
        CALL    TICKIT
        XOR     A
        LD      (CNT),A
LOOPY:  SYSSUK  SENTRY          ; UPI SENse TRansition Y
        DW      ALKEYS          ; ... Keymask Address (ALL KEYS)
        ;
        SYSSUK  DOIT            ; UPI DOIT table, branch to translation handler
        DW      THETBL          ; ... Table Address (THE TaBLe)
        ;
        JR      LOOPY                   ; (-$0a)

  ; ****** THE DOIT TaBLe
THETBL:
        RC   SCT0, ACTION, $00
        RC   SJ0, MOVJOY, $00
        RC   SJ1, MOVJOY, $00
        RC   SJ2, MOVJOY, $00
        RC   SJ3, MOVJOY, $00
        RC   SKYD, CALPIZ, ENDx

ACTION:
        CALL    TICKIT
  ; INCREMENT THE CURRENT PLAYER INDEX BY 1 UNTIL
  ; AN ACTIVE PLAYER IS FOUND THEN UPDATE HIM
INCIX:  LD      A,(PLIX)
        INC     A
        AND     $03
        LD      (PLIX),A        ; CURR PLAYER IX <- CURR PL IX+1 M
        CALL    LDPLIX
        BIT     ACTBIT,(IX+PSTAT) ; TEST FOR ACTIVE PLAYER
        JR      Z,INCIX                 ; (-$12)
        JP      MOVEIT          ; THE MAJOR EVENT

MOVJOY:
        SUB     SJ0             ; TAKE OFF WHATEVER
        SRL     A               ; DIVIDE BY 2
        JP      STALL

CALPIZ:
        CALL    TICKIT
        SYSTEM  PIZBRK          ; UPI take a PIZza BReaK
        ;
        RET

UPMUZK:
        LD      A,(PLIX)
        CALL    LDPLIY
        LD      A,(IX+AROT)
        LD      B,$03
TSTBIT: RRCA
        JR      C,GOTBIT                ; (+$02)
        DJNZ    TSTBIT                  ; (-$05)
GOTBIT: LD      C,B
        LD      B,$00
        ADD     IY,BC
        LD      A,(IY+$00)
        OUT     (TONEC),A
        LD      A,MUSVOL
        OUT     (VOLC),A
        LD      A,OA4
        OUT     (TONMO),A
        RET

  ; *** CLEAR FIELD subroutine
CLEARF:
        SYSSUK  FILL            ; UPI FILL memory with data
        DW      STARTS          ; ... Memory Address
        DW      ALLBYT          ; ... Byte Count
        DB      $00             ; ... Data = 0
        ;
        RET

  ; ****** THIS ROUTINE UPDATES A PLAYER'S POSITION
  ; Input PARAS Are: IX = POINTER TO PLAYERS PACKET
  ; DURING ROUTINE B = CURRENT SWITCH, C = LAST SWITCH
MOVEIT:
        LD      C,(IX+LASTSW)
        LD      B,(IX+CURSW)
        BIT     HUMBIT,(IX+PSTAT)
        JR      NZ,NOCUR                ; (+$03) IF NOT HUMAN
ZSW:    XOR     A               ; CLEAR A
        LD      B,A             ; CLEAR CURRENT SWITCH
        LD      C,A             ; CLEAR LAST SWITCH ENDIF
NOCUR:  LD      A,B             ; IF CURR SW = 0
        OR      A
        JR      NZ,RANTST               ; (+$01)
        LD      B,C             ; THEN CURR SW <- LAST SW ENDIF
RANTST: LD      (IX+LASTSW),B   ; SAVE LAST SW
        LD      A,B             ; IF CURR SW = 0
        OR      A
        JR      NZ,GOTSW                ; (+$05)
        LD      C,$00           ; LAST SW <- 0
        CALL    RANMOV          ; GET RANDOM MOVE ENDIF
GOTSW:  LD      A,(IX+LASTMV)   ; GET LAST MOVE
        CALL    MOVTST
        JR      Z,GOTMOV                ; (+$13)
  ; ANY MOVE AND CURR SW
        CALL    MOVANY
        JR      Z,GOTMOV                ; (+$0e)
        LD      B,C             ; TRY LAST SW
  ; ANY MOVE
        CALL    MOVANY
        JR      Z,GOTMOV                ; (+$08)
        LD      B,(IX+LASTMV)   ; TRY LAST MOVE
  ; ANY MOVE
        CALL    MOVANY
        JR      NZ,CRASH                ; (+$3c)
  ; A LEGIT MOVE HAS BEEN FOUND SO UPDATE THE GUY
GOTMOV: LD      (IX+LASTMV),A   ; SAVE ACTUAL MOVE FOR LATER
        LD      (IX+AROT),A     ; ARROW ROTATION AMOUNT <- THE MOVe
        LD      D,(IX+ARRY)
        LD      E,(IX+ARRX)
        CALL    ERASE
        LD      H,(IY+PPATH)
        LD      L,(IY+PPATL)
        LD      BC,(PATYSZ SHL 8) + PATXSZ
        LD      A,WRITOR
  ; WRITE PLAYER PATTERN OVER ARROw
        SYSTEM  WRIT            ; UPI WRITe with sizes provided
        ;
        LD      A,(TARRX)
        LD      (IX+ARRX),A     ; SAVE NEW ARROW X
        LD      A,(TARRY)
        LD      (IX+ARRY),A     ; SAVE NEW ARROW Y
        CALL    ANIARR          ; ANIMATE THE ARROW
        JP      UPMUZK

ERASE:
        PUSH    DE
        SYSSUK  RELAB1          ; UPI RELative to non-magic ABSolute
        DB      $00             ; ... Magic = 0
        ;
        EX      DE,HL
        LD      B,$00
        LD      DE,(PATYSZ SHL 8) + PATXSZ
        SYSTEM  BLANK           ; UPI BLANK area
        ;
        POP     DE
        RET

  ; A PLAYER HAS CRASHED.  DESTROY HIS ARROW AND
  ; ELIMINATE HIM FROM THE GAME.
CRASH:  LD      BC,EXPATS
        LD      DE,EXCOLS       ; DE <- EXPLODE COLOR TABLE ADDR
        LD      A,$05
        LD      HL,EXPSND
EXLOOP: PUSH    AF              ; PUSH LOOP COUNT
        PUSH    BC              ; PUSH EXT PAT ADDRESS
        PUSH    DE              ; PUSH EXPLODE COLOR TABLE ADDR
        PUSH    HL              ; PUSH EXPLODE SOUND ADDRESS
        LD      A,(DE)          ; A <- EXPLODE COLOR
if BALLY or WHITE ; ### conditional ###
        OUT     (COL0R),A       ; # > # FLASH Right side
else ; ASTRO ####### conditional ###### else...
        OUT     (COL0L),A       ; # < # FLASH Left side
endif ; ##### end conditional #########
        PUSH    BC
  ;WRIT EXPLOSION
        LD      D,(IX+ARRY)
        LD      E,(IX+ARRX)
        CALL    ERASE
        POP     HL              ; PATTERN ADDRESS
        LD      A,$10
        LD      BC,(PATYSZ SHL 8) + PATXSZ
        SYSTEM  WRIT            ; UPI WRITe EXPLOSION
        ;
        SYSSUK  PAWS            ; UPI PAUSE
        DB      $07             ; ... Interrupts = 7
        ;
        POP     HL              ; GET EXPLODE SOUND ADDR
        LD      BC,(SBLEN SHL 8) + SNDBX
        OTIR
        POP     DE
        POP     BC
        POP     AF
        DEC     A
        JR      Z,EXPFIN                ; (+$07) LOOP COUNT EXPIRED
        INC     DE              ; INC TO NEXT COLOR
        INC     BC              ; BUMP UP TO NEXT PAT ADDR
        INC     BC
        INC     BC
        INC     BC
        JR      EXLOOP                  ; (-$2f)

EXPFIN:
        LD      D,(IY+PPATH)
        LD      E,(IY+PPATL)    ; DE <- PLAYER PAT ADDR
        LD      IY,$0000
        ADD     IY,DE           ; IY <- PLAYER PAT ADDR
        LD      DE,($04 SHL 8) + $00  ; D <- LOOP COUNT
STOMP:  LD      A,(IY+$00)      ; A <- BYTE OF PLAYER PATTERN
        LD      HL,STARTS
        LD      BC,ALLBYT
STLOOP: CPIR
        JR      NZ,RESTOM               ; (+$05)
        INC     BC
        DEC     HL
        LD      (HL),E
        JR      STLOOP                  ; (-$09)

RESTOM:
        INC     IY
        DEC     D
        JR      NZ,STOMP                ; (-$17)
        BIT     HUMBIT,(IX+PSTAT)
        JR      Z,KILLST                ; (+$04) IF HUMAN
        LD      HL,CNOHUM
        DEC     (HL)            ; DEC CURRENT # HUMANS END IF zero
KILLST: RES     ACTBIT,(IX+PSTAT) ; KILL STATUS
  ; INC ALL ACTIVE PLAYERS SCORES
        LD      C,$04
BUMPEM: DEC     C
        LD      A,C
        CALL    LDPLIX
        BIT     ACTBIT,(IX+PSTAT)
        JR      Z,BUMPCK                ; (+$18)
        LD      B,$00
        PUSH    BC
        LD      A,C
        LD      HL,CURSCR
        ADD     HL,BC
        ADD     HL,BC
        ADD     HL,BC
        SCF
        CALL    DISPSC
        POP     BC
        SYSTEM  INCSCR          ; UPI INCrement SCoRe
        ;
        LD      A,C
        PUSH    BC
        OR      A
        CALL    DISPSC
        POP     BC
BUMPCK: SYSSUK  PAWS            ; UPI PAUSE
        DB      $1E             ; ... Interrupts = 30
        ;
        LD      A,C
        OR      A
        JR      NZ,BUMPEM               ; (-$2a)
  ; DEC CURR # PLAYERS
  ; IF CURR # PLAYERS .LEQ. 0, GO TO END OF GAME
        LD      HL,CNOPL
        DEC     (HL)
        DEC     (HL)
        JR      Z,ENDCHK                ; (+$02)
        INC     (HL)
        RET

ENDCHK:
        LD      A,(CT7)
        DEC     A
        DAA
        LD      (CT7),A
        JP      NZ,FIREIT
        SYSTEM  QUIT            ; UPI QUIT cassette execution
        ;
  ; ****** DISPLAY SCORE
  ; A = PLAYER #
  ; HL -> LAST BYTE OF SCORE
DISPSC:
        LD      C,(IY+PSDOP)
        JR      NC,NOTXOR               ; (+$04)
        RES     MROR,C
        SET     MRXOR,C
NOTXOR: LD      E,(IY+PSPOSX)
        LD      D,(IY+PSPOSY)
        LD      A,$0C
        ADD     A,E
        LD      E,A
        INC     D
        LD      B,01000011B     ; for 3 small digits
        LD      IX,FNTSML
        SYSTEM  DISNUM          ; UPI DISplay NUMber
        ;
        RET

  ; ****** ANIMATE THE ARROW subroutine
  ; Input AND Output IS IX WHO STAYS THE SAME
  ; DESTROYS ALL OTHER REGISTERS
ANIARR:
        BIT     ACTBIT,(IX+PSTAT)
        RET     Z               ; EXIT IF NOT ACTIVE
        LD      A,(IX+AROT)
        CALL    GETROT          ; HL <- ARROW PAT ADDR
        LD      D,(IX+ARRY)
        LD      E,(IX+ARRX)
        PUSH    HL
        CALL    ERASE
        POP     HL
        LD      BC,(PATYSZ SHL 8) + PATXSZ
        LD      A,WRITOR
        SYSTEM  WRIT            ; UPI WRITe with sizes provided
        ;
        RET

  ; ****** THIS ROUTINE TAKES CARE OF ARROW ANIMATION
  ; AND SHOWING A PLAYER HIS CURRENT JOYSTICK POSITion
  ; A = WHICH PLAYER
  ; B = JOYSTICK BITS
STALL:
        CALL    LDPLIX          ; IX <- ADDR OF PLAYER PACKET
        XOR     A
        OR      B
        JR      NZ,STORIT               ; (+$03)
        LD      A,(IX+CURSW)
STORIT: LD      (IX+CURSW),A
        XOR     (IX+LASTMV)     ; A <- DIFFERENCE FROM LAST MOVE
        JR      Z,GETLM                 ; (+$12) IF DIFFERENCE = 0 USE LAST MOVE
        XOR     RLMOVE
        JR      Z,GETLM                 ; (+$0e)
        XOR     RLMOVE
        XOR     UDMOVE
        JR      Z,GETLM                 ; (+$08)
        XOR     UDMOVE
HUMCHK: BIT     HUMBIT,(IX+PSTAT)
        JR      NZ,GOTIT                ; (+$03) IF HUMAN WE'VE GOT IT
GETLM:  LD      A,(IX+LASTMV)   ; GET LAST MOVE
GOTIT:  LD      (IX+AROT),A     ; STORE ARROW ROTATION
        JR      ANIARR                  ; (-$4a)

  ; *** TICK COUNT <- (8-CURR # PLAYERS)
TICKIT: LD      A,(CNOHUM)
        OR      A
        LD      A,$02
        JR      Z,STICK                 ; (+$06)
        LD      HL,CNOPL
        LD      A,$08
        SUB     (HL)
STICK:  LD      (CT0),A
        RET

  ; ****** LOAD IY WITH POINTER TO CURR PLAYER ROM DATA
  ; LOAD IX WITH A POINTER TO CURRENT PLAYER PACKET
  ; A = PLAYER #  MUST BE .GEQ. 0 & .LEQ. 3  [0 through 3]
LDPLIX:
LDPLIY:
        PUSH    DE
        PUSH    HL
        SYSSUK  INDEXW          ; UPI INDEX Word by A
        DW      ROMTBL          ; ... Base Address (ROM TaBLe)
        ;
        PUSH    DE
        POP     IY
        SYSSUK  INDEXW          ; UPI INDEX Word by A
        DW      RAMTBL          ; ... Base Address (RAM TaBLe)
        ;
        PUSH    DE
        POP     IX
        POP     HL
        POP     DE
        RET

ROMTBL:
        DW      PLROM0
        DW      PLROM1
        DW      PLROM2
        DW      PLROM3

RAMTBL:
        DW      PLAY0
        DW      PLAY1
        DW      PLAY2
        DW      PLAY3

  ; ****** GENERATE A RANDOM MOVE FOR THE PLAYER PACKET POInter
  ; Input AND Output:
  ; B = CURRENT SWITCH, C = LAST SWITCH
RANMOV:
        SYSSUK  RANGED          ; UPI RANGED random number
        DB      $20             ; ... Cutoff = 32
        ;
        OR      A               ; TIME TO CHANGE DIRECTION?
        JR      Z,NEWMOV                ; (+$08)
        LD      B,(IX+LASTMV)   ; USE LAST MOVE
        LD      A,B
        CALL    MOVTST
        RET     Z               ; LAST MOVE IS GOOD ENOUGH
NEWMOV: SYSSUK  RANGED          ; UPI RANGED random number
        DB      $04             ; ... Cutoff = 4
        ;
        LD      B,A
        INC     B
        LD      A,$80
SHFTIT: RLCA
        DJNZ    SHFTIT                  ; (-$03)
        LD      B,A
RANFIN: LD      A,$08
        CALL    MOVTST
        JR      NZ,ANYMOV               ; (+$02)
        LD      B,A
        RET

ANYMOV:
        LD      B,$0F           ; TRY ALL MOVES
        LD      A,$08
        CALL    MOVTST
        LD      B,A
        RET

MOVANY:
        LD      A,AMOVE
  ; ****** TEST THE NEW MOVE FOR VALIDITY routine
  ; The Inputs and Outputs:
  ; B = A SET ??? MOVES TO BE TESTED (IS UNCHANGED)
  ; C = UNCHANGED
  ; A = FIRST MOVE TO TEST, VALUE OF GOOD MOVE ON OUTPUt
  ; DE, HL = RETURNED UNCHANGED
  ; D = # ROTATES
  ; Z FLAG = Z IF GOOD MOVE FOUND (A CONTAINS FIRST GOOd move)
  ; Z FLAG = ??? IF NO GOOD MOVES FOUND (IN B)
MOVTST:
        PUSH    DE
        LD      D,$08           ; INIT # ROTATES
ROTMSK: RRCA                    ; ROTATE TO NEXT MOVE
        LD      E,A
        AND     B
        CALL    CHKMOV          ; CHECK MOVE
        LD      A,E
        JR      Z,MOVEXT                ; (+$06) FOUND ONE
        DEC     D               ; DEC # ROTATES
        JR      NZ,ROTMSK               ; (-$0c)
        SCF                     ; NO GOOD MOVES
        RL      D               ; SET Z FLAG=NZ
MOVEXT: POP     DE
        RET

  ; ****** CHECK THE MOVE IN A FOR BEING UNOCCUPIED routine
  ; Input AND Output:
  ; A = UP, DOWN, RIGHT OR LEFT BIT (RETURN UNCHANGED)
  ; Z FLAG = Z IF MOVE IN A IS TO AN EMPTY POSITION
  ; Z FLAG = NZ IF MOVE IN A IS BAD
  ; BC, DE, HL RETURNED UNTOUCHED
  ; IX = POINTER TO CURRENT PLAYER PACKET
  ; Local to this routine:
  ;    D = TEMP X COORD OF ARROW
  ;    E = TEMP Y COORD OF ARROW
CHKMOV:
        PUSH    BC
        PUSH    DE
        PUSH    HL
        PUSH    AF
        LD      D,(IX+ARRX)     ; GET X COORD OF ARROW
        LD      E,(IX+ARRY)     ; GET Y COORD OF ARROW
TLEFT:  BIT     CHLEFT,A
        JR      Z,TRIGHT                ; (+$0a)
        LD      A,D             ; GET A LEFT MOVE
        CP      LOWX
        JR      Z,BADMOV                ; (+$2f) ALREADY AT LOWEST X
        SUB     WIDTH           ; DEC TEMP X BY 1 POSITION
        LD      D,A
        JR      LOOKSQ                  ; (+$30)

TRIGHT:
        BIT     CHRIGH,A
        JR      Z,TUP                   ; (+$0a)
        LD      A,D             ; GOT A RIGHT MOVE
        CP      XMAX
        JR      NC,BADMOV               ; (+$21) ALREADY .GEQ. MAX X
        ADD     A,WIDTH
        LD      D,A
        JR      LOOKSQ                  ; (+$22)

TUP:
        BIT     CHUP,A
        JR      Z,TDOWN                 ; (+$0a)
        LD      A,E             ; GOT AN UP MOVE
        CP      LOWY
        JR      Z,BADMOV                ; (+$13) ALREADY AT LOWEST Y
        SUB     HEIGHT          ; DEC TEMP Y BY 1 POSITION
        LD      E,A
        JR      LOOKSQ                  ; (+$14)

TDOWN:
        BIT     CHDOWN,A
        JR      Z,BADMOV                ; (+$0a)
        LD      A,E             ; GOT A DOWN MOVE
        CP      YMAX
        JR      Z,BADMOV                ; (+$05) ALREADY AT HIGHEST Y
        ADD     A,HEIGHT        ; INC TEMP Y BY 1 POSITION
        LD      E,A
        JR      LOOKSQ                  ; (+$06)

BADMOV:
        POP     AF
        SCF
        RL      D               ; SET ??? FLAG = NZ
        JR      MOVEND                  ; (+$23)

  ; *** SEE IF THE NEW SQUARE IS OCCUPIED
LOOKSQ:
        PUSH    DE
        PUSH    DE
        POP     BC
        LD      D,C             ; REVERSE X,Y FOR SYSTEM
        LD      E,B
        SYSSUK  RELAB1          ; UPI RELative to non-magic ABSolute
        DB      $00             ; ... Magic = 0
        ;
        POP     HL
        EX      DE,HL
        LD      A,(HL)
        OR      A               ; TEST SQUARE
        JR      NZ,BADMOV               ; (-$14)
        LD      BC,BYTEPL
        ADD     HL,BC
        LD      A,(HL)
        OR      A
        JR      NZ,BADMOV               ; (-$1c)
        LD      A,D
        LD      (TARRX),A       ; STORE TEMP ARROW X COORD
        LD      A,E
        LD      (TARRY),A       ; STORE TEMP ARROW Y COORD
        POP     AF
        LD      D,$00
        SRL     D               ; SET Z FLAG=Z
MOVEND: POP     HL
        POP     DE
        POP     BC
        RET

  ; *** HL <- BASE ADDR OF ROTATED PATTERN
  ; A <- DIRECTION OF ROTATION
  ; IF A HAS MORE THAN 1 BIT SET THEN ONLY ONE IS US
GETROT:
        LD      HL,AUP
        BIT     CHUP,A
        RET     NZ
        LD      HL,ADOWN
        BIT     CHDOWN,A
        RET     NZ
        LD      HL,ARIGHT
        BIT     CHRIGH,A
        RET     NZ
        LD      HL,ALEFT
        RET

  ; ****** START OF ROM DATA FOR EACH PLAYER.
  ; CONTAINS 4 PLAYER NOTES, PLAYER PATTERN ADDR
  ;  , PLAYER CHAR DISP OPT
  ; PLAYER SCORE POSITION [X varies, Y always = 1]
  ; AND PLAYER SCORE DISP OPT
PLROM0:
PNOTE0: DB      G0,GS0,A0,AS0
PPADR0: DW      PPAT0
PCDOP0: DB      00011000B       ; (OR color 2)
if BALLY or WHITE ; ### conditional ###
PSPOS0: DB      $04             ; # < #
else ; ASTRO ####### conditional ###### else...
PSPOS0: DB      $0C             ; # > #
endif ; ##### end conditional #########
        DB      $01
PSDOP0: DB      00011000B       ; (OR color 2)

PLROM1:
PNOTE1: DB      B0,C1,CS1,D1
PPADR1: DW      PPAT1
PCDOP1: DB      00011100B       ; (OR color 3)
if BALLY or WHITE ; ### conditional ###
PSPOS1: DB      $85             ; # > #
else ; ASTRO ####### conditional ###### else...
PSPOS1: DB      $7D             ; # < #
endif ; ##### end conditional #########
        DB      $01
PSDOP1: DB      00011100B       ; (OR color 3)

PLROM2:
PNOTE2: DB      DS1,E1,F1,FS1
PPADR2: DW      PPAT2
PCDOP2: DB      00011100B       ; (OR color 3)
PSPOS2: DW      ($01 SHL 8) + $2D
PSDOP2: DB      00011100B       ; (OR color 3)

PLROM3:
PNOTE3: DB      G1,GS1,A1,AS1
PPADR3: DW      PPAT3
PCDOP3: DB      00011000B       ; (OR color 2)
if BALLY or WHITE ; ### conditional ###
PSPOS3: DB      $5D             ; # > #
else ; ASTRO ####### conditional ###### else...
PSPOS3: DB      $59             ; # < #
endif ; ##### end conditional #########
        DB      $01
PSDOP3: DB      011000B         ; (OR color 2)

  ; ****** EXPLOSION color PATTERNS
EXPATS:
EXPAT1:
        DB      $00             ; 0 0 0 0       
        DB      00010100B       ; 0 1 1 0       
        DB      00010100B       ; 0 1 1 0       
        DB      $00             ; 0 0 0 0       

EXPAT2:
        DB      $00             ; 0 0 0 0       
        DB      01000101B       ; 1 0 1 1       
        DB      01010001B       ; 1 1 0 1       
        DB      $00             ; 0 0 0 0       

EXPAT3:
        DB      00000101B       ; 0 0 1 1       
        DB      01000000B       ; 1 0 0 0       
        DB      00000001B       ; 0 0 0 1       
        DB      01010000B       ; 1 1 0 0       

EXPAT4:
        DB      00010001B       ; 0 1 0 1       
        DB      01000000B       ; 1 0 0 0       
        DB      00000001B       ; 0 0 0 1       
        DB      01000100B       ; 1 0 1 0       

EXPAT5:
        DB      $00             ; 0 0 0 0       
        DB      $00             ; 0 0 0 0       
        DB      $00             ; 0 0 0 0       
        DB      $00             ; 0 0 0 0       

  ; ****** EXPLOSION COLORS sequence
EXCOLS:
        DB      $07             ; White
        DB      $03             ; Gray
        DB      $07             ; White
        DB      $03             ; Gray
if BALLY or WHITE ; ### conditional ###
        DB      $77             ; # ~ # Orange
else ; ASTRO ####### conditional ###### else...
        DB      $87             ; # ~ # Yellow
endif ; ##### end conditional #########

  ; *** COUNT DOWN DISPLAY PACKET
CDCOLR: DW      ($00 SHL 8) + 00000100B

  ; *** TIMER DISPLAY PACKET
TDPACK: DW      (10000000B SHL 8) + 00000001B

  ; *** ARROW ANIMATION PATTERNS FOR EACH ROTATION
AUP:
        DB      00010100B       ; 0 1 1 0       
        DB      01010101B       ; 1 1 1 1       
        DB      01000001B       ; 1 0 0 1       
        DB      $00             ; 0 0 0 0       

ARIGHT:
        DB      00010100B       ; 0 1 1 0       
        DB      00000101B       ; 0 0 1 1       
        DB      00000101B       ; 0 0 1 1       
        DB      00010100B       ; 0 1 1 0       

ADOWN:
        DB      $00             ; 0 0 0 0       
        DB      01000001B       ; 1 0 0 1       
        DB      01010101B       ; 1 1 1 1       
        DB      00010100B       ; 0 1 1 0       

ALEFT:
        DB      00010100B       ; 0 1 1 0       
        DB      01010000B       ; 1 1 0 0       
        DB      01010000B       ; 1 1 0 0       
        DB      00010100B       ; 0 1 1 0       

  ; ****** PLAYER color PATTERNS
PPAT0:
        DB      00001000B       ; 0 0 2 0       
        DB      10101000B       ; 2 2 2 0       
        DB      00101010B       ; 0 2 2 2       
        DB      00100000B       ; 0 2 0 0       

PPAT1:
        DB      11111111B       ; 3 3 3 3       
        DB      11000011B       ; 3 0 0 3       
        DB      11000011B       ; 3 0 0 3       
        DB      11111111B       ; 3 3 3 3       

PPAT2:
        DB      00001100B       ; 0 0 3 0       
        DB      11111100B       ; 3 3 3 0       
        DB      00111111B       ; 0 3 3 3       
        DB      00110000B       ; 0 3 0 0       

PPAT3:
        DB      10101010B       ; 2 2 2 2       
        DB      10000010B       ; 2 0 0 2       
        DB      10000010B       ; 2 0 0 2       
        DB      10101010B       ; 2 2 2 2       

  ; COLOR BLOCK (# X #'s are color 1 value and Not used)
CBLOCK:
if BALLY or WHITE ; ### conditional ###
        DB      $F8             ; # X # (left side NOT used)
        DB      $F8             ; # X #
        DB      $F8             ; # = #'s are unchanged
        DB      $F8             ; # X #
        DB      $B5             ; # L # GREEN color 3 (players 2 & 3)
        DB      $52             ; # L # RED color 2 (players 1 & 4)
        DB      $F8             ; # = # dark BLUE (border and arrow heads)
        DB      $77             ; # L # YELLOW color 0 right (playfield)
else ; ASTRO ###### conditional ####### else...
        DB      $B3             ; # R # darker GREEN color 3 left
        DB      $62             ; # R # different RED color 2
        DB      $F8             ; # = # dark BLUE color 1
        DB      $87             ; # R # different YELLOW color 0
        DB      $F8             ; # X #
        DB      $F8             ; # X #
        DB      $F8             ; # = #
        DB      $F8             ; # X # (NO right side)
endif ; ##### end conditional #########

  ; ****** EXPLOSION SOUNDS
EXPSND: DB      $EF,$FF,$3F,$00,$FF,$FD,$F5,$F5
        DB      $8F,$EE,$3E,$00,$FF,$FD,$F5,$F5
        DB      $4E,$88,$38,$00,$FF,$FD,$F5,$F5
        DB      $48,$44,$34,$00,$FF,$FD,$F5,$F5
        DB      $00,$00,$00,$00,$00,$00,$00,$00

  ;
  ;       END OF CHECKMATE PROGRAM
  ;

; FIVE FILLER BYTES BETWEEN CHECKMATE AND GUNFIGHT TO MATCH ROM
        DB      $FF,$FF,$FF,$FF,$FF

  ;
  ; ***************************
  ; * H V G   G U N F I G H T * ; VER. 1.0 - DEC. 15, 2003
  ; ***************************
  ;

  ; Program Name - Gun Fight
  ;                for the Bally Astrocade
  ; Retyped by Adam Trionfo

  ; ====== GFSTRT: GunFight STaRT equate
; L17DE:
        JP    INIT

  ; *********************
  ; * GUN FIGHT EQUATES *
  ; *********************
  ; GUNFIGHT BACKGROUND JOB
  ; CONSISTING OF INITIALIZATION, PRE-ROUND DISPLAY,
  ; MONITORING OF CONTROLS AND VECTOR DELTA CHANGING
  ; DEATH, POST ROUND STUFF AND END GAME

  ; EQUATES
LNX     EQU     $08             ; LEFT NUMBER X
BSY     EQU     $02             ; BANNER STRINGS Y
RNX     EQU     $88             ; RIGHT NUMBER X = 136
LBULX   EQU     $20             ; LEFT BULLETS X = 32
RBULX   EQU     $68             ; RIGHT BULLETS X = 104
STMRX   EQU     $4C             ; SUB TIMER X = 76
GRX     EQU     $2C             ; GET READY X = 44
GRY     EQU     $01             ; GET READY Y = 1
DRX     EQU     $40             ; DRAW X = 64
TCACY   EQU     $14             ; TOP CACTUS Y = 20
TTREEY  EQU     TCACY - 5
MCACY   EQU     $2A             ; MID CACTUS Y = 42
BCACY   EQU     $46             ; BOTTOM CACTUS Y = 70
BTREEY  EQU     BCACY - 5
LCACX   EQU     $40             ; LEFT CACTUS X = 64
RCACX   EQU     $58             ; RIGHT CACTUS X = 88
CCACX   EQU     $4C             ; CENTER CACTUS X = 76
WAGX    EQU     $48             ; WAGON X = 72
COWX    EQU     RCACX + 8       ; OTHER COWBOYS WINDOW X

TLINE   EQU     $0A             ; TOP LINE OF GUNSPACE = 10
ALINE   EQU     TLINE - 1
BLINE   EQU     $5C             ; BOTTOM LINE OF GUNSPACE = 92

BULVSZ  EQU     $12             ; BULLET VECTOR SIZE = 18
GFVSIZ  EQU     $17             ; GunFight Vector SIZe = 23
WAGVSZ  EQU     $12             ; WAGON VECTOR SIZE = 18

WINBND  EQU     $32             ; TOP BOTTOM WINDOW BOUNDARY = 50
TOPLIN  EQU     $35 * 2         ; TOP WINDOWS LINE = 106
BOTLIN  EQU     $00             ; BOTTOM WINDOW LINE = 0
LFRLIN  EQU     $64 * 2         ; LOW PRIORITY FOREGROUND LINE = 200

NEXT    EQU     -1              ; NEXT LINE FOR QUEUES negative offset
VBARM   EQU     VBOAH + 1       ; ARM STATE
VBOARM  EQU     VBARM + 1       ; LAST ARM PATTERN WRITTEN
VBLEGT  EQU     VBOARM + 1      ; LEG TIMER
VBLEG   EQU     VBLEGT + 1      ; LEG LINK
VBCOMP  EQU     VBLEG + 1       ; TIMER FOR COMPUTER CONTROL
 
  ; BITS
VBSWAG  EQU     +0              ; WAGON BIT
VBSCHG  EQU     +3              ; CHANGE STATUS BIT
VBSNOM  EQU     +4              ; NOT MOVING STATUS
VBSINT  EQU     +5              ; INTERCEPTED/DEAD STATUS

  ; *************
  ; * RAM CELLS *
  ; *************

STACK   EQU     $4F06           ; (150 bytes) START STACK HERE
MSTACK  EQU     $4F12           ; (12 bytes) Music STACK
STRRAM  EQU     $4F12
WRITQ   EQU     $4F12           ; WRITE Q HEADER
VECQ    EQU     $4F15           ; VECTOR Q HEADER
VECSTR  EQU     $4F18
BULV1   EQU     $4F18           ; BULLET VECTOR 1
BULV2   EQU     $4F2A           ; BULLET VECTOR 2
BULV3   EQU     $4F3C           ; BULLET VECTOR 3
BULV4   EQU     $4F4E           ; BULLET VECTOR 4

  ; LEFT COMWBOY LINK
LCOWB   EQU     $4F61           ; LEFT GUNFIGHTER
  
  ; RIGHT COWBOY LINK
RCOWB   EQU     $4F78           ; RIGHT GUNFIGHTER
WAGVEC  EQU     $4F8F           ; WAGON VECTOR
WAGON   EQU     $4F90           ; WAGVEC+VBSTAT
ENDRAM  EQU     $4FA1
LBULS   EQU     $4FDA           ; CT5
RBULS   EQU     $4FDB           ; CT6
RFIELD  EQU     $4FA1
LSCORE  EQU     $4FA2
LFIELD  EQU     $4FA5
RSCORE  EQU     $4FA6

  ; ***************
  ; * SUBROUTINES *
  ; ***************

  ; ****** DISPLAY CLOCK AND UPDATE CT4
DCLOCK:
        DI
        SYSSUK  DECCTS          ; UPI DECrement CT'S under mask
        DB      $80             ; ... Counters = 10000000B
        ;
        LD      IX,FNTSML
        LD      A,(CT7)
        OR      A
        JR      Z,DCOUT                 ; (+$08)
        SYSSUK  DISNUM          ; UPI DISplay NUMber
        DB      STMRX           ; ... X = 76
        DB      BSY             ; ... Y = 2
        DB      TIME            ; ... Options (color 2 on color 3)
        DB      01000010B       ; ... Extended (for 2 small digits)
        DW      CT7             ; ... Number Address
        ;
DCOUT:  XOR     A
        OUT     (MAGIC),A
        LD      (WASTE),A
        EI
        RET

  ; ****** FIRE BULLETS
  ; *** LEFT COWBOY
FIRE0:
        SYSSUK  SUCK            ; UPI SUCK inline args into context block
        DB      11011100B       ; ... Mask (for IX, BC, and HL) 
        DW      LCOWB           ; ... IX (at Left COWboy array)
        DW      LBULS           ; ... BC (point at CT5)
        DW      BULV1 + 1       ; ... HL (BULlet Vector #1 + 1)
        ;
        JR      ZORE                    ; (+$09)

  ; *** RIGHT COWBOY
FIRE1:
        SYSSUK  SUCK            ; UPI SUCK inline args into context block
        DB      11011100B       ; ... Mask (for IX, BC, and HL) 
        DW      RCOWB           ; ... IX (at Right COWboy array)
        DW      RBULS           ; ... BC (point at CT6)
        DW      BULV3 + 1       ; ... HL (BULlet Vector #3 + 1)
        ;
ZORE:   LD      A,(IY+CBB)
        OR      A
        RET     Z
        LD      A,(BC)          ; GET BULLET COUNT
        OR      A
        RET     Z
        LD      A,(HL)          ; CHECK IF BULLET IS AVAILABLE
        OR      A
        JR      Z,ZOK                   ; (+$09)
        LD      DE,BULVSZ       ; DELTA TO NEXT BULLET
        ADD     HL,DE
        LD      A,(HL)
        OR      A
        JR      Z,ZOK                   ; (+$01)
        RET

  ; *** NOW HL -> BULLET
  ;     IX -> COWBOY
  ; SUB 1 FROM BULLET COUNT
ZOK:
        LD      A,(BC)
        DEC     A
        LD      (BC),A
  ; SET SUB TIMER IF OUT OF BULLETS
        JR      NZ,BERASE               ; (+$0d)
        LD      A,(CT7)
        OR      A
        LD      A,$10
        JR      Z,STSEC                 ; (+$02)
        LD      A,$02
STSEC:  LD      (CT7),A
BERASE: PUSH    HL
        PUSH    IX
        LD      A,(BC)
        LD      L,A
        LD      H,$00
        ADD     HL,HL           ; times 4
        ADD     HL,HL
        LD      DE,(BSY SHL 8) + RBULX
        BIT     MRFLOP,(IX+VBMR)
        LD      A,$40           ; FLOPPED MR
        JR      Z,RITB                  ; (+$01)
        XOR     A               ; NORMAL MR
  ; NOW POSITION AND ERASE
RITB:   ADD     HL,DE
        EX      DE,HL
        SYSTEM  RELAB1          ; UPI RELative to non-magic ABSolute
        ;
        EX      DE,HL
        LD      B,$05
        LD      DE,$0028        ; INC TO NEXT LINE
BELP:   LD      (HL),$FF        ; ERASE LINE
        ADD     HL,DE           ; GO DOWN A LINE
        DJNZ    BELP                    ; (-$05)
        LD      D,$00
        LD      E,(IX+VBARM)    ; GET CURRENT ARM POS
        LD      H,D
        LD      L,E
        ADD     HL,HL           ; times 2
        ADD     HL,DE           ; times 3
        LD      DE,BULTAB
        ADD     HL,DE           ; -> BULTAB(ARM)
        EX      DE,HL
        POP     BC              ; BC <== IX
        POP     HL              ; BUL [STAT]
        PUSH    HL              ; SAVE FOR ACTIVATE
        INC     HL              ; BUL [DEL TIME]
        LD      (HL),$01        ; MAKE BULIT JUMP OUT
        INC     HL              ; BUL [DEL XLOW]
        INC     BC              ; COW [STAT]
        INC     BC              ; COW [DEL TIME]
        INC     BC              ; COW [DX LO]
        CALL    PUTVEC
        INC     BC              ; COW [XCHK]
        INC     BC              ; COW [DY LO]
        INC     HL              ; BUL [XCHK]
        LD      (HL),$01        ; LIMIT CHECK
        INC     HL              ; BUL [DY LO]
        CALL    PUTVEC
        POP     HL              ; BUL [STAT]
        LD      (HL),$80        ; ACTIVE
        SYSSUK  BMUSIC          ; UPI Begin playing MUSIC
        DW      MSTACK          ; ... Music Stack
        DB      $01             ; ... Voices = JUST NOISE
        DW      GUNSHOT         ; ... Score Address
        ;
        RET

  ; ****** TAKE A COFFEE BREAK macro-routine
NBRK:
        DONT  PIZBRK            ; SEE IF I CARE
        ;
        DO    MRET
        ;

  ; *** CONVERT JOYSTICKS
JOY0:
        LD      IX,LCOWB
        JR      PJOY                    ; (+$04)

JOY1:
        LD      IX,RCOWB
  ;   CONVERT   JOYSTICKS
PJOY:   LD      C,(IX+VBMR)
        LD      DE,$0080        ; for 0.5 pixels per X step
        LD      HL,$0080        ; for 0.5 pix per Y step
        SYSTEM  MSKTD           ; UPI MaSK joystick in B To Deltas
        ;
        LD      (IX+VBDYH),H    ; Save DELTAS
        LD      (IX+VBDYL),L
        LD      (IX+VBDXH),D
        LD      (IX+VBDXL),E
        RET

PPOT1:
        LD      IX,RCOWB
        LD      A,B             ; POT MUST BE FLOPPED BECAUSE
        CPL                     ; ARM IS FLOPPED
        JR      PPOT                    ; (+$05)

PPOT0:
        LD      IX,LCOWB
        LD      A,B
  ; CONVERT POT AND STORE
PPOT:   AND     $E0
        RRCA
        RRCA
        RRCA
        RRCA
        CP      $0E
        JR      NZ,KART                 ; (+$02)
        LD      A,$0C           ; IF KNOB = 7 THEN SET TO 6
KART:   LD      (IX+VBARM),A    ; SET ARM POSITION
        RET

  ; *** CHECK IF BULLET HIT ANYTHING
HITCHK:
        LD      A,(IX+VBSTAT)
        AND     $60
        CP      $20             ; CHECK ONLY IF BLANKED
        JR      Z,HIT                   ; (+$0f)
        RET     NC              ; RETURN IF NOT BLANKED YET
        BIT     VBCLAT,(IX+VBXCHK)
        RET     Z
        LD      (IX+VBSTAT),$00 ; BULLET HIT WALL
        LD      (IX+VBXCHK),$01 ; SET LIMIT CHECK
        RET

HIT:
        LD      A,(IX+VBXH)     ; CHECK WHAT PART OF SCR ITS ON
        CP      WAGX
        JR      NC,HIT1                 ; (+$0e)
        LD      (IX+VBTIMB),$02 ; MAKE IT JUMP OUT
        LD      (IX+VBSTAT),$80 ; RE ACTIVATE
        LD      HL,BULLMT
        SYSTEM  VECT            ; UPI VECTor move coordinate pair
        ;
        RET

HIT1:
        LD      (IX+VBSTAT),$00 ; BULLET DIES FROM WAGON ON
        CP      RCACX
        JR      NC,HIT2                 ; (+$1d)
        LD      A,(WAGON)
        OR      A               ; IS IT A CACTII?
        RET     NZ              ; NOPE ITS A WAGON
        LD      E,CCACX         ; LOAD X
  ; ERASE OBJECT BULLET HITS
ERASE_: LD      D,(IX+VBYH)     ; LOAD Y
        DEC     D
        SYSSUK  RELAB1          ; UPI RELative to non-magic ABSolute
        DB      $00             ; ... Magic = 0
        ;
        EX      DE,HL
        LD      DE,$FFD7        ; minus 41
        LD      B,$00
ELOP:   LD      A,(HL)
        LD      (HL),B          ; ZERO THE SCREEN BYTE
        INC     HL
        OR      (HL)
        LD      (HL),B
        ADD     HL,DE
        JR      NZ,ELOP                 ; (-$08)
        RET

HIT2:
        CP      RCACX+8         ; GUNFTR SPACE
        JR      NC,DIE                  ; (+$0c)
        LD      E,LCACX
        BIT     MRFLOP,(IX+VBMR)
        JR      NZ,ERASE_               ; (-$22)
        LD      E,RCACX
        JR      ERASE_                  ; (-$26)

DIE:
        BIT     MRFLOP,(IX+VBMR) ; WHO DIED?
        JR      Z,DLEFT                 ; (+$0c)
        SYSSUK  SUCK            ; UPI SUCK inline args into context block
        DB      11011101B       ; ... Mask (IX, D, BC, and HL)
        DW      LCOWB           ; ... IX (at Left COWboy array)
        DB      $08             ; ... D = 8
        DW      TAPS            ; ... BC (point at music score)
        DW      RSCORE          ; ... HL (point at Right SCORE)
        ;
        JR      DIE1                    ; (+$0a)

DLEFT:  SYSSUK  SUCK            ; UPI SUCK inline args into context block
        DB      11011101B       ; ... Mask (IX, D, BC, and HL)
        DW      RCOWB           ; ... IX (at Right COWboy array)
        DB      $64             ; ... D = 100
        DW      FUNERL          ; ... BC (point at music score)
        DW      LSCORE          ; ... HL (point at Left SCORE)
        ;
DIE1:   LD      (IX+VBLEGT),$06 ; SET FIRST CELL TIME
        LD      (IX+VBLEG),KIL1 AND $FF ; ??
        LD      (IX+VBSTAT),$68 ; KILL HIM
        LD      A,(IX+VBYH)     ; WHERE TO WRITE GOT ME
        SUB     $08
        CP      TLINE + 9
        JR      NC,DIE4                 ; (+$02)
        ADD     A,$20
DIE4:   LD      D,A
        SYSTEM  INCSCR          ; UPI INCrement SCoRe
        ;
        DEC     HL
        LD      A,(HL)          ; FIELD
        CP      $05
        ADC     A,$00
        LD      (HL),A
  ; PLAY DEATH SONG
        LD      H,B
        LD      L,C
        LD      IX,MSTACK
        LD      A,11000000B     ; for TONe A only
        SYSTEM  BMUSIC          ; UPI Begin playing MUSIC
        ;
        LD      C,LARG2
        LD      HL,GOTME
        DI
        SYSTEM  STRDIS          ; UPI STRing DISplay (in color 3)
        ;
        SYSSUK  PAWS            ; UPI PAUSE
        DB      $FA             ; ... Interrupts = 250
        ;
        LD      A,$01
        LD      (SEMI4S),A      ; SET FLAG0
        RET

  ; *** FIELD PUTS UP THE SACTII APPROP TO SCORE
  ; A = SCORE OF OPP PLAYER UP TO 6
  ; BC -> ARRAY OF Y POSITIONS
FIELD:
        LD      HL,CACTUS       ; -> CACTUS PATTERN
        PUSH    AF
        LD      A,00001000B
        OUT     (XPAND),A
        POP     AF
        CP      $01
        RET     C
        CP      $04
        JR      NC,TCAC                 ; (+$03)
        CALL    CACW
TCAC:   INC     BC
        CP      $02
        RET     C
        CP      $05
        JR      NC,MCAC                 ; (+$03)
        CALL    CACW
MCAC:   CP      $03
        RET     C
        INC     BC
        EX      AF,AF'
        LD      A,$81           ; ACIVATE WAGON
        LD      (WAGON),A
        EX      AF,AF'
        CALL    CACW
        CP      $04
        RET     C
        INC     BC
        LD      HL,TREE
        PUSH    AF
        LD      A,00001100B
        OUT     (XPAND),A
        POP     AF
        CALL    CACW
        CP      $05
        RET     C
        INC     BC
CACW:   PUSH    AF
        PUSH    DE
        LD      A,(BC)
        LD      D,A
        LD      A,$08           ; EXPAND
        SYSTEM  WRITP           ; UPI WRITe with Pattern size lookup
        ;
        POP     DE
        POP     AF
        RET

  ; *** PUT DEL X,Y INTO BULLET VECTORS
PUTVEC:
        LD      A,(DE)          ; TABLE [D LO]
        LD      (HL),A          ; BUL [D LO]
        INC     DE              ; TAB [D HI]
        INC     BC              ; COW [D HI]
        INC     HL              ; BUL [D HI]
        LD      A,(DE)
        LD      (HL),A
        INC     HL              ; BUL [LO]
        INC     DE              ; TAB [HI]
        INC     BC              ; COW [LO]
        LD      (HL),$00
        INC     BC              ; COW [HI]
        INC     HL              ; BUL [HI]
        LD      A,(BC)
        EX      DE,HL
        ADD     A,(HL)
        EX      DE,HL
        LD      (HL),A          ; BUL [HI] = COW [HI] + TAB [HI]
        INC     DE              ; TAB [D HI]
        RET

  ; ****** GUNFIGHT START UP ROUTINE (ONCE PER GAME)
INIT:
        SYSSUK  GETPAR          ; UPI GET game PARameter from user
        DW      MXSCR           ; ... Prompt Address for "MaX SCoRe"
if BALLY or ASTRO ; ### conditional ###
        DB      10000100B       ; # + # ... Digits = 4 (ZERO SUPPress)
else ; WHITE ####### conditional ###### else...
        DB      10000010B       ; # - # ... Digits = 2 (ZERO SUPPress)
endif ; ##### end conditional #########
        DW      ENDSCR          ; ... Parameter Address
        ;
        LD      SP,STACK
        SYSTEM  INTPC           ; UPI INTerPret with Context create
        ;
        DO      FILL            ; UPI FILL memory with data
        DW      STACK           ; ... Memory Address
        DW      CT7 - STACK     ; ... Byte Count = 214
        DB      $00             ; ... Data = 0
        ;
        DO      SETB            ; UPI SET Byte
        DB      2 * GSBSCR      ; ... Data = 2
        DW      GAMSTB          ; ... Memory Address
        ;
        DO      SETOUT          ; UPI SET UP game OUTput ports
        DB      BLINE * 2       ; ... VERBL*2 = BOTTOM LINE - VRT BLK
        DB      $D6             ; ... HORCB/4 = HORZ BOUNDS
        DB      $08             ; ... INMOD = 8 (screen only)
        ;
        DO      COLSET          ; UPI COLors SET
        DW      GFCOLS          ; ... Table Address (GunFight COLorS)
        ;
        DO      BMUSIC          ; UPI Begin MUSIC "STREETS OF LOREDO"
        DW      MSTACK          ; ... Music Stack
        DB      $C0             ; ... Voices = 11000000B (ON VOICE A)
        DW      HOME            ; ... Score Address (HOME on the range)
        ;
        DONT    XINTC           ; UPI eXit INTerpreter with Context
        ;

  ; *****
  ; ONCE A ROUND START UP ROUTINE
  ; *****
STRND:
        DI
        SYSTEM  INTPC           ; UPI INTerPret with Context create
        ;
  ; INIT HANDLES, BULLETS, TIMERS
        DO      MOVE            ; UPI MOVE block transfer
        DW      CT5             ; ... Destination Address
        DW      $000C           ; ... Byte Count = 12
        DW      SINIT           ; ... Source Address
        ;
  ; COLOR BANNER
        FILLq NORMEM, BYTEPL*ALINE, $FF
        ;
  ; ERASE SCREEN
        FILLq NORMEM+BYTEPL*ALINE, BYTEPL*(BLINE-ALINE), $00
        ;
  ; RESET VECTORS
        FILLq STRRAM, ENDRAM-STRRAM, $00
        ;
  ; SHOW SCORES
        DO      SUCK            ; UPI SUCK inline args into context block
        DB      00010000B       ; ... Mask (IX only)
        DW      FNTSML          ; ... IX = $020D
        ;
        DO      DISNUM          ; UPI DISplay NUMber (4 digits)
        DB      LNX             ; ... X = 8
        DB      BSY             ; ... Y = 2
        DB      TIME            ; ... Options (color 2 on color 3)
        DB      11000100B       ; ... Extended (ZERO SUPRS, SMALL)
        DW      LSCORE          ; ... Number Address (Left SCORE)
        ;
        DO      DISNUM          ; UPI DISplay NUMber (4 digits)
        DB      RNX             ; ... X = 136
        DB      BSY             ; ... Y = 2
        DB      TIME            ; ... Options (color 2 on color 3)
        DB      11000100B       ; ... Extended (ZERO SUPRS, SMALL)
        DW      RSCORE          ; ... Number Address (Right SCORE)
        ;
  ; CHECK FOR END GAME
        DO      RCALL           ; UPI Real CALL asm language subroutine
        DW      ENDGAM          ; ... Real Address
        ;
  ; STRDIS+1, "GET ReaDY", X = 44, Y = 1, Options = 00001011B
        TEXTD GETRDY, GRX, GRY, LARGE ; (in color 2 on color 3)
        ;
        DONT    XINTC           ; UPI eXit INTerpreter with Context
        ;
        XOR     A               ; SET UP WAGON
        LD      (WAGON),A       ; STOP WAGON
  ; PUT UP PLAY FIELD
        LD      A,(RFIELD)      ; NUMBER OF CACTII
        LD      E,RCACX         ; RIGHT CAC COLUMN
        LD      BC,RFTAB        ; POSITIONS OF TABLE OF CACTII
        CALL    FIELD           ; PUT THE CACTII UP
        LD      A,(LFIELD)
        LD      E,LCACX
        LD      BC,LFTAB
        CALL    FIELD
  ; INITIALIZE Q POINTERS
INITQ:  LD      A,LCOWB SHR 8
        LD      (WRITQ+2),A
        LD      (VECQ+2),A
  ; SETUP VECTORS SO COWBOYS WALK OUT
        LD      IX,LCOWB        ; LEFT COWBOY VECTOR
        LD      (IX+VBMR),$10
        LD      HL,VECQ
        CALL    COWINT
        LD      IX,RCOWB        ; RIGHT COWBOY VECTOR
        LD      (IX+VBMR),$50
        CALL    COWINT
        LD      A,(WAGON)       ; IF WAGON IS ON
        OR      A
        JR      Z,MIDC                  ; (+$1d)
        LD      IX,WAGVEC       ; THEN ACTIVATE WAGON
        LD      (IX+VBMR),$10
        LD      (IX+VBYCHK),$03
        LD      (IX+VBDYL),$40
        LD      (IX+VBXH),$48
        LD      (IX+VBYH),TLINE
        CALL    ADDTQ
        JR      BORG                    ; (+$0b)

  ; ELSE PUT UP A CACTUS
MIDC:
        LD      A,$08
        OUT     (XPAND),A
        SYSSUK  WRITP           ; UPI WRITe with Pattern size lookup
        DB      CCACX           ; ... X = 76
        DB      MCACY           ; ... Y = 42
        DB      $08             ; ... Magic (EXPAND on)
        DW      CACTUS          ; ... Pattern Address
        ;
  ; INITIALIZE BULLET VECTORS
BORG:   LD      DE,BULVSZ
        LD      IX,BULV1
        LD      BC,($04 SHL 8) + $20
        LD      A,$02
BULLP:  CP      B
        JR      NZ,TIYU                 ; (+$02)
        LD      C,$60
TIYU:   LD      (IX+VBMR),C
        LD      (IX+VBXCHK),$01
        LD      (IX+VBYCHK),$03
        ADD     IX,DE
        DJNZ    BULLP                   ; (-$14)
  ; FIRE UP INTERRUPTS (BALLY standard IM 2 is already DONE IN MENU)
        LD      A,INTTBL SHR 8
        LD      I,A
        LD      A,LFRVEC AND 0FFH
        OUT     (INFBK),A
  ; ***
  ; LET COWBOYS WALK OUT
  ; ***
WALK:
        SYSSUK  PAWS            ; UPI PAUSE
        DB      $64             ; ... Interrupts = 100
        ;
        DI
        LD      IX,FNTSML
        SYSTEM  INTPC           ; UPI INTerPret with Context create
        ;
  ; ERASE GET READY
        DO      BLANK           ; UPI BLANK area
        DB      $12             ; ... Width = 18 bytes
        DB      $08             ; ... Height = 8 lines
        DB      $FF             ; ... Data = 255
        DW      $4033           ; ... Screen Address (NORMEM + 1*BYTEPL + 11)
        ;
  ; STRDIS+1, "DRAW", X = 64, Y = 1, Options = 00001011B
        TEXTD DRAW, DRX, GRY, LARGE ; (in color 2 on color 3)
        ;
        DO      CHRDIS          ; UPI CHaRacter DISplay
        DB      LBULX           ; ... X = 32
        DB      BSY             ; ... Y = 2
        DB      BULT            ; ... Options (color 2 on color 3)
        DB      $BB             ; ... Character (BULLET)
        ;
        DO      MCALL           ; UPI Macro CALL interpreter subroutine
        DW      BULRIT          ; ... Macro Address (for 5 MORE)
        ;
        DO      SUCK            ; UPI SUCK inline args into context block
        DB      00000001B       ; ... Mask (E only)
        DB      RBULX           ; ... X = 104 (to DO THE RIGHT ONES)
        ;
  ; DISPLAY THE FIRST ONE
        DONT    CHRDIS          ; UPI CHaRacter DISplay
        ;
        DO      MCALL           ; UPI Macro CALL interpreter subroutine
        DW      BULRIT          ; ... Macro Address (to DISPLAY THR OTHER FIVE)
        ;
        DO      PAWS            ; UPI PAUSE
        DB      $3C             ; ... Interrupts = 60
        ;
        DO      BLANK           ; UPI BLANK area to Color 3
        DB      $08             ; ... Width = 8 bytes
        DB      $08             ; ... Height = 8 lines
        DB      $FF             ; ... Data = 255 TO BLANK TO
        DW      $4038           ; ... Screen Address (NORMEM + 1*BYTEPL + 16)
        ;
        DONT    XINTC           ; UPI eXit INTerpreter with Context
        ;
  ; *****
  ; MAIN LOOP DURING ROUND
  ; GETS HANDLES, SETS VECTORS AND CHECKS BULLETS
LOOP:   SYSTEM  INTPC           ; UPI INTerPret with Context create
        ;
        DO      SENTRY          ; UPI SENse TRansition Y
        DW      ALKEYS          ; ... Keymask Address (ALL KEYS)
        ;
        DO      DOIT            ; UPI DOIT table, branch to translation handler
        DW      DTAB            ; ... Table Address (Doit TABle)
        ;
        DONT    XINTC           ; UPI eXit INTerpreter with Context
        ;
  ; CHECK FOR DEATHS
; DEATH:
        LD      IX,BULV1
        LD      DE,BULVSZ
        LD      B,$04
LPPP2:  PUSH    BC
        PUSH    DE
        CALL    HITCHK
        POP     DE
        POP     BC
        ADD     IX,DE
        LD      A,(SEMI4S)      ; CHECK IF DEATH MODE
        DEC     A
        JR      Z,LOOP                  ; (-$21)
        DJNZ    LPPP2                   ; (-$11)
        JR      LOOP                    ; (-$25)

  ; *** round done
ENDRND:
        DONT    XINTC           ; UPI eXit INTerpreter with Context
        JP      STRND

  ; *** all done
ENDGAM:
        LD      A,(GAMSTB)
        BIT     GSBEND,A
        RET     Z
        SYSTEM  QUIT            ; UPI QUIT cassette execution
        ;

  ; ****** Doit TABle
DTAB    JMPd     SCT7, ENDRND, $00
        JMPd     SF0, ENDRND, $00
        RC       SP0, PPOT0, $00
        RC       SP1, PPOT1, $00
        RC       SJ0, JOY0, $00
        RC       SJ1, JOY1, $00
        MC       SKYD, NBRK, $00
        RC       ST0, FIRE0, $00
        RC       ST1, FIRE1, $00
        RC       SSEC, DCLOCK, ENDx

BULRIT:
        DONT     CHRDIS
        DONT     CHRDIS
        DONT     CHRDIS
        DONT     CHRDIS
        DONT     CHRDIS
        DONT     MRET

  ; ************************************
  ; * GUNFIGHT WRITE INTERRUPT ROUTINE *
  ; ************************************
GFWRIT:
        EX      AF,AF'
        EXX
        PUSH    IX
BEGINT: LD      A,LFRVEC AND $FF ; ESTABLISH TICKS INT
        OUT     (INFBK),A
        LD      A,LFRLIN
        OUT     (INLIN),A
        LD      HL,WRITQ        ; GET FIRST WRITE Q ENTRY
        CALL    FIRST
        CALL    DELQ            ; DROP FROM WRITE Q
        XOR     A
        LD      (WASTE),A
        BIT     VBSWAG,(IX+VBSTAT) ; WAGON?
        JR      NZ,GFWRT1               ; (+$28) JUMP IF YES
  ; GUNFIGHTER - BLANKETH HIM
        LD      DE,$1405        ; LOAD BLANKING PARMS
        SYSTEM  VBLANK          ; UPI Vector BLANK area
        ;
  ; AND WRITE LEG
        LD      H,LEG0 SHR 8    ; point at LEG PATTERN
        LD      L,(IX+VBLEG)
        INC     L               ; SKIP OVER LINK AND TIME
        INC     L
        SYSTEM  VWRITR          ; UPI Vector WRITe Relative
        ;
  ; IS GUNFIGHTER DEAD?
        BIT     VBSINT,(IX+VBSTAT)
        JR      NZ,GFWRT5               ; (+$30) JUMP IF SO
        LD      HL,ARMTBL       ; LOOKUP ARM PATTERN
        LD      D,$00
        LD      E,(IX+VBARM)
        ADD     HL,DE
        LD      E,(HL)
        INC     HL
        LD      D,(HL)
        EX      DE,HL
  ; WRITE ARM PATTERN
        SYSTEM  VWRITR          ; UPI Vector WRITe Relative
        ;
        LD      HL,GFBODY       ; LOAD BODY PATTERN
        JR      GFWRT2                  ; (+$08)

  ; *** BLANK THE WAGON
GFWRT1:
        LD      DE,$1604        ; LOAD WAGON SIZE
        SYSTEM  VBLANK          ; UPI Vector BLANK area
        ;
        LD      HL,WAGPAT
GFWRT2: SYSTEM  VWRITR          ; UPI Vector WRITe Relative
        ;
GFWRT4: LD      (IX+VBOAH),D
        LD      (IX+VBOAL),E
GFWRT3: LD      HL,VECQ         ; ADD VECTOR TO VECTOR Q
        CALL    ADDTQ
        POP     IX
        EX      AF,AF'
        EXX
EIRE:   EI
        RET

GFWRT5:
        LD      HL,NULPAT
        JR      GFWRT2                  ; (-$19)

  ; ***********************************
  ; * GUNFIGHT LOW FOREGROUND ROUTINE *
  ; ***********************************
GFLFR:
        PUSH    AF
        PUSH    BC
        PUSH    DE
        PUSH    HL
        PUSH    IX
  ; BUMP TIME BASES OF ACTIVE OR INTERCEPTED VECTORS
        LD      HL,BULV1 + VBSTAT
        LD      DE,BULVSZ - 1
        LD      B,$04
        CALL    TBUMP
        INC     HL              ; SKIP LINK FIELD
        LD      DE,GFVSIZ - 1
        LD      B,$03
        CALL    TBUMP
  ; LOOP TO UNWRITE, THE WRITE ALL FOUR BULLETS
  ; BUT FIRST, A WORD TO OUR SHIFTER
        XOR     A
        LD      (WASTE),A
        LD      B,$04
        LD      IX,BULV1
  ; UNWRITE THIS GUY?
WRBUL1: BIT     6,(IX+VBSTAT)
        JR      Z,WRBUL2                ; (+$11) JUMP IF NOT
        LD      H,(IX+VBOAH)
        LD      L,(IX+VBOAL)
        LD      A,(IX+VBARM)    ; GET LAST MR
        OUT     (MAGIC),A
        LD      (HL),$C0        ; UNWRITE BULLET
        RES     VBBLNK,(IX+VBSTAT) ; CLEAR BLANK BIT
  ; SHALL WE WRITE THIS GUY?
WRBUL2: BIT     VBSACT,(IX+VBSTAT)
        JR      Z,WRBUL4                ; (+$2b)
        LD      D,(IX+VBYH)
        LD      E,(IX+VBXH)
        LD      A,(IX+VBMR)
        SYSTEM  RELABS          ; UPI RELative to ABSolute conversion
        ;
        LD      (IX+VBOAH),D
        LD      (IX+VBOAL),E
        LD      (IX+VBARM),A
        LD      HL,NORMEM
        ADD     HL,DE
        LD      A,(HL)
        EX      DE,HL
        LD      (HL),$C0
        OR      A
        JR      Z,WRBUL3                ; (+$08) JUMP IF NOT
        RES     VBSACT,(IX+VBSTAT) ; KILL ACTIVE BIT
        SET     VBSINT,(IX+VBSTAT) ; SET INTERCEPT LINE
WRBUL3: SET     VBBLNK,(IX+VBSTAT) ; SET BLANK BIT
  ; STEP TO NEXT BULLET VECTOR, LOOP BACK IF DONE
WRBUL4: LD      DE,BULVSZ
        ADD     IX,DE
        DJNZ    WRBUL1                  ; (-$4f)
  ; GET NEXT PATTERN TO WRITE, AND SCHEDULE HIM
        LD      HL,WRITQ
        CALL    FIRST
        JR      Z,WRBL5A                ; (+$12) JUMP IF EMPTY Q
        LD      A,WRTVEC AND $FF ; SET FEEDBACK REG
        OUT     (INFBK),A
        LD      A,(IX+VBYH)     ; WHICH WINDOW TO USE?
        CP      WINBND          ; COMPARE TO WINDOW BOUNDRY
        LD      A,BOTLIN        ; ASSUME BOTTOM LINE
        JR      NC,WRBUL5               ; (+$02) JUMP IF GOOD GUESS
        LD      A,TOPLIN        ; WRONG - USE TOP
WRBUL5: OUT     (INLIN),A       ; SET LINE REGISTER
        EI
  ; LOOP THRU VECTORING THOSE BULLETS
WRBL5A: LD      IX,BULV1
        LD      B,$04
        LD      HL,BULLMT       ; HL = BULLET LIMITS TABLE
        LD      DE,BULVSZ
WRBUL6: BIT     VBSACT,(IX+VBSTAT) ; ACTIVE BULLET?
        JR      Z,WRBUL7                ; (+$0c)
        SYSTEM  VECT            ; UPI VECTor move coordinate pair
        ;
        BIT     VBCLAT,(IX+VBXCHK) ; DID Y HIT EDGE?
        JR      Z,WRBUL7                ; (+$04) NOPE
        RES     VBSACT,(IX+VBSTAT) ; DEACTIVATE BULLET
WRBUL7: ADD     IX,DE
        DJNZ    WRBUL6                  ; (-$16) LOOP BACK
  ; NOW PUT SOMETHING ON THE WRITE Q
        LD      B,$02           ; MAX 2 TIMES THRU
        LD      HL,VECQ
GVECT:  CALL    FIRST           ; FIRST VECTOR Q ENTRY
        JP      Z,GVECT4        ; JUMP IS Q EMPTY
        CALL    DELQ            ; DROP FROM VECTOR Q
        EI
  ; WAGON?
        BIT     VBSWAG,(IX+VBSTAT)
        JP      NZ,GVECT5       ; JUMP ON WAGON
  ; DEAD?
        BIT     VBSINT,(IX+VBSTAT)
        JR      NZ,GVECT1               ; (+$25)
  ; ZERO VELOCITY?
        LD      A,(IX+VBDXL)
        OR      (IX+VBDXH)
        OR      (IX+VBDYL)
        OR      (IX+VBDYH)
        JR      NZ,GVECT1               ; (+$17) GVECT1 IF NONZERO
        LD      (IX+VBTIMB),A   ; ZERO TIME BASE
        BIT     VBSNOM,(IX+VBSTAT) ; ALREADY STATIONARY
        JR      NZ,GVEC3A               ; (+$36)
  ; SET STATIONARY LEGS
        LD      (IX+VBLEG),LEG0 AND 0FFH
        SET     VBSCHG,(IX+VBSTAT) ; SET CHANGED
        SET     VBSNOM,(IX+VBSTAT) ; AND STATIONARY
        JR      GVEC3A                  ; (+$28) JUMP TO ARM CHECK
  ; MOVING GUNFIGHTER VECTOR
GVECT1: LD      HL,GUNLMT       ; LOAD GF LIMITS
        SYSTEM  VECT            ; UPI VECTor move coordinate pair
        ;
        JR      Z,GVECT2                ; (+$08)
        SET     VBSCHG,(IX+VBSTAT) ; SET CHANGED BIT
        RES     VBSNOM,(IX+VBSTAT) ; CLEAR NOT MOVING STATUS
  ; NEED WE GO TO NEXT CELL IN ANIMATION SEQUENCE?
GVECT2: LD      A,(IX+VBLEGT)   ; A = ANIMATION TIMER
        SUB     C               ; SUBTRACT TIME BASE
        JP      P,GVECT3        ; JUMP IF NOT COUNTED DOWN
  ; GET NEXT CELL
        LD      E,(IX+VBLEG)    ; GET LINK
        LD      D,LEG0 SHR 8    ; SET H.0. PART
        LD      A,(DE)          ; A = NEXT
        LD      (IX+VBLEG),A
        INC     DE              ; STEP TO TIMER
        LD      A,(DE)          ; GET NEW TIMER
        SET     VBSCHG,(IX+VBSTAT) ; SET CHANGED BIT
GVECT3: LD      (IX+VBLEGT),A   ; STORE BACK TIMER
  ; DID ARM CHANGE?
GVEC3A: LD      A,(IX+VBARM)
        CP      (IX+VBOARM)     ; COMPARE TO OLD ARM
        JR      Z,GVEC3B                ; (+$07) JUMP IF NO CHANGE
        SET     VBSCHG,(IX+VBSTAT) ; SET CHANGED BIT
        LD      (IX+VBOARM),A
  ; ADD ITEM TO WRITE Q?
GVEC3B: BIT     VBSCHG,(IX+VBSTAT)
        JR      NZ,GVECT6               ; (+$20) YES GVECT6
  ; NO CHANGE - LINK TO VECTOR Q
        LD      HL,VECQ
        CALL    ADDTQ
        DEC     B
        JP      NZ,GVECT        ; SUB FOR DJNZ
GVECT4: EI
        CALL    STIMER
        POP     IX
        POP     HL
        POP     DE
        POP     BC
        POP     AF
        RET

  ; *** VECTOR AND Q WAGON
GVECT5:
        LD      HL,WAGLMT
        SYSTEM  VECT            ; UPI VECTor move coordinate pair
        ;
        LD      HL,VECQ
        CALL    DELQ            ; REMOVE FROM VECTOR Q
GVECT6: RES     VBSCHG,(IX+VBSTAT)
        LD      HL,WRITQ
        CALL    ADDTQ
        JR      GVECT4                  ; (-$22) JUMP BACK TO QUIT

  ; *** ROUTINE TO BUMP TIME BASE OF VECTORS
TBUMP:  LD      A,(HL)          ; GET STATUS
        INC     HL
        AND     $A0             ; ACTIVE OR INTERCEPTED?
        JR      Z,TBUMP1                ; (+$01) NO - TBUMP1
        INC     (HL)            ; BUMP THE TIME BASE
TBUMP1: ADD     HL,DE
        DJNZ    TBUMP                   ; (-$0a)
        RET

  ; *** SUBROUTINE TO DELETE ENTRY AT FRONT OF Q
  ; Entry:        HL = HEAD-TAIL, IX = OBJECT, A = CLOBBEREd
DELQ:   DI
        LD      A,(IX+NEXT)     ; HEAD = NEXT(OBJECT)
        LD      (HL),A
        AND     A               ; IS HEAD NOW NIL?
        RET     NZ              ; QUIT IF NOT
        INC     HL              ; YES - SET TAIL = NIL TOO
        LD      (HL),A
        DEC     HL
        RET

  ; ****** COWboy INiTialize routine
COWINT:
        LD      (IX+VBDXL),$32  ; SLOW WALK OUT
        LD      (IX+VBSTAT),$80 ; ACTIVE
        LD      (IX+VBXCHK),$01
        LD      (IX+VBYCHK),$01
        LD      (IX+VBXH),$04
        LD      (IX+VBYH),$28
        LD      (IX+VBARM),$06  ; SET ARM STRAIGHT
        LD      (IX+VBLEG),LEG0 AND $FF
  ; *** SUBROUTINE TO APPEND ENTRY TO END OF Q
  ; Entry:        HL = HEAD-TAIL BYTES, IX = OBJECT, A, DE, C
ADDTQ:  PUSH    IX              ; DE = ENTRY
        POP     DE
        DI
        LD      (IX+NEXT),$00   ; NEXT(OBJ) = NIL
        INC     HL
        LD      A,(HL)          ; A = OLD TAIL
        LD      (HL),E          ; SET TAIL = .OBJ
        AND     A               ; WAS OLD TAIL NIL?
        JR      Z,ADDTQ1                ; (+$06) JUMP IF SO
  ; NONNIL OLD TAIL  SET NEXT(OLDTAIL) = .OBJ
        LD      E,A             ; DE = .NEXT(OLDTAIL)
        LD      A,(HL)          ; A = OBJ (FROM NEW TAIL)
        DEC     HL
        DEC     DE
        LD      (DE),A
        RET

  ; NIL OLD TAIL CASE
ADDTQ1:
        DEC     HL              ; BACKUP TO HEAD
        LD      (HL),E          ; HEAD = .OBJ
        RET

  ; *** SUBROUTINE TO POINT IX AT FIRST ENTRY ON A Q
  ; Entry:        HL = Q HEAD-TAIL
  ; Exit:         IX, DE = OBJECT, A = L.O. BYTE OF OBJECT
  ;               NONZERO STATUS SET IF Q NOT EMPTY
FIRST:
        DI
        LD      E,(HL)
        INC     HL
        INC     HL
        LD      D,(HL)          ; D = H.O. ADDR. BYTE
        DEC     HL
        DEC     HL
        LD      A,E             ; E = HEAD OF Q
        AND     A
        PUSH    DE
        POP     IX
        RET

  ; **********************
  ; * GUNFIGHT CONSTANTS *
  ; **********************

INTTBL:         ; INTerrupt TaBLe
LFRVEC: DW      GFLFR           ; Low Foreground Routine VECtor

WRTVEC: DW      GFWRIT          ; GunFighter WRiTe VECtor

  ; ****** WAGON LIMITS TABLE
WAGLMT:
        DB      TLINE
        DB      BLINE - $18

  ; ****** "GET ReaDY" text string
GETRDY:
        DB      'G','E','T',$20,'R','E','A','D','Y'
  ; terminated by next byte below ...

  ; ****** GUNFIGHTER LIMITS
GUNLMT:
        DB      $00
        DB      LCACX - $11
        DB      TLINE
        DB      BLINE - $14

  ; ****** "DRAW" text string
DRAW:
        DB      'D','R','A','W'
  ; terminated by next byte below ...

  ; ****** BULLET LIMITS
BULLMT:
        DB      $00
        DB      $9F
        DB      ALINE
        DB      BLINE - $01

  ; *** MACRO for BULlet TABle
BN      macro   DX, ARMX, DY, ARMY
        DW      DX
        DB      ARMX
        DW      DY
        DB      ARMY
        endm

  ; *** Bullet trajectories for 7 arms
BULTAB:
        BN      $0300, $0F, $0300, $0F
        BN      $0400, $0F, $0200, $0C
        BN      $0400, $0F, $0100, $0B
        BN      $0400, $0F, $0000, $08
        BN      $0400, $0F, $FF00, $06
        BN      $0400, $0F, $FE00, $04
        BN      $0300, $0F, $FD00, $03

  ; *** fixed Cactus Y location TABles
LFTAB:
        DEF5    $48, $16, $2C, $43, $0E

RFTAB:
        DEF5    $12, $44, $28, $0D, $3F

GFCOLS: ; (with # + #'s lighter, and # - #'s darker)
if BALLY ; ### conditional assembly ###
        DB      $9D             ; # + # light GREEN (Border, Tree)
        DB      $76             ; # + # TAN (Faces, Score, Cactus)
        DB      $FC             ; # + # BLUE (Left Cowboy)
        DB      $87             ; # = #'s not changed
        DB      $9D             ; # + # light GREEN
        DB      $76             ; # + # TAN
        DB      $6C             ; # = # RED (Right Cowboy)
        DB      $87             ; # = # YELLOW (Playfield)
endif ; ##### end conditional ######### or...
if WHITE ; ### conditional assembly ###
        DB      $9B             ; # . # GREEN (color 3 left)
        DB      $76             ; # + # TAN (color 2)
        DB      $FC             ; # + # BLUE (color 1)
        DB      $87             ; # = # YELLOW (color 0)
        DB      $9B             ; # . # GREEN (color 3 right)
        DB      $76             ; # + # TAN (color 2)
        DB      $6C             ; # = # RED (color 1)
        DB      $87             ; # = # YELLOW (color 0)
endif ; ##### end conditional ######### or...
if ASTRO ; ### conditional assembly ###
        DB      $A2             ; # - # darker GREEN
        DB      $7D             ; # - # darker TAN
        DB      $0B             ; # - # darker BLUE
        DB      $87             ; # = # YELLOW
        DB      $A2             ; # - # darker GREEN
        DB      $7D             ; # - # darker TAN
        DB      $6C             ; # = # RED
        DB      $87             ; # = # YELLOW
endif ; ##### end conditional #########

  ; for CT5-7, CNT, SEMI4S, OPOT1-4, KEYSEX, and OSW1-2
SINIT:        
        DEF8    $06, $06, $00, $00, $00, $30, $30, $00
        DEF4    $00, $80, $0F, $0F

  ; ****** COLOR MASK equates
; NUMB    EQU     00000111B       ; color 1 on color 3 (not used)

BULT    EQU     00001011B       ; color 2 on color 3

TIME    EQU     00001011B       ; color 2 on color 3

LARGE   EQU     00001011B       ; color 2 on color 3

LARG2   EQU     00001100B       ; color 3 on color 0

  ; **********************
  ; * GUN FIGHT PATTERNS *
  ; **********************

  ; ****** PATTERN TABLES
ARMTBL:
        DW      ARM0            ; fully CW
        DW      ARM1
        DW      ARM2
        DW      ARM3
        DW      ARM4
        DW      ARM5
        DW      ARM6            ; fully CCW

TREE:           ; one Monochrome pattern (drawn in color 3)
        DEF2    $01, $11        ; Width = 1 byte, Height = 17
        DB      00001000B       ; . . . . X . . .
        DB      00011100B       ; . . . X X X . .
        DB      00111110B       ; . . X X X X X . 
        DB      01101011B       ; . X X . X . X X 
        DB      00001000B       ; . . . . X . . . 
        DB      00001000B       ; . . . . X . . . 
        DB      00111100B       ; . . X X X X . . 
        DB      01111110B       ; . X X X X X X . 
        DB      10101001B       ; X . X . X . . X 
        DB      00001000B       ; . . . . X . . . 
        DB      00111100B       ; . . X X X X . . 
        DB      01111110B       ; . X X X X X X . 
        DB      11101011B       ; X X X . X . X X 
        DB      10001001B       ; X . . . X . . X 
        DB      00001000B       ; . . . . X . . . 
        DB      00011100B       ; . . . X X X . . 
        DB      10101110B       ; X . X . X X X . 

ARM0:           ; all the rest are Color patterns (except Cactus)
        DEF2    $0A, $0A        ; Offsets = 10 left, 10 down
        DEF2    $02, $05        ; Width = 2 bytes, Height = 5
        DEF2    $40, $00        ; 1 . . . . . . . [arms in color 1 only]
        DEF2    $51, $00        ; 1 1 . 1 . . . .
        DEF2    $04, $00        ; . . 1 . . . . .
        DEF2    $01, $00        ; . . . 1 . . . .
        DEF2    $00, $40        ; . . . . 1 . . .

ARM1:
        DEF2    $0A, $0A        ; Offsets = 10 left, 10 down
        DEF2    $02, $03        ; Width = 2 bytes, Height = 3
        DEF2    $50, $00        ; 1 1 . . . . . .
        DEF2    $14, $00        ; . 1 1 . . . . .
        DEF2    $01, $40        ; . . . 1 1 . . .

ARM2:
        DEF2    $0A, $0A        ; Offsets = 10 left, 10 down
        DEF2    $02, $02        ; Width = 2 bytes, Height = 2
        DEF2    $54, $00        ; 1 1 1 . . . . .
        DEF2    $55, $40        ; 1 1 1 1 1 . . .

ARM3:
        DEF2    $0A, $07        ; Offsets = 10 left, 7 down
        DEF2    $02, $04        ; Width = 2 bytes, Height = 4
        DEF2    $10, $00        ; . 1 . . . . . .
        DEF2    $05, $40        ; . . 1 1 1 . . .
        DEF2    $54, $00        ; 1 1 1 . . . . .
        DEF2    $50, $00        ; 1 1 . . . . . .

ARM4:
        DEF2    $0A, $06        ; Offsets = 10 left, 6 down
        DEF2    $02, $05        ; Width = 2 bytes, Height = 5
        DEF2    $00, $40        ; . . . . 1 . . .
        DEF2    $45, $00        ; 1 . 1 1 . . . .
        DEF2    $10, $00        ; . 1 . . . . . .
        DEF2    $50, $00        ; 1 1 . . . . . .
        DEF2    $40, $00        ; 1 . . . . . . .

ARM5:
        DEF2    $0A, $05        ; Offsets = 10 left, 5 down
        DEF2    $02, $06        ; Width = 2 bytes, Height = 6
        DEF2    $00, $40        ; . . . . 1 . . .
        DEF2    $01, $00        ; . . . 1 . . . .
        DEF2    $05, $00        ; . . 1 1 . . . .
        DEF2    $14, $00        ; . 1 1 . . . . .
        DEF2    $54, $00        ; 1 1 1 . . . . .
        DEF2    $50, $00        ; 1 1 . . . . . .

ARM6:
        DEF2    $0A, $05        ; Offsets = 10 left, 5 down
        DEF2    $01, $05        ; Width = 1 byte, Height = 5
        DB      $01             ; . . . 1       
        DB      $44             ; 1 . 1 .       
        DB      $10             ; . 1 . .       
        DB      $40             ; 1 . . .       
        DB      $40             ; 1 . . .       

  ; **** NOTE ****
  ; THE FOLLOWING PATTERNS ARE CONSTRAINED TO EXIST ON THE
  ; PAGE.  <ORG ($+255) AND 0FF00H> WILL DO IT FOR EXPERIMENTAL
  ; PATTERNS ARE: LEG0, LEG1, LEG2, KIL1, KIL2
  ; ***
LEG0:
        DB      LEG1 AND $FF    ; Next in animation cycle
        DB      $04             ; Delay time setting = 4
        DEF2    $00, $0F        ; Offsets = 0, 15 down
        DEF2    $03, $05        ; Width = 3 bytes, Height = 5
        DEF3    $01, $55, $00   ; . . . 1 1 1 1 1 . . . . [... legs too]
        DEF3    $05, $45, $40   ; . . 1 1 1 . 1 1 1 . . .     
        DEF3    $15, $01, $40   ; . 1 1 1 . . . 1 1 . . .     
        DEF3    $50, $01, $40   ; 1 1 . . . . . 1 1 . . .     
        DEF3    $15, $00, $54   ; . 1 1 1 . . . . 1 1 1 .     

LEG1:
        DB      LEG2 AND $FF    ; Next in cycle
        DB      $04             ; Delay time
        DEF2    $02, $0F        ; Offsets = 2 left, 15 down
        DEF2    $02, $05        ; Width = 2 bytes, Height = 5
        DEF2    $15, $50        ; . 1 1 1 1 1 . . 
        DEF2    $54, $50        ; 1 1 1 . 1 1 . . 
        DEF2    $50, $50        ; 1 1 . . 1 1 . . 
        DEF2    $50, $50        ; 1 1 . . 1 1 . . 
        DEF2    $55, $15        ; 1 1 1 1 . 1 1 1 

LEG2:
        DB      LEG0 AND $FF    ; LOOP back to start
        DB      $04             ; Delay
        DEF2    $03, $0F        ; Offsets = 3 left, 15 down
        DEF2    $02, $05        ; Width = 2 bytes, Height = 5
        DEF2    $55, $00        ; 1 1 1 1 . . . .  
        DEF2    $15, $00        ; . 1 1 1 . . . .  
        DEF2    $15, $00        ; . 1 1 1 . . . .  
        DEF2    $14, $00        ; . 1 1 . . . . .  
        DEF2    $05, $40        ; . . 1 1 1 . . .  

KIL1:
        DB      KIL2 AND $FF    ; Next in animation cycle
        DB      $14             ; Delay time = 20 tics (1/3 sec.)
        DEF2    $00, $01        ; Offsets = 0, 1 down
        DEF2    $04, $13        ; Width = 4 bytes, Height = 19
        DEF4    $01, $10, $00, $00  ; . . . 1 . 1 . . . . . . . . . .           
        DEF4    $45, $54, $40, $00  ; 1 . 1 1 1 1 1 . 1 . . . . . . .           
        DEF4    $55, $55, $40, $00  ; 1 1 1 1 1 1 1 1 1 . . . . . . .           
        DEF4    $0A, $A8, $00, $00  ; . . 2 2 2 2 2 . . . . . . . . .           
        DEF4    $0A, $A2, $00, $01  ; . . 2 2 2 2 . 2 . . . . . . . 1           
        DEF4    $0A, $AA, $80, $14  ; . . 2 2 2 2 2 2 2 . . . . 1 1 .           
        DEF4    $02, $AA, $00, $50  ; . . . 2 2 2 2 2 . . . . 1 1 . .           
        DEF4    $00, $A8, $05, $40  ; . . . . 2 2 2 . . . 1 1 1 . . .           
        DEF4    $05, $55, $54, $00  ; . . 1 1 1 1 1 1 1 1 1 . . . . .           
        DEF4    $15, $55, $50, $00  ; . 1 1 1 1 1 1 1 1 1 . . . . . .           
        DEF4    $54, $55, $50, $00  ; 1 1 1 . 1 1 1 1 1 1 . . . . . .           
        DEF4    $50, $05, $54, $00  ; 1 1 . . . . 1 1 1 1 1 . . . . .           
        DEF4    $50, $01, $55, $00  ; 1 1 . . . . . 1 1 1 1 1 . . . .           
        DEF4    $10, $01, $55, $40  ; . 1 . . . . . 1 1 1 1 1 1 . . .           
        DEF4    $10, $00, $05, $50  ; . 1 . . . . . . . . 1 1 1 1 . .           
        DEF4    $00, $00, $01, $50  ; . . . . . . . . . . . 1 1 1 . .           
        DEF4    $00, $00, $00, $40  ; . . . . . . . . . . . . 1 . . .           
        DEF4    $00, $00, $01, $40  ; . . . . . . . . . . . 1 1 . . .           
        DEF4    $00, $00, $00, $54  ; . . . . . . . . . . . . 1 1 1 .           
  
KIL2:
        DB      KIL2 AND $FF    ; LOOP animation cycle here
        DB      $3C             ; Delay = 60 tics (one second)
        DEF2    $00, $0D        ; Offsets = 0, 13 down
        DEF2    $04, $07        ; Width = 4 bytes, Height = 7
        DEF4    $01, $10, $00, $00  ; . . . 1 . 1 . . . . . . . . . .           
        DEF4    $45, $54, $40, $00  ; 1 . 1 1 1 1 1 . 1 . . . . . . .           
        DEF4    $55, $55, $40, $00  ; 1 1 1 1 1 1 1 1 1 . . . . . . .           
        DEF4    $0A, $A8, $00, $00  ; . . 2 2 2 2 2 . . . . . . . . .           
        DEF4    $0A, $88, $15, $01  ; . . 2 2 2 . 2 . . 1 1 1 . . . 1           
        DEF4    $16, $A5, $55, $41  ; . 1 1 2 2 2 1 1 1 1 1 1 1 . . 1           
        DEF4    $15, $55, $55, $55  ; . 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1           

CACTUS:         ;  another Monochrome pattern (drawn in color 2)
        DEF2    $01, $0C        ; Width = 1 byte, Height = 12
        DB      00100000B       ; . . X . . . . . 
        DB      00110000B       ; . . X X . . . .
        DB      00111000B       ; . . X X X . . .
        DB      00110000B       ; . . X X . . . . 
        DB      10110010B       ; X . X X . . X .
        DB      11110010B       ; X X X X . . X .   
        DB      11110110B       ; X X X X . X X .
        DB      00111100B       ; . . X X X X . .  
        DB      00111100B       ; . . X X X X . .  
        DB      00110000B       ; . . X X . . . . 
        DB      00110000B       ; . . X X . . . .
        DB      00110000B       ; . . X X . . . .
 
  ; ****** "GOT ME" text string
GOTME:
        DB      'G','O','T',$20,'M','E'
  ; terminated by next byte below ...

NULPAT:
        DEF2    $00, $00        ; Offsets = 0, 0
        DEF2    $01, $01        ; Width = 1 byte, Height = 1
  ; note: pattern used is the first NULL byte below

GFBODY:         ; note: Hot Spot for parts displacement marked with ()'s
        DEF2    $00, $00        ; Offsets = 0, 0
        DEF2    $03, $0F        ; Width = 3 bytes, Height = 15
        DEF3    $00, $44, $00   ;(.). . . 1 . 1 . . . . .     
        DEF3    $11, $55, $10   ; . 1 . 1 1 1 1 1 . 1 . .     
        DEF3    $15, $55, $50   ; . 1 1 1 1 1 1 1 1 1 . .     
        DEF3    $02, $AA, $00   ; . . . 2 2 2 2 2 . . . .     
        DEF3    $02, $A2, $00   ; . . . 2 2 2 . 2 . . . .     
        DEF3    $02, $AA, $80   ; . . . 2 2 2 2 2 2 . . .     
        DEF3    $00, $AA, $00   ; . . . . 2 2 2 2 . . . .     
        DEF3    $00, $A8, $00   ; . . . . 2 2 2 . . . . .     
        DEF3    $15, $55, $00   ; . 1 1 1 1 1 1 1 . . . .     
        DEF3    $55, $55, $50   ; 1 1 1 1 1 1 1 1 1 1 . .     
        DEF3    $51, $55, $50   ; 1 1 . 1 1 1 1 1 1 1 . .     
        DEF3    $41, $55, $00   ; 1 . . 1 1 1 1 1 . . . .     
        DEF3    $41, $55, $00   ; 1 . . 1 1 1 1 1 . . . .     
        DEF3    $45, $55, $00   ; 1 . 1 1 1 1 1 1 . . . .     
        DB      $01, $55        ; . . . 1 1 1 1 1 [below]

WAGPAT:
        DEF2    $00, $00        ; Offsets = 0, 0
        DEF2    $04, $16        ; Width = 4 bytes, Height = 22
        DEF4    $00, $05, $50, $00  ;(.). . . . . 1 1 1 1 . . . . . .           
        DEF4    $00, $55, $55, $00  ; . . . . 1 1 1 1 1 1 1 1 . . . .           
        DEF4    $01, $55, $55, $40  ; . . . 1 1 1 1 1 1 1 1 1 1 . . .           
        DEF4    $05, $55, $55, $50  ; . . 1 1 1 1 1 1 1 1 1 1 1 1 . .           
        DEF4    $15, $54, $15, $54  ; . 1 1 1 1 1 1 . . 1 1 1 1 1 1 .           
        DEF4    $15, $50, $05, $54  ; . 1 1 1 1 1 . . . . 1 1 1 1 1 .           
        DEF4    $15, $40, $01, $54  ; . 1 1 1 1 . . . . . . 1 1 1 1 .           
        DEF4    $15, $40, $01, $54  ; . 1 1 1 1 . . . . . . 1 1 1 1 .           
        DEF4    $15, $50, $05, $54  ; . 1 1 1 1 1 . . . . 1 1 1 1 1 .           
        DEF4    $05, $54, $15, $50  ; . . 1 1 1 1 1 . . 1 1 1 1 1 . .           
        DEF4    $01, $55, $55, $40  ; . . . 1 1 1 1 1 1 1 1 1 1 . . .           
        DEF4    $00, $55, $55, $00  ; . . . . 1 1 1 1 1 1 1 1 . . . .           
        DEF4    $00, $15, $54, $00  ; . . . . . 1 1 1 1 1 1 . . . . .           
        DEF4    $02, $AA, $AA, $80  ; . . . 2 2 2 2 2 2 2 2 2 2 . . .           
        DEF4    $00, $AA, $AA, $00  ; . . . . 2 2 2 2 2 2 2 2 . . . .           
        DEF4    $12, $AA, $AA, $84  ; . 1 . 2 2 2 2 2 2 2 2 2 2 . 1 .           
        DEF4    $10, $A8, $2A, $04  ; . 1 . . 2 2 2 . . 2 2 2 . . 1 .           
        DEF4    $10, $20, $08, $04  ; . 1 . . . 2 . . . . 2 . . . 1 .           
        DEF4    $52, $AA, $AA, $85  ; 1 1 . 2 2 2 2 2 2 2 2 2 2 . 1 1           
        DEF4    $10, $20, $08, $04  ; . 1 . . . 2 . . . . 2 . . . 1 .           
        DEF4    $10, $00, $00, $04  ; . 1 . . . . . . . . . . . . 1 .           
        DEF4    $10, $00, $00, $04  ; . 1 . . . . . . . . . . . . 1 .           

  ; ****** Checksum FUDGe #4
FUDG4:
if BALLY ; ### conditional assembly ###
        DB      $37             ; # # # CHECKSUM #4
endif ; ##### end conditional ######### or...
if WHITE ; ### conditional assembly ###
        DB      $3D             ; # # # CHECKSUM #4
endif ; ##### end conditional ######### or...
if ASTRO ; ### conditional assembly ###
        DB      $00             ; # 0 # CHECKSUM #4
endif ; ##### end conditional #########

  ; *** Music Scores for GunFight
MSET:
        MASTER  $11
        VOLUME  $09, $00
        RET

  ; ****** HOME ON THE RANGE music score
HOME:
        CALL    MSET
        NOTE1   $24, G1
        NOTE1   $0C, F1
        NOTE1   $12, E1
        NOTE1   $06, D1
        NOTE1   $24, E1
        QUIET

  ; ****** TAPS music score
TAPS:
        CALL    MSET
        NOTE1   $12, C1
        NOTE1   $06, C1
        NOTE1   $24, F1
        NOTE1   $12, C1
        NOTE1   $06, F1
        NOTE1   $24, A1
        QUIET

  ; ****** FUNERAL music score
FUNERL:
        CALL    MSET
        NOTE1   $18, A0
        NOTE1   $12, A0
        NOTE1   $06, A0
        NOTE1   $18, A0
        NOTE1   $12, C1
        NOTE1   $06, B0
        NOTE1   $12, B0
        NOTE1   $06, A0
        NOTE1   $12, A0
        NOTE1   $06, GS0
        NOTE1   $12, A0
        QUIET

GUNSHOT:
        OUTPUT  $18, $F0, $F5, $FD, $FF, $00, $3F, $FF, $EF
        LEGSTA
        VOLUME  $FF, $3F
        REST    $05
        NOTE1   $05, $8F
        NOTE1   $05, $4C
        QUIET

  ;
  ;       END OF GUNFIGHT PROGRAM
  ;

; TWENTY-ONE FILLER BYTES BETWEEN GUNFIGHT AND END OF ROM
; L1FEF:                        ; LEND EQU LASTB
        DB      $FF,$FF,$FF,$FF,$FF
        DB      $FF,$FF,$FF,$FF,$FF
        DB      $FF,$FF,$FF,$FF,$FF
        DB      $FF,$FF,$FF,$FF,$FF,$FF
